
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <float.h>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_statistics_double.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_errno.h>


#define NUM_THREADS 4

#define AIC 0
#define BIC 1
#define MDL 2
#define LIC 3

#define MIN_LOGLAMBDA_COEFF log(0.001)
#define MIN_LOGLAMBDA_COVAR log(0.001)
#define MIN_LOGLAMBDA_COEFF_ALASSO log(0.0001)
#define MIN_LOGLAMBDA_COVAR_ALASSO log(0.0001)
#define MAX_LOGLAMBDA 100.0
#define SUBGRID_WIDTH 3
#define GRAD_STEP 5.0e-4
#define PARAM_SCAD 3.7// As in Fan and Li
#define PARAM_MCP 1.5// 1.4; 1.7; // > 1.0, as in Zhang

//Penalties (some are shared by coefficients and covariance)
#define NONE 0
#define LASSO 1
#define ADALASSO 2
#define GROUP_ADALASSO 3
#define SCAD 4
#define MCP 5
#define ORACLE 8
#define LASSO_INVCOVAR 9
#define ADALASSO_INVCOVAR 10
#define SCAD_INVCOVAR 11
#define MCP_INVCOVAR 12
#define GRAPHICAL_LASSO 15
#define GRAPHICAL_ADALASSO 16
#define GRAPHICAL_SCAD 17
#define GRAPHICAL_MCP 18
#define GRAPHICAL_ORACLE 19
#define ORACLE_INVCOVAR 20
#define INVCOVAR_GRAPHICAL(X)  ((X==GRAPHICAL_LASSO)||(X==GRAPHICAL_ADALASSO)||(X==GRAPHICAL_SCAD)||(X==GRAPHICAL_MCP)||(X==GRAPHICAL_ORACLE))
#define INVCOVAR_REG(X)  ((X==LASSO_INVCOVAR)||(X==ADALASSO_INVCOVAR)||(X==SCAD_INVCOVAR)||(X==MCP_INVCOVAR)||(X==ORACLE_INVCOVAR)||(INVCOVAR_GRAPHICAL(X)))


typedef struct msvar_parameters
{
  gsl_matrix * Pr;  //Probability matrix
  gsl_matrix * A;   //Matrix of coefficients for all regimes
  gsl_matrix * SigmaL;  //Cholesky factor of covariance matrix, all regimes
  gsl_matrix * Sigma_inv;  //Inverse, all regimes
  size_t D; //Data dimension
  size_t M; //Current Number of regimes
  size_t T; //Number of observations
  size_t p; //Lag order
  size_t initial_state; //Initial state of the Markov chain
  size_t switching_coeff;   //Indicator for regime-dependent coefficients
  size_t switching_covar;   //Indicator for regime-dependent covariances
  double loglik;
  double pen;
} msvar_parameters;

typedef struct em_workspace
{
    //Forward and backward recursion auxiliary memory
    gsl_vector * temp_vec;
    gsl_vector * mean_vec;
    gsl_vector * temp_vecm;
    gsl_vector * temp_vecx;
    gsl_vector_view temp_vecm_v;
    gsl_vector_view temp_vecx_v;
    
    // Transition probabs auxiliary memory
    gsl_matrix * tmpm;
    gsl_matrix_view tmpm_v;
    
    //Coefficients auxiliary memory
    gsl_matrix * tmp;
    gsl_matrix * Ctmp;
    gsl_matrix * Ctmp1HH;
    gsl_matrix * Ctmp2;
    gsl_matrix * C1;
    gsl_matrix * C2;
    gsl_vector * b;
    
    //Covariance auxiliary memory
    gsl_matrix * U;
    gsl_matrix * Usw;
    gsl_matrix * tmpc;
    gsl_matrix * tmpcsw;
    gsl_matrix_view U_v;
    gsl_matrix_view tmpc_v;
    gsl_eigen_symm_workspace * eigw;
    
    // EM auxiliary memory
    gsl_matrix * xi_smooth;
    gsl_matrix * xi_filter;
    gsl_matrix * xi_tt1;
    gsl_matrix_view xi_smooth_v;
    gsl_matrix_view xi_filter_v;
    gsl_matrix_view xi_tt1_v;
    
    msvar_parameters * estimates_prev;
    
} em_workspace;

typedef struct fista_workspace
{
    gsl_matrix * Z;
    gsl_matrix * B_prev;
    gsl_matrix * B_hat_init;
    gsl_matrix * Grad;
    
    gsl_matrix * XtY;
    gsl_matrix * XtXZ;
    gsl_matrix * XtX;
    gsl_matrix * XB;
    gsl_matrix * XZ;
    
    gsl_matrix * BmZ;
    
} fista_workspace;

typedef struct regu_workspace
{
    double ** A_mm;
    fista_workspace * fista_wspace;
    //fista_workspace_sp * fista_wspace_sp;
    gsl_matrix * Li;
    gsl_matrix * R;
    
    gsl_matrix * Ys;
    gsl_matrix * E;
    gsl_matrix * lambda_weights_tr;
    gsl_matrix * groups_tr;
    gsl_vector * eval_vec;
    gsl_eigen_symm_workspace * eigw;

} regu_workspace;

typedef struct regu_cov_matrices
{
    gsl_matrix * S;
    gsl_matrix * S_prev;
    gsl_matrix * S_0;
    gsl_matrix * S_inv;
    gsl_matrix * S_0_inv;
    gsl_matrix * Tmp;
    gsl_matrix * Th;
    gsl_matrix * U;
    gsl_matrix * S_tmp;
    
    gsl_matrix * W11;
    gsl_matrix * W;
    gsl_matrix * Beta;
    gsl_matrix * beta_prev;
    gsl_matrix * w;
    gsl_matrix * w_prev;
    gsl_matrix * s;
    
} regu_cov_matrices;

typedef struct regu_cov_workspace
{
    double ** S_mm;
    double ** S_inv_mm;
    regu_cov_matrices * regu_cov_mats;
    //regu_cov_spmatrices * regu_cov_spmats;
    //gsl_matrix * Se;
    gsl_vector * eval_vec;
    gsl_eigen_symm_workspace * eigw;
    gsl_eigen_symmv_workspace * eigvw;
    
} regu_cov_workspace;


int gsl_matrix_scale_rows_0 ( gsl_matrix * a, const gsl_vector * x );
double gsl_vector_sum_0 (const gsl_vector * a);


double scad_penalty ( const gsl_matrix * A, double lambda, double a, size_t isCovar );

double mcp_penalty ( const gsl_matrix * A, double lambda, double gamma, size_t isCovar );

double l1_norm ( const gsl_matrix * A, const gsl_matrix * weights, size_t isCovar );

double group_l2_norm ( const gsl_matrix * A, const gsl_matrix * groups, const gsl_matrix * weights );

double compute_penalty ( size_t penalty, const gsl_matrix * B, double lambda, const gsl_matrix * weights, double a, double gamma, double N, const gsl_matrix * groups, size_t isCovar );

double compute_penalties ( const msvar_parameters * estimates, size_t coeff_penalty, size_t covar_penalty, double coeff_lambda, double covar_lambda, const gsl_matrix * coeff_lambda_weights, const gsl_matrix * covar_lambda_weights, double a, double gamma, double N, const gsl_matrix * groups );

double inner_product ( const gsl_matrix * A, const gsl_matrix * B );

double norm_diff ( const gsl_vector * a, const gsl_vector * b );

double frobenius_norm_diff ( const gsl_matrix * a, const gsl_matrix * b );

int kronecker_product ( CBLAS_TRANSPOSE_t TransA, double alpha, const gsl_matrix * A, const gsl_matrix * B, gsl_matrix * C );

int msvar_parameters_read_constants ( const char * filename, msvar_parameters * parameters );

int msvar_parameters_read_matrices ( const char * filename, msvar_parameters * parameters );

int msvar_parameters_save ( const char * filename, const msvar_parameters * parameters );

int generate_data ( const msvar_parameters * parameters, gsl_matrix * Y, gsl_matrix * YT, gsl_vector * mean_vec, const gsl_rng * rng, gsl_vector * regimes );

int generate_initial_transition_probabilities ( msvar_parameters * estimates, const gsl_rng * rng );

int forward_recursion ( em_workspace * em_wspace, const gsl_matrix * X, const gsl_matrix * Y, msvar_parameters * estimates );

int backward_recursion ( em_workspace * em_wspacea, const gsl_matrix * Ft );

int estimate_transition_probabilities ( em_workspace * em_wspace, msvar_parameters * estimates, double C );

int estimate_coefficients ( em_workspace * em_wspace, const gsl_matrix * Y, const gsl_matrix * X, const gsl_vector * y, msvar_parameters * estimates ) ;

int estimate_switching_coefficients ( em_workspace * em_wspace, const gsl_matrix * Y, const gsl_matrix * X, msvar_parameters * estimates ) ;

int estimate_sparse_coefficients ( em_workspace * em_wspace, regu_workspace * regu_wspace,
                                   const gsl_matrix * Y, const gsl_matrix * X, msvar_parameters * estimates,
                                   size_t penalty, double lambda, gsl_matrix * lambda_weights,
                                 size_t max_iters, double tol, double L_c, const gsl_matrix * groups );

int estimate_covariance_precision ( em_workspace * em_wspace, const gsl_matrix * Y, const gsl_matrix * X, msvar_parameters * estimates );

int estimate_switching_covariance_precision ( em_workspace * em_wspace, const gsl_matrix * Y, const gsl_matrix * X, msvar_parameters * estimates, int m, size_t chol );

int estimate_regularized_covariance_precision ( em_workspace * em_wspace, regu_cov_workspace * regu_cov_wspace, const gsl_matrix * Y, const gsl_matrix * X,
                    msvar_parameters * estimates, size_t penalty, double lambda, gsl_matrix * lambda_weights, size_t max_iters, double tol );

int estimate_sparse_covariance_precision ( em_workspace * em_wspace, regu_cov_workspace * regu_cov_wspace, const gsl_matrix * Y, const gsl_matrix * X,
                    msvar_parameters * estimates, size_t penalty, double lambda, gsl_matrix * lambda_weights, size_t max_iters, double tol );

int compute_initial_estimates ( em_workspace * em_wspace, const gsl_matrix * Y, const gsl_matrix * X, msvar_parameters * estimates, int sort_i );

int expectation_maximization ( em_workspace * em_wspace, regu_workspace * regu_wspace, regu_cov_workspace * regu_cov_wspace, msvar_parameters * estimates,
                               const gsl_matrix * Y, const gsl_matrix * X, const gsl_vector * y, size_t em_max_iters, size_t max_iters, double tol, size_t coeff_penalty,
                               size_t covar_penalty, double coeff_lambda, double covar_lambda, gsl_matrix * coeff_lambda_weights, gsl_matrix * covar_lambda_weights,
                               double L_c, const gsl_matrix * groups );

int compute_ar_sum ( size_t m, size_t t, const msvar_parameters * parameters, const gsl_matrix * X, const gsl_matrix * Y, gsl_vector * Y_t );

int compute_regression_matrix_vector ( regu_workspace * regu_wspace, const gsl_matrix * Y, const msvar_parameters * estimates, gsl_matrix * X, gsl_vector * y,
                                       double * L_c );

void em_replications ( const msvar_parameters * parameters, msvar_parameters * estimates,
                       gsl_matrix * estimates_matrix_A, gsl_matrix * estimates_matrix_Pr, gsl_matrix * estimates_matrix_Sigma,
                       gsl_matrix * estimates_matrix_loglik, gsl_matrix * estimates_matrix_smooth_p, gsl_matrix * estimates_matrix_lambdas,
                       size_t replications, size_t em_max_iters, size_t max_iters, double tol, const gsl_rng * rng, size_t first_datapoint_t, const char * oracle_dir,
                       size_t coeff_penalty, double coeff_loglambda_step,
                       size_t covar_penalty, double covar_loglambda_step, size_t k_folds, const char * groups_file );

void set_mem_matrices ( size_t i, int ic_c, const msvar_parameters * estimates,
                        double ** estimates_mem_A, double ** estimates_mem_Pr, double ** estimates_mem_Sigma, int is_precision );

void compute_df ( const msvar_parameters * estimates, int * params_A, int * params_S, int * params_A_min, int * params_S_min, gsl_vector * df_m,
                  size_t coeff_penalty, size_t covar_penalty, double tol, const gsl_matrix * coeff_lambda_weights, const gsl_matrix * groups );

double compute_ic ( em_workspace * em_wspace, const msvar_parameters * estimates, const gsl_matrix * Y, size_t ic, size_t params_A, size_t params_S,
                    gsl_vector * T_m, const gsl_vector * df_m );

void compute_information_descent_optimum ( em_workspace * em_wspace, regu_workspace * regu_wspace, regu_cov_workspace * regu_cov_wspace, msvar_parameters * estimates,
                                const gsl_matrix * A_mle, const gsl_matrix * SL_mle, const gsl_matrix * S_inv_mle, const gsl_matrix * Pr_mle,
                                const gsl_matrix * Y_tr, const gsl_matrix * Y_vd, gsl_matrix * X, gsl_vector * y, double ** ic_arr, size_t ic_c,
                                size_t coeff_penalty, double coeff_loglambda_step, size_t covar_penalty, double covar_loglambda_step,
                                gsl_matrix * coeff_lambda_weights, gsl_matrix * covar_lambda_weights, size_t em_max_iters, size_t max_iters,
                                double tol, int * coeff_lambda_ind, int * covar_lambda_ind, double grid_width, const gsl_matrix * groups, double resol_, double L_c );

double log_likelihood ( em_workspace * em_wspace, const gsl_matrix * Y, const msvar_parameters * estimates, size_t t_1, const gsl_vector * s_0 );

void min_max_sym_eigenvalue ( gsl_matrix * M, gsl_vector * eval_vec, gsl_eigen_symm_workspace * eigw, double * lmin, double * lmax);

void em_workspace_alloc ( em_workspace * emd, const msvar_parameters * estimates );

void em_workspace_free ( em_workspace * emd, size_t M );

void msvar_parameters_alloc_matrices ( msvar_parameters * parameters );

void msvar_parameters_free_matrices ( msvar_parameters * parameters );

int read_alloc_groups ( const char * groups_file, gsl_matrix ** groups, size_t A_cols );

void printf_matrix ( const char * info, const gsl_matrix * m );

void printf_vector ( const char * info, const gsl_vector * m );

int fprintf_matrix ( const char * filename, const char * info, const gsl_matrix * m );

int fprintf_vector ( const char * filename, const char * info, const gsl_vector * m );

int fscanf_matrix ( const char * filename, const char * info, gsl_matrix * m );

int fscanf_vector ( const char * filename, const char * info, gsl_vector * m );

/*--*/

void fista_workspace_alloc ( fista_workspace * fista_wspace, size_t T, size_t size1, size_t size2 );

void fista_workspace_free ( fista_workspace * fista_wspace );

void regu_workspace_alloc ( regu_workspace * regu_wspace, const msvar_parameters * parameters, size_t penalty, size_t g_size );

void regu_workspace_free ( regu_workspace * regu_wspace, const msvar_parameters * parameters );

void regu_cov_matrices_alloc ( regu_cov_matrices * regu_cov_mats, const msvar_parameters * parameters, size_t penalty );

void regu_cov_matrices_free ( regu_cov_matrices * regu_cov_mats, size_t penalty );

void regu_cov_workspace_alloc ( regu_cov_workspace * regu_cov_wspace, const msvar_parameters * parameters, size_t penalty );

void regu_cov_workspace_free ( regu_cov_workspace * regu_cov_wspace, size_t penalty, const msvar_parameters * parameters );

void set_penalty_weights ( size_t D, const gsl_matrix * A, const gsl_matrix * SL, const gsl_matrix * S_inv, size_t coeff_penalty, size_t covar_penalty,
                         gsl_matrix * coeff_lambda_weights, gsl_matrix * covar_lambda_weights, const gsl_matrix * groups );

void set_coeff_penalty_weights ( const gsl_matrix * A, size_t D, size_t coeff_penalty, gsl_matrix * coeff_lambda_weights, double coeff_lambda, const gsl_matrix * groups );

void set_covar_penalty_weights ( const gsl_matrix * SL, const gsl_matrix * S_inv, size_t covar_penalty, gsl_matrix * covar_lambda_weights, double covar_lambda );

void fista_estimation ( fista_workspace * fista_wspace, const gsl_matrix * X, const gsl_matrix * Y, const gsl_matrix * groups, double lambda,
                        const gsl_matrix * lambda_weights, gsl_matrix * B_init_hat, double L_0, size_t max_iters, double tol, size_t penalty, double N );

void covariance_optimization ( regu_cov_workspace * regu_cov_wspace, const gsl_matrix * Sigma, gsl_matrix * S_init, msvar_parameters * estimates, double lambda,
                               const gsl_matrix * lambda_weights, size_t penalty, size_t max_iters, double tol, double L_0, double N );

void precision_optimization ( regu_cov_workspace * regu_cov_wspace, const gsl_matrix * Sigma, gsl_matrix * S_init, msvar_parameters * estimates, double lambda,
                              const gsl_matrix * lambda_weights, size_t penalty, size_t max_iters, double tol, double L_0, double N );

double compute_min_delta ( double a, double c, size_t max_iters, double tol, int isPrecision );
double h ( double x, double a, int is_precision );
double dh ( double x, double a, int is_precision );

void proximal_operator ( gsl_matrix * B, size_t penalty, double lambda, const gsl_matrix * lambda_weights, double eta, double a_param, double gamma_param,
                         double c_pen, double N, const gsl_matrix * groups, int isCovar );

void p_operator_lasso ( gsl_matrix * B_p, double lambda, const gsl_matrix * lambda_weights, double eta, double N, int isCovar );

void p_operator_group_lasso ( gsl_matrix * B_p, double lambda, const gsl_matrix * lambda_weights, double eta, double N, const gsl_matrix * groups );

void p_operator_scad ( gsl_matrix * B_p, double lambda, double eta, double a, double c_pen, double N, int isCovar );

void p_operator_mcp ( gsl_matrix * B_p, double lambda, double eta, double gamma, double c_pen, double N, int isCovar );

double Quad ( fista_workspace * fista_wspace, const gsl_matrix * B, const gsl_matrix * Y, const gsl_matrix * groups,
              double lambda, const gsl_matrix * lambda_weights, double t, size_t penalty, double a, double gamma, double N );

double ls_objective_function ( fista_workspace * fista_wspace, const gsl_matrix * B, const gsl_matrix * Y, const gsl_matrix * X, const gsl_matrix * groups,
                            double lambda, const gsl_matrix * lambda_weights, size_t penalty, double a, double gamma, double N );

///////
////
//

