
#This script performs parameter estimation for 50 randomly generated samples from
#an MSVAR model, with parameter values given in Section 5 of our paper, corresponding to dimension d = 40.

#It generates Figure 1(a) and Figure 2(a) of our paper.
#The tuning parameters are provided to save computational time.

#Make sure the working directory is set to the path of this script.

#-----------------------------------------------------------------------------

#Compile and load functions if not done for this R session:

Rcpp::sourceCpp ( "../C/msvar_functions.cpp", showOutput = F, verbose = F )

#Read the pre-specified parameter values (Section 5 of the main paper) for an MSVAR model:
#The parameters are the same for all sample sizes.

#The AR coefficients and covariance matrices are stacked regime-wise so that
#first d rows correspond to regime 1, second d rows correspond to regime 2, etc.
#The first column of Ap below contains the intercepts.

#AR coefficients, transition probabilities, and covariance matrices:

Ap = as.matrix(read.table( "../msvar_data_d40_n300/mle_star/msvar_parameters_A"))
Prp = as.matrix(read.table( "../msvar_data_d40_n300/mle_star/msvar_parameters_Pr"))
Sp = as.matrix(read.table( "../msvar_data_d40_n300/mle_star/msvar_parameters_Sigma"))

#The following sets the sample sizes (n = 300, 400, 500), the penalties, and number of replications.

n_vec = c(300, 400, 500)
pen_vec = c("lasso", "alasso", "scad", "mcp", "mle_star")
num_reps = 50 #50 as in the main paper

#The following, for num_reps = 50, takes around 150 minutes on an AMD Ryzen Pro 9, and
#Note: it runs serially; this can be run in parallel along penalties and sample sizes

#For each randomly generated dataset from an MSVAR model, the code obtains the MPLE and MLE*
#(referred to as pen = "oracle" below, which means no penalty, as described in the main paper).

tm = system.time(
for(n in n_vec)
{
  for(pen in pen_vec)
  {
    print(paste0(pen, ", n = ", n))
    datadir = paste0("../msvar_data_d40_n", n, "/", pen, "/")
    
    lambs = matrix()
    
    #Read tuning parameters (already provided, obtained using BIC):
    
    if(pen != "mle_star")
    {
      lambs = as.matrix(read.table( paste0(datadir, "msvar_estimates_lambdas_matrix_input")))[1:num_reps,]
    } else
      pen = "oracle"
    
    #Do estimation:
    
    mats = msvar_replications(ar_coeff = Ap, covar = Sp, trans_probabs = Prp, n = n, M = 2, p = 1, coeff_penalty_str = pen, covar_penalty_str = pen, replications = num_reps, lambdas_matrix = lambs)
    
    #Save output matrices; each row corresponds to one replication:
    
    MASS::write.matrix( mats$ar_coeff_matrix, file = paste0(datadir, "msvar_estimates_coeff_matrix"))
    MASS::write.matrix( mats$covar_prec_matrix, file = paste0(datadir, "msvar_estimates_covar_prec_matrix"))
    MASS::write.matrix( mats$trans_probabs_matrix, file = paste0(datadir, "msvar_estimates_probabs_matrix"))
    MASS::write.matrix( mats$lambdas_matrix, file = paste0(datadir, "msvar_estimates_lambdas_matrix"))
  }
}
)
tm

#Generate and save plots and tables; this uses the output files generated above:
source("./utils/plots.R")

#The plots (used to create Figure 1 in the Supplement) in pdf format are now saved in the working directory.

##
#
