
#This script exemplifies how to use our code to fit an MSVAR model to a dataset.

#Make sure the working directory is set to the path of this script.

#-----------------------------------------------------------------------------

#Compile and load functions if not done for this R session:

Rcpp::sourceCpp ( "../C/msvar_functions.cpp", showOutput = F, verbose = F )

#Read the pre-specified parameter values for an MSVAR model to generate a synthetic dataset.
#If using a user-provided dataset, jump to line 44.

#Below, M is the number of regimes, p is the AR-lag order, and d is the data dimension.

#The AR coefficients and covariance matrices are stacked regime-wise so that
#first d rows correspond to regime 1, second d rows correspond to regime 2, etc.
#The first column of A below contains the intercepts.

#AR coefficients ( dM x (dp + 1) matrix), transition probabilities (M x M matrix), and covariance matrices (dM x d matrix).

A = as.matrix(read.table("../msvar_data_d5_n200/msvar_parameters_A"))
Pr = as.matrix(read.table("../msvar_data_d5_n200/msvar_parameters_Pr"))
S = as.matrix(read.table("../msvar_data_d5_n200/msvar_parameters_Sigma"))

#Generate a dataset of n = 200 observations from an MSVAR model with M = 2, p = 1, and d = 5.

YR = msvar_data_sample( ar_coeff = A, covar = S, trans_probabs = Pr, n = 200, seed = 1 )

#Plot generated data: n = 200 observations of 5-dimensional (d = 5) time series

ts.plot(YR$Y, col = 1:5, lty = 1:5)

#Indicate regime by adding points coloured by regime.
#This is possible for a simulated dataset since the regimes are known.

points(x = 1:200, y = rep(min(YR$Y), 200), col = c("purple", "pink")[YR$regimes], pch = 15)

#Fit an MSVAR model to the generated dataset.
#Possible options for the penalty are c("none", "lasso", "alasso", "scad", "mcp").
#BIC is used for tuning parameter (lambda) selection.

#Set the (n x d) data matrix. If user-provided, set Yt accordingly here:

Yt = YR$Y

#Fit an MSVAR model with M = 2 regimes and AR-lag order p = 1:

fit1 = msvar_fit ( Y = Yt, M = 2, p = 1, coeff_penalty_str = "lasso", covar_penalty_str = "lasso")

#Estimated AR coefficients are stacked, so that first d rows correspond to regime 1, etc.
#Similar for covariance matrices. To show each regime-specific parameter estimates:

reg_selection = 1 # Between 1 and M

#Regime-specific AR intercept (first column) and AR coefficients:

fit1$ar_coeff [(1 + (reg_selection - 1) * ncol(Yt)):(reg_selection * ncol(Yt)),]

#Regime-specific covariance matrix:

fit1$covar [(1 + (reg_selection - 1) * ncol(Yt)):(reg_selection * ncol(Yt)),]

#Alternatively, we can look at all the elements of the output:

fit1$ar_coeff
fit1$covar
fit1$trans_probabs

#The selected tuning parameters:
fit1$lambdas

#Loglikelihood and penalty value:
fit1$loglik
fit1$pen

#-----------------------------------------------------

#Above, the initial parameter estimates for the EM algorithm are computed internally in
#msvar_fit as described in the Supplement, Section A1.4.

#We can generate initial parameter estimates explicitly:

vars = msvar_init_estimates ( Y = Yt, M = 2, p = 1 )

#If one wants to use specific initial parameter estimates, one can set the
#AR coefficients ( dM x (dp + 1) matrix), transition probabilities (M x M matrix),
#and covariance matrices (dM x d matrix) as:

#vars$ar_coeff = ...
#vars$covar = ...
#vars$trans_probabs = ...

#And fit with:

fit2 = msvar_fit (Y = Yt, ar_coeff = vars$ar_coeff, covar = vars$covar, trans_probabs = vars$trans_probabs, M = 2, p = 1, coeff_penalty_str = "lasso", covar_penalty_str = "lasso" )

#-----------------------------------------------------

#Above, the tuning parameters were selected using BIC. One can use specific tuning parameters values (no selection of tuning parameters is done) as input:

lams = c(0.1, 0.1)
fit3 = msvar_fit (Y = Yt, lambdas = lams, M = 2, p = 1, coeff_penalty_str = "lasso", covar_penalty_str = "lasso" )
