
// [[Rcpp::depends(RcppGSL)]]

#include <RcppGSL.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_sort_vector.h>
#include "msvar_functions.h"

//#define PARVAR 1

#ifdef PARVAR
    #include <sys/mman.h>
    #include <sys/types.h>
    #include <sys/wait.h>
    #include <unistd.h>
#endif

#define ICRIT BIC

// [[Rcpp::export]]
Rcpp::List msvar_data_sample ( const RcppGSL::Matrix & ar_coeff, const RcppGSL::Matrix & covar, const RcppGSL::Matrix & trans_probabs, int n = 200, int cond_state = 1, int first_datapoint_t = 400, int seed = 1 )
{
    int N = n;
    if ( cond_state < 1 || first_datapoint_t < 1 || seed < 0 )
    {
        printf ( "Invalid input constants\n" );
        return Rcpp::List();
    }

    //Check matrix sizes
    if ( covar.ncol() < 2 || covar.nrow() < 2 || ar_coeff->size2 < 3 )
    {
        printf ( "Invalid input matrix dimensions\n" );
        return Rcpp::List();
    }
    int D = covar->size2;
    int M_star = covar->size1 / D;
    int p_star = ( ar_coeff->size2 - 1 ) / D;
    if ( ar_coeff->size1 != D * M_star || ar_coeff->size2 != D * p_star + 1 || trans_probabs->size1 != trans_probabs->size2 || trans_probabs->size1 != M_star )
    {
        printf ( "Invalid input matrix dimensions\n" );
        return Rcpp::List();
    }
    for ( size_t i = 0; i < M_star; ++i )
    {
        gsl_vector_const_view row = gsl_matrix_const_row ( trans_probabs, i );
        for ( size_t j = 0; j < M_star; ++j )
        {
            double pr = gsl_vector_get ( & row.vector, j );
            if (  ( pr <= 1.0e-4 ) || ( pr >= 1.0 - 1.0e-4 && M_star > 1 ) )
            {
                printf ( "Invalid probability matrix element\n" );
                return Rcpp::List();
            }
        }
        // ( Absolute ) sum of row
        if ( fabs ( gsl_blas_dasum ( & row.vector )  - 1.0 ) > 1.0e-5 )
        {
            printf ( "Invalid probability matrix row\n" );
            return Rcpp::List();
        }
    }

    gsl_rng * rng = gsl_rng_alloc ( gsl_rng_ranlxd1 ) ;
    gsl_rng_set ( rng, seed ) ;

    msvar_parameters * parameters = ( msvar_parameters * ) calloc ( 1, sizeof ( msvar_parameters )  );
    parameters->D = D;
    parameters->M = M_star;
    parameters->p = p_star;
    parameters->T = N - p_star;
    parameters->initial_state = cond_state;
    parameters->switching_coeff = 1;
    parameters->switching_covar = 1;
    msvar_parameters_alloc_matrices ( parameters );
    gsl_matrix_memcpy ( parameters->A, ar_coeff );
    gsl_matrix_memcpy ( parameters->SigmaL, covar );
    gsl_matrix_memcpy ( parameters->Pr, trans_probabs );
    for ( size_t j = 0; j < parameters->M; ++j )
    {
        gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( parameters->SigmaL, j * parameters->D, 0, parameters->D, parameters->D );
        gsl_matrix_view Sigma_inv_subview = gsl_matrix_submatrix ( parameters->Sigma_inv, j * parameters->D, 0, parameters->D, parameters->D );
        gsl_linalg_cholesky_decomp1 ( & SigmaL_subview.matrix );
        gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, & SigmaL_subview.matrix );
        gsl_linalg_cholesky_invert ( & Sigma_inv_subview.matrix );
    }

    gsl_matrix * Y = gsl_matrix_alloc ( parameters->T + parameters->p, parameters->D );    //observations are stored as row vectors
    gsl_matrix * YT = gsl_matrix_alloc ( parameters->T + parameters->p + first_datapoint_t - 1, parameters->D );
    gsl_vector * mean_vec = gsl_vector_alloc ( parameters->D );   //zero vector of size K
    gsl_vector * regimes = gsl_vector_alloc ( YT->size1 );

    generate_data ( parameters, Y, YT, mean_vec, rng, regimes );

    RcppGSL::Matrix Y_out ( Y->size1, Y->size2 );
    RcppGSL::Vector regimes_out ( Y->size1 );
    gsl_matrix_memcpy ( Y_out , Y );
    gsl_vector_const_view regimes_v = gsl_vector_const_subvector ( regimes, first_datapoint_t - 1, Y->size1 );
    gsl_blas_dcopy ( & regimes_v.vector, regimes_out );

    msvar_parameters_free_matrices ( parameters );
    free ( parameters );
    gsl_rng_free ( rng ) ;
    gsl_matrix_free ( Y );
    gsl_matrix_free ( YT );
    gsl_vector_free ( mean_vec );
    gsl_vector_free ( regimes );
    printf("Data sampled\n");

    return Rcpp::List::create ( Rcpp::Named("Y") = Y_out, Rcpp::Named("regimes") = regimes_out );
}

// [[Rcpp::export]]
Rcpp::List msvar_init_estimates ( const RcppGSL::Matrix & Y, int M = 2, int p = 1, int cond_state = 1 )
{
    if ( M < 1 || p < 1 || cond_state < 1)
    {
        printf ( "Invalid input constants\n" );
        return Rcpp::List();
    }
    msvar_parameters * estimates = ( msvar_parameters * ) calloc ( 1, sizeof ( msvar_parameters )  );
    estimates->D = Y->size2;
    estimates->M = M;
    estimates->T = Y->size1 - p;
    estimates->p = p;
    estimates->initial_state = cond_state;
    estimates->switching_coeff = 1;
    estimates->switching_covar = 1;
    msvar_parameters_alloc_matrices ( estimates );

    RcppGSL::Matrix RA ( estimates->A->size1, estimates->A->size2 );
    RcppGSL::Matrix RPr ( estimates->Pr->size1, estimates->Pr->size2 );
    RcppGSL::Matrix RS ( estimates->SigmaL->size1, estimates->SigmaL->size2 );

    gsl_matrix * X  = gsl_matrix_alloc ( estimates->T, 1 + estimates->D * estimates->p );
    gsl_vector * y = gsl_vector_alloc ( estimates->D * estimates->T );

    em_workspace * em_wspace = ( em_workspace * ) calloc ( 1, sizeof ( em_workspace )  );
    em_workspace_alloc ( em_wspace, estimates );

    double L_c = 0;
    compute_regression_matrix_vector ( NULL, Y, estimates, X, y, & L_c );

    generate_initial_transition_probabilities ( estimates, NULL );
    compute_initial_estimates ( em_wspace, Y, X, estimates, 0 );

    gsl_matrix_memcpy ( RA, estimates->A );
    gsl_matrix_memcpy ( RPr, estimates->Pr );
    //Compute every Sigma sub-matrix from their Cholesky decomposition; (upper triangle already contains Sigma)
    for ( size_t j = 0; j < estimates->M; ++j )
    {
        gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( estimates->SigmaL, j * estimates->D, 0, estimates->D, estimates->D );
        for ( size_t ir = 0; ir < estimates->D; ++ir )
        {
            double val;
            gsl_vector_view SigmaL_vecview = gsl_matrix_subrow ( & SigmaL_subview.matrix, ir, 0, ir + 1 );
            gsl_blas_ddot ( & SigmaL_vecview.vector, & SigmaL_vecview.vector, & val );
            gsl_vector_set ( & SigmaL_vecview.vector, ir, val );
            for ( size_t ic = 0; ic < ir; ++ic )
                gsl_matrix_set ( & SigmaL_subview.matrix, ir, ic, gsl_matrix_get ( & SigmaL_subview.matrix, ic, ir )  );
        }
    }
    gsl_matrix_memcpy ( RS, estimates->SigmaL );
    double loglik = estimates->loglik;
    double pen = estimates->pen;

    em_workspace_free( em_wspace, estimates->M );
    free( em_wspace);
    msvar_parameters_free_matrices ( estimates );
    free ( estimates );
    gsl_matrix_free ( X );
    gsl_vector_free ( y );
    printf("Estimates initialized\n");
    return Rcpp::List::create ( Rcpp::Named("ar_coeff") = RA, Rcpp::Named("covar") = RS, Rcpp::Named("trans_probabs") = RPr, Rcpp::Named("loglik") = loglik, Rcpp::Named("pen") = pen );
}

// [[Rcpp::export]]
Rcpp::List msvar_fit ( const RcppGSL::Matrix & Y, int M = 2, int p = 1, int cond_state = 1, RcppGSL::Matrix ar_coeff = RcppGSL::Matrix(1,1), RcppGSL::Matrix covar = RcppGSL::Matrix(1,1), RcppGSL::Matrix trans_probabs = RcppGSL::Matrix(1,1), Rcpp::NumericVector lambdas = Rcpp::NumericVector::create(0.0), const char * coeff_penalty_str = "lasso", const char * covar_penalty_str = "lasso", const char * precision_penalty_str = "none", double coeff_loglambda_step = 0.25, double covar_loglambda_step = 0.25, int max_iters = 800, int em_max_iters = 50, double tol = 1e-5, RcppGSL::Matrix Rgroups = RcppGSL::Matrix(1,1) )
{
    if ( M < 1 || p < 1 || cond_state < 1 || max_iters < 1 || em_max_iters < 1 || tol < 0.0 || coeff_loglambda_step < 0.0 || covar_loglambda_step < 0.0  )
    {
        printf ( "Invalid input constants\n" );
        return Rcpp::List();
    }
    size_t coeff_penalty, covar_penalty;
    if ( coeff_penalty_str[0] == 'l' )
        coeff_penalty = LASSO;
    else if ( coeff_penalty_str[0] == 'a' )
        coeff_penalty = ADALASSO;
    else if ( coeff_penalty_str[0] == 'g' )
    {
        coeff_penalty = GROUP_ADALASSO;
        if ( Rgroups.nrow() == 1 && Rgroups.ncol() == 1 )
        {
            printf ( "Matrix with groups must be provided as input argument\n" );
            return Rcpp::List();
        }
    }
    else if ( coeff_penalty_str[0] == 's' )
        coeff_penalty = SCAD;
    else if ( coeff_penalty_str[0] == 'm' )
        coeff_penalty = MCP;
    else
        coeff_penalty = NONE;

    if ( covar_penalty_str[0] == 'l' )
        covar_penalty = LASSO;
    else if ( covar_penalty_str[0] == 'a' )
        covar_penalty = ADALASSO;
    else if ( covar_penalty_str[0] == 's' )
        covar_penalty = SCAD;
    else if ( covar_penalty_str[0] == 'm' )
        covar_penalty = MCP;
    else
        covar_penalty = NONE;

    if ( covar_penalty == NONE )
    {
        if ( precision_penalty_str[0] == 'l' )
            covar_penalty = LASSO_INVCOVAR;
        else if ( precision_penalty_str[0] == 'a' )
            covar_penalty = ADALASSO_INVCOVAR;
        else if ( precision_penalty_str[0] == 's' )
            covar_penalty = SCAD_INVCOVAR;
        else if ( precision_penalty_str[0] == 'm' )
            covar_penalty = MCP_INVCOVAR;
        else
            covar_penalty = NONE;
    }

    msvar_parameters * estimates = ( msvar_parameters * ) calloc ( 1, sizeof ( msvar_parameters )  );
    estimates->D = Y->size2;
    estimates->M = M;
    estimates->T = Y->size1 - p;
    estimates->p = p;
    estimates->initial_state = cond_state;
    estimates->switching_coeff = 1;
    estimates->switching_covar = 1;
    msvar_parameters_alloc_matrices ( estimates );

    gsl_matrix * X = gsl_matrix_alloc ( estimates->T, 1 + estimates->D * estimates->p ); //Design matrix for generalized regression
    gsl_matrix * A_mle = gsl_matrix_alloc ( estimates->A->size1, estimates->A->size2 );
    gsl_matrix * SL_mle = gsl_matrix_alloc ( estimates->SigmaL->size1, estimates->SigmaL->size2 );
    gsl_matrix * S_inv_mle = gsl_matrix_alloc ( estimates->Sigma_inv->size1, estimates->Sigma_inv->size2 );
    gsl_matrix * Pr_mle = gsl_matrix_alloc ( estimates->Pr->size1, estimates->Pr->size2 );
    gsl_vector * y = gsl_vector_alloc ( estimates->D * estimates->T );

    em_workspace * em_wspace = ( em_workspace * ) calloc ( 1, sizeof ( em_workspace )  );
    em_workspace_alloc ( em_wspace, estimates );
    regu_workspace * regu_wspace = NULL;
    regu_cov_workspace * regu_cov_wspace = NULL;
    gsl_matrix * coeff_lambda_weights = NULL;
    gsl_matrix * covar_lambda_weights = NULL;
    gsl_matrix * groups = NULL;
    if ( coeff_penalty != NONE && coeff_penalty != ORACLE  )
    {
        size_t g_size = 0;
        if ( coeff_penalty == GROUP_ADALASSO )
        {
            groups = gsl_matrix_alloc ( Rgroups.nrow() , Rgroups.ncol() );
            gsl_matrix_memcpy ( groups, Rgroups );
            g_size = groups->size1 / estimates->D;
            coeff_lambda_weights = gsl_matrix_alloc ( estimates->M, g_size );
        }
        else
            coeff_lambda_weights = gsl_matrix_alloc ( estimates->A->size1, estimates->A->size2 - 1 );
        regu_wspace = ( regu_workspace * ) calloc ( 1, sizeof ( regu_workspace )  );
        regu_workspace_alloc ( regu_wspace, estimates, coeff_penalty, g_size );
    }
    else
    {
        coeff_penalty = NONE;
    }
    if ( covar_penalty != NONE && covar_penalty != ORACLE && covar_penalty != ORACLE_INVCOVAR && covar_penalty != GRAPHICAL_ORACLE && estimates->D > 1 )
    {
        regu_cov_wspace = ( regu_cov_workspace * ) calloc ( 1, sizeof ( regu_cov_workspace )  );
        regu_cov_workspace_alloc ( regu_cov_wspace, estimates, covar_penalty );
        covar_lambda_weights = gsl_matrix_alloc ( estimates->SigmaL->size1, estimates->SigmaL->size2 );
    }
    else
    {
        covar_penalty = NONE;
    }

    int max_i = 1;
    int max_j = 1;
    double min_loglambda_coeff = MIN_LOGLAMBDA_COEFF;
    double min_loglambda_covar = MIN_LOGLAMBDA_COVAR;
    if ( coeff_penalty == ADALASSO || coeff_penalty == GROUP_ADALASSO )
        min_loglambda_coeff = MIN_LOGLAMBDA_COEFF_ALASSO;
    if ( covar_penalty == ADALASSO || covar_penalty == ADALASSO_INVCOVAR || covar_penalty == GRAPHICAL_ADALASSO )
        min_loglambda_covar = MIN_LOGLAMBDA_COVAR_ALASSO;

    if ( coeff_penalty != NONE )
        max_i = ( int ) ( ( MAX_LOGLAMBDA - min_loglambda_coeff ) / coeff_loglambda_step );
    if ( covar_penalty != NONE  )
        max_j = ( int ) ( ( MAX_LOGLAMBDA - min_loglambda_covar ) / covar_loglambda_step );

    int i_ic = 0;
    int j_ic = 0;
    double coeff_lambda_ic = 0.0;
    double covar_lambda_ic = 0.0;

    double ** ic_array;
    ic_array = ( double ** ) calloc ( max_i, sizeof ( double * ) );
    for ( size_t i = 0; i < max_i; ++i )
    {
        ic_array[i] = ( double * ) calloc ( max_j, sizeof ( double ) );
        for ( int j = 0; j < max_j; ++j )
            ic_array[i][j] = GSL_NEGINF;
    }

    size_t pn = ICRIT;
    double L_c = 0.0;
    compute_regression_matrix_vector ( regu_wspace, Y, estimates, X, y, & L_c );
    generate_initial_transition_probabilities ( estimates, NULL );
    compute_initial_estimates ( em_wspace, Y, X, estimates, 0 );

    if ( ! ( ar_coeff.nrow() != estimates->D * estimates->M || ar_coeff.ncol() != estimates->D * estimates->p + 1 ) )
        gsl_matrix_memcpy ( estimates->A, ar_coeff );
    if ( ! ( covar.nrow() != estimates->D * estimates->M || covar.ncol() != estimates->D ) )
    {
        gsl_matrix_memcpy ( estimates->SigmaL, covar );
        for ( size_t j = 0; j < estimates->M; ++j )
        {
            gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( estimates->SigmaL, j * estimates->D, 0, estimates->D, estimates->D );
            gsl_matrix_view Sigma_inv_subview = gsl_matrix_submatrix ( estimates->Sigma_inv, j * estimates->D, 0, estimates->D, estimates->D );
            gsl_linalg_cholesky_decomp1 ( & SigmaL_subview.matrix );
            gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, & SigmaL_subview.matrix );
            gsl_linalg_cholesky_invert ( & Sigma_inv_subview.matrix );
        }
    }
    if ( ! ( trans_probabs.nrow() != estimates->M || trans_probabs.ncol() != estimates->M ) )
    {
        int normal = 1;
        for ( size_t i = 0; i < estimates->M; ++i )
        {
            gsl_vector_const_view row = gsl_matrix_const_row ( trans_probabs, i );
            for ( size_t j = 0; j < estimates->M; ++j )
            {
                double pr = gsl_vector_get ( & row.vector, j );
                if (  ( pr <= 1.0e-4 ) || ( pr >= 1.0 - 1.0e-4 && estimates->M > 1 ) )
                {
                    normal = 0;
                    i = estimates->M;
                    break;
                }
            }
            // ( Absolute ) sum of row
            if ( fabs ( gsl_blas_dasum ( & row.vector )  - 1.0 ) > 1.0e-5 )
            {
                normal = 0;
                break;
            }
        }
        if ( normal )
            gsl_matrix_memcpy ( estimates->Pr, trans_probabs );
    }
    //MLE
    expectation_maximization ( em_wspace, NULL, NULL, estimates, Y, X, y, em_max_iters, max_iters, tol, NONE, NONE, 0.0, 0.0, NULL, NULL, 0.0, NULL );

    if ( lambdas.size() == 2 && lambdas [0] > 0.0 && lambdas [1] > 0.0 )
    {
        coeff_lambda_ic = lambdas [0];
        covar_lambda_ic = lambdas [1];
    }
    else
    {
        if ( pn != LIC )
        {
            gsl_matrix_memcpy ( A_mle, estimates->A );
            gsl_matrix_memcpy ( Pr_mle, estimates->Pr );
            gsl_matrix_memcpy ( SL_mle, estimates->SigmaL );
            gsl_matrix_memcpy ( S_inv_mle, estimates->Sigma_inv );
            set_penalty_weights ( estimates->D, estimates->A, estimates->SigmaL, estimates->Sigma_inv, coeff_penalty, covar_penalty, coeff_lambda_weights, covar_lambda_weights, groups );

            printf(" Performing tuning parameter selection...\n");

            double resol_step;
            resol_step = 4.0;
            compute_information_descent_optimum ( em_wspace, regu_wspace, regu_cov_wspace, estimates, A_mle, SL_mle, S_inv_mle, Pr_mle, Y, NULL, X, y, ic_array, pn,
                    coeff_penalty, coeff_loglambda_step, covar_penalty, covar_loglambda_step, coeff_lambda_weights, covar_lambda_weights,
                    em_max_iters, max_iters, tol, & i_ic, & j_ic, 0, groups, resol_step, L_c );
            resol_step = 2.0;
            compute_information_descent_optimum ( em_wspace, regu_wspace, regu_cov_wspace, estimates, A_mle, SL_mle, S_inv_mle, Pr_mle, Y, NULL, X, y, ic_array, pn,
                    coeff_penalty, coeff_loglambda_step, covar_penalty, covar_loglambda_step, coeff_lambda_weights, covar_lambda_weights,
                    em_max_iters, max_iters, tol, & i_ic, & j_ic, SUBGRID_WIDTH, groups, resol_step, L_c );
            resol_step = 1.0;
            compute_information_descent_optimum ( em_wspace, regu_wspace, regu_cov_wspace, estimates, A_mle, SL_mle, S_inv_mle, Pr_mle, Y, NULL, X, y, ic_array, pn,
                    coeff_penalty, coeff_loglambda_step, covar_penalty, covar_loglambda_step, coeff_lambda_weights, covar_lambda_weights,
                    em_max_iters, max_iters, tol, & i_ic, & j_ic, SUBGRID_WIDTH, groups, resol_step, L_c );

            coeff_lambda_ic = ( i_ic > 0 ) * exp ( min_loglambda_coeff + ( i_ic - 1 ) * coeff_loglambda_step );
            covar_lambda_ic = ( j_ic > 0 ) * exp ( min_loglambda_covar + ( j_ic - 1 ) * covar_loglambda_step );

            gsl_matrix_memcpy ( estimates->A, A_mle );
            gsl_matrix_memcpy ( estimates->Pr, Pr_mle );
            gsl_matrix_memcpy ( estimates->SigmaL, SL_mle );
            gsl_matrix_memcpy ( estimates->Sigma_inv, S_inv_mle );
        }
    }
    set_penalty_weights ( estimates->D, estimates->A, estimates->SigmaL, estimates->Sigma_inv, coeff_penalty, covar_penalty, coeff_lambda_weights, covar_lambda_weights, groups );
    expectation_maximization ( em_wspace, regu_wspace, regu_cov_wspace, estimates, Y, X, y, em_max_iters, max_iters, tol, coeff_penalty, covar_penalty,
                                coeff_lambda_ic, covar_lambda_ic, coeff_lambda_weights, covar_lambda_weights, L_c, groups );

    RcppGSL::Vector Rlambdas(2);
    RcppGSL::Matrix Rcoeff ( estimates->A->size1, estimates->A->size2 );
    RcppGSL::Matrix Rcovar ( estimates->SigmaL->size1, estimates->SigmaL->size2 );
    RcppGSL::Matrix Rprec ( estimates->Sigma_inv->size1, estimates->Sigma_inv->size2 );
    RcppGSL::Matrix Rprobabs ( estimates->Pr->size1, estimates->Pr->size2 );

    Rlambdas[0] = coeff_lambda_ic;
    Rlambdas[1] = covar_lambda_ic;
    gsl_matrix_memcpy ( Rcoeff, estimates->A );
    gsl_matrix_memcpy ( Rprobabs, estimates->Pr );
    gsl_matrix_memcpy ( Rprec, estimates->Sigma_inv );
    //Compute every Sigma sub-matrix from their Cholesky decomposition; (upper triangle already contains Sigma)
    for ( size_t j = 0; j < estimates->M; ++j )
    {
        gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( estimates->SigmaL, j * estimates->D, 0, estimates->D, estimates->D );
        for ( size_t ir = 0; ir < estimates->D; ++ir )
        {
            double val;
            gsl_vector_view SigmaL_vecview = gsl_matrix_subrow ( & SigmaL_subview.matrix, ir, 0, ir + 1 );
            gsl_blas_ddot ( & SigmaL_vecview.vector, & SigmaL_vecview.vector, & val );
            gsl_vector_set ( & SigmaL_vecview.vector, ir, val );
            for ( size_t ic = 0; ic < ir; ++ic )
                gsl_matrix_set ( & SigmaL_subview.matrix, ir, ic, gsl_matrix_get ( & SigmaL_subview.matrix, ic, ir )  );
        }
    }
    gsl_matrix_memcpy ( Rcovar, estimates->SigmaL );

    for ( size_t i = 0; i < max_i; ++i )
        free ( ic_array[i] );
    free ( ic_array );

    em_workspace_free ( em_wspace, estimates->M );
    free ( em_wspace );
    if ( coeff_penalty != NONE )
    {
        gsl_matrix_free ( coeff_lambda_weights );
        regu_workspace_free ( regu_wspace, estimates );
        free ( regu_wspace );
        if ( coeff_penalty == GROUP_ADALASSO )
            gsl_matrix_free ( groups );
    }
    if ( covar_penalty != NONE )
    {
        gsl_matrix_free ( covar_lambda_weights );
        regu_cov_workspace_free ( regu_cov_wspace, covar_penalty, estimates );
        free (regu_cov_wspace);
    }

    double loglik = estimates->loglik;
    double pen = estimates->pen;

    msvar_parameters_free_matrices ( estimates );
    free ( estimates );
    gsl_matrix_free ( A_mle );
    gsl_matrix_free ( SL_mle );
    gsl_matrix_free ( S_inv_mle );
    gsl_matrix_free ( Pr_mle );
    gsl_matrix_free ( X );
    gsl_vector_free ( y );
    printf("MSVAR fitted\n");
    return Rcpp::List::create ( Rcpp::Named("ar_coeff") = Rcoeff, Rcpp::Named("covar") = Rcovar,  Rcpp::Named("prec") = Rprec, Rcpp::Named("trans_probabs") = Rprobabs, Rcpp::Named("lambdas") = Rlambdas, Rcpp::Named("loglik") = loglik, Rcpp::Named("pen") = pen );
}

// [[Rcpp::export]]
Rcpp::List msvar_replications ( const RcppGSL::Matrix & ar_coeff, const RcppGSL::Matrix & covar, const RcppGSL::Matrix & trans_probabs, int n = 200, int M = 0, int p = 0, int cond_state = 1, const char * coeff_penalty_str = "lasso", const char * covar_penalty_str = "lasso", const char * precision_penalty_str = "none", RcppGSL::Matrix lambdas_matrix = RcppGSL::Matrix(1,1), int replications = 50, int em_max_iters = 50, int max_iters = 800, double tol = 1.0e-5, int first_datapoint_t = 400, double coeff_loglambda_step = 0.25, double covar_loglambda_step = 0.25, int seed = 1 )
{
    int N = n;
    if ( cond_state < 1 || max_iters < 1 || em_max_iters < 1 || tol < 0.0 || coeff_loglambda_step < 0.0 || covar_loglambda_step < 0.0 || replications < 1 || first_datapoint_t < 1 || seed < 0 )
    {
        printf ( "Invalid input constants\n" );
        return Rcpp::List();
    }
    size_t coeff_penalty, covar_penalty;
    if ( coeff_penalty_str[0] == 'l' )
        coeff_penalty = LASSO;
    else if ( coeff_penalty_str[0] == 'a' )
        coeff_penalty = ADALASSO;
    else if ( coeff_penalty_str[0] == 's' )
        coeff_penalty = SCAD;
    else if ( coeff_penalty_str[0] == 'm' )
        coeff_penalty = MCP;
    else if ( coeff_penalty_str[0] == 'o' )
        coeff_penalty = ORACLE;
    else
        coeff_penalty = NONE;

    if ( covar_penalty_str[0] == 'l' )
        covar_penalty = LASSO;
    else if ( covar_penalty_str[0] == 'a' )
        covar_penalty = ADALASSO;
    else if ( covar_penalty_str[0] == 's' )
        covar_penalty = SCAD;
    else if ( covar_penalty_str[0] == 'm' )
        covar_penalty = MCP;
    else if ( covar_penalty_str[0] == 'o' )
        covar_penalty = ORACLE;
    else
        covar_penalty = NONE;

    if ( covar_penalty == NONE )
    {
        if ( precision_penalty_str[0] == 'l' )
            covar_penalty = LASSO_INVCOVAR;
        else if ( precision_penalty_str[0] == 'a' )
            covar_penalty = ADALASSO_INVCOVAR;
        else if ( precision_penalty_str[0] == 's' )
            covar_penalty = SCAD_INVCOVAR;
        else if ( precision_penalty_str[0] == 'm' )
            covar_penalty = MCP_INVCOVAR;
        else if ( precision_penalty_str[0] == 'o' )
            covar_penalty = ORACLE_INVCOVAR;
        else
            covar_penalty = NONE;
    }

    //Check matrix sizes
    if ( covar.ncol() < 2 || covar.nrow() < 2 || ar_coeff->size2 < 3 )
    {
        printf ( "Invalid input matrix dimensions\n" );
        return Rcpp::List();
    }
    int D = covar->size2;
    int M_star = covar->size1 / D;
    int p_star = ( ar_coeff->size2 - 1 ) / D;
    if ( M <= 0 )
        M = M_star;
    if ( p <= 0)
        p = p_star;
    if ( ar_coeff->size1 != D * M_star || ar_coeff->size2 != D * p_star + 1 || trans_probabs->size1 != trans_probabs->size2 || trans_probabs->size1 != M_star )
    {
        printf ( "Invalid input matrix dimensions\n" );
        return Rcpp::List();
    }
    for ( size_t i = 0; i < M_star; ++i )
    {
        gsl_vector_const_view row = gsl_matrix_const_row ( trans_probabs, i );
        for ( size_t j = 0; j < M_star; ++j )
        {
            double pr = gsl_vector_get ( & row.vector, j );
            if (  ( pr <= 1.0e-4 ) || ( pr >= 1.0 - 1.0e-4 && M_star > 1 ) )
            {
                printf ( "Invalid probability matrix element\n" );
                return Rcpp::List();
            }
        }
        // ( Absolute ) sum of row
        if ( fabs ( gsl_blas_dasum ( & row.vector )  - 1.0 ) > 1.0e-5 )
        {
            printf ( "Invalid probability matrix row\n" );
            return Rcpp::List();
        }
    }

    gsl_rng * rng = gsl_rng_alloc ( gsl_rng_ranlxd1 ) ;
    gsl_rng_set ( rng, seed ) ;

    msvar_parameters * parameters = ( msvar_parameters * ) calloc ( 1, sizeof ( msvar_parameters )  );
    msvar_parameters * estimates = ( msvar_parameters * ) calloc ( 1, sizeof ( msvar_parameters )  );
    parameters->D = D;
    parameters->M = M_star;
    parameters->T = N;
    parameters->p = p_star;
    parameters->initial_state = cond_state;
    parameters->switching_coeff = 1;
    parameters->switching_covar = 1;
    estimates->D = D;
    estimates->M = M;
    estimates->T = N;
    estimates->p = p;
    estimates->initial_state = cond_state;
    estimates->switching_coeff = 1;
    estimates->switching_covar = 1;
    msvar_parameters_alloc_matrices ( parameters );
    msvar_parameters_alloc_matrices ( estimates );
    gsl_matrix_memcpy ( parameters->A, ar_coeff );
    gsl_matrix_memcpy ( parameters->SigmaL, covar );
    gsl_matrix_memcpy ( parameters->Pr, trans_probabs );
    for ( size_t j = 0; j < parameters->M; ++j )
    {
        gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( parameters->SigmaL, j * parameters->D, 0, parameters->D, parameters->D );
        gsl_matrix_view Sigma_inv_subview = gsl_matrix_submatrix ( parameters->Sigma_inv, j * parameters->D, 0, parameters->D, parameters->D );
        gsl_linalg_cholesky_decomp1 ( & SigmaL_subview.matrix );
        gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, & SigmaL_subview.matrix );
        gsl_linalg_cholesky_invert ( & Sigma_inv_subview.matrix );
    }

    if ( coeff_penalty == NONE || coeff_loglambda_step <= 0.0 )
    {
        coeff_penalty = NONE;
        coeff_loglambda_step = 0.0;
    }
    if ( covar_penalty == NONE || covar_loglambda_step <= 0.0 )
    {
        covar_penalty = NONE;
        covar_loglambda_step = 0.0;
    }

    gsl_matrix * estimates_matrix_Pr = gsl_matrix_alloc ( replications, estimates->M * estimates->M );
    gsl_matrix * estimates_matrix_A = gsl_matrix_alloc ( replications, estimates->M * estimates->D * ( estimates->D * estimates->p + 1 ) );
    gsl_matrix * estimates_matrix_SigmaL = gsl_matrix_alloc ( replications, estimates->M * estimates->D * estimates->D );
    gsl_matrix * estimates_matrix_loglik = gsl_matrix_alloc ( replications, 1 );
    gsl_matrix * estimates_matrix_smooth_p = gsl_matrix_alloc ( replications, estimates->M );
    gsl_matrix * estimates_matrix_lambdas = gsl_matrix_alloc ( replications, 2 );

    if ( lambdas_matrix.nrow() == estimates_matrix_lambdas->size1 && lambdas_matrix.ncol() == estimates_matrix_lambdas->size2 )
        gsl_matrix_memcpy ( estimates_matrix_lambdas, lambdas_matrix );
    else
        gsl_matrix_set_zero ( estimates_matrix_lambdas );

    em_replications ( parameters, estimates, estimates_matrix_A, estimates_matrix_Pr, estimates_matrix_SigmaL, estimates_matrix_loglik, estimates_matrix_smooth_p,
                          estimates_matrix_lambdas, replications, em_max_iters, max_iters, tol, rng, first_datapoint_t, "",
                          coeff_penalty, coeff_loglambda_step, covar_penalty, covar_loglambda_step, 0, NULL );

    RcppGSL::Matrix coeff_out ( estimates_matrix_A->size1, estimates_matrix_A->size2 );
    RcppGSL::Matrix covar_out ( estimates_matrix_SigmaL->size1, estimates_matrix_SigmaL->size2 );
    RcppGSL::Matrix probabs_out ( estimates_matrix_Pr->size1, estimates_matrix_Pr->size2 );
    RcppGSL::Matrix lambdas_out ( estimates_matrix_lambdas->size1, estimates_matrix_lambdas->size2 );
    gsl_matrix_memcpy ( coeff_out , estimates_matrix_A );
    gsl_matrix_memcpy ( covar_out , estimates_matrix_SigmaL );
    gsl_matrix_memcpy ( probabs_out , estimates_matrix_Pr );
    gsl_matrix_memcpy ( lambdas_out , estimates_matrix_lambdas );

    msvar_parameters_free_matrices ( parameters );
    msvar_parameters_free_matrices ( estimates );
    free ( parameters );
    free ( estimates );

    gsl_rng_free ( rng ) ;
    gsl_matrix_free ( estimates_matrix_A );
    gsl_matrix_free ( estimates_matrix_Pr );
    gsl_matrix_free ( estimates_matrix_SigmaL );
    gsl_matrix_free ( estimates_matrix_loglik );
    gsl_matrix_free ( estimates_matrix_smooth_p );
    gsl_matrix_free ( estimates_matrix_lambdas );

    printf("Done\n");

    return Rcpp::List::create ( Rcpp::Named("ar_coeff_matrix") = coeff_out, Rcpp::Named("covar_prec_matrix") = covar_out, Rcpp::Named("trans_probabs_matrix") = probabs_out, Rcpp::Named("lambdas_matrix") = lambdas_out );
}

int gsl_matrix_scale_rows_0 ( gsl_matrix * a, const gsl_vector * x )
{
    const size_t M = a->size1;

    if (x->size != M)
    {
        GSL_ERROR ("x must match number of rows of A", GSL_EBADLEN);
    }
    else
    {
        size_t i;

        for (i = 0; i < M; ++i)
        {
            const double xi = gsl_vector_get (x, i);
            gsl_vector_view v = gsl_matrix_row (a, i);
            gsl_vector_scale (&v.vector, xi);
        }
        return GSL_SUCCESS;
    }
}

double gsl_vector_sum_0 (const gsl_vector * a)
{
    const size_t N = a->size;
    double sum = 0;
    size_t i;
    for (i = 0; i < N; i++)
    {
        sum += gsl_vector_get (a, i);
    }
    return sum;
}

//------

double scad_penalty ( const gsl_matrix * A, double lambda, double a, size_t is_covar )
{
    double s = 0;
    if ( ! is_covar )
    {
        for(size_t i = 0; i < A->size1; ++i )
            for(size_t j = 0; j < A->size2; ++j )
            {
                double v = gsl_matrix_get ( A, i, j );
                if ( fabs ( v ) <= lambda )
                    s = s + lambda * fabs( v );
                else if ( fabs ( v ) <= a * lambda )
                    s = s + ( 2 * a * lambda * fabs ( v ) - v * v - lambda * lambda ) / ( 2 * ( a - 1 ) );
                else
                    s = s + lambda * lambda * ( a + 1 ) / 2;
            }
    }
    else
    {
        for(size_t i = 0; i < A->size1; ++i )
            for(size_t j = 0; j < i; ++j )
            {
                double v = gsl_matrix_get ( A, j, i );
                if ( fabs ( v ) <= lambda )
                    s = s + 2.0 * lambda * fabs( v );
                else if ( fabs ( v ) <= a * lambda )
                    s = s + 2.0 * ( 2.0 * a * lambda * fabs ( v ) - v * v - lambda * lambda ) / ( 2.0 * ( a - 1 ) );
                else
                    s = s + 2.0 * lambda * lambda * ( a + 1 ) / 2.0;
            }
    }
    return s;
}

double mcp_penalty ( const gsl_matrix * A, double lambda, double gamma, size_t is_covar )
{
    double s = 0;
    if ( ! is_covar )
    {
        for(size_t i = 0; i < A->size1; ++i )
            for(size_t j = 0; j < A->size2; ++j )
            {
                double v = gsl_matrix_get ( A, i, j );
                if ( fabs ( v ) <= lambda * gamma )
                    s = s + lambda * fabs( v ) - 0.5 * gsl_pow_2 ( v ) / gamma;
                else
                    s = s + 0.5 * gamma * gsl_pow_2 ( lambda );
            }
    }
    else
    {
        for(size_t i = 0; i < A->size1; ++i )
            for(size_t j = 0; j < i; ++j )
            {
                double v = gsl_matrix_get ( A, j, i );
                if ( fabs ( v ) <= lambda * gamma )
                    s = s + 2.0 * ( lambda * fabs( v ) - 0.5 * gsl_pow_2 ( v ) / gamma );
                else
                    s = s + 2.0 * 0.5 * gamma * gsl_pow_2 ( lambda );
            }
    }
    return s;
}

double l1_norm ( const gsl_matrix * A, const gsl_matrix * weights, size_t is_covar )
{
    double s = 0;
    if ( ! is_covar )
    {
        for(size_t i = 0; i < A->size1; ++i )
            for(size_t j = 0; j < A->size2; ++j )
            {
                s += gsl_matrix_get ( weights, i, j ) * fabs ( gsl_matrix_get ( A, i, j ) );
            }
    }
    else
    {
        for(size_t i = 0; i < A->size1; ++i )
            for(size_t j = 0; j < i; ++j )
            {
                s += 2.0 * gsl_matrix_get ( weights, j, i ) * fabs ( gsl_matrix_get ( A, j, i ) );
            }
    }
    return s;
}

double group_l2_norm ( const gsl_matrix * A, const gsl_matrix * groups, const gsl_matrix * weights )
{
    if ( weights->size1 == 1 )
    {
        double nr = 0.0;
        for ( size_t g = 0; g < weights->size2; ++g )
        {
            //Weight for group
            double w = gsl_matrix_get ( weights, 0, g );
            //Get group
            gsl_matrix_const_view group_view = gsl_matrix_const_submatrix ( groups, g * A->size1, 0, A->size1, A->size2 );
            double s = 0.0;
            for ( size_t i = 0; i < A->size1; ++i )
            {
                for ( size_t j = 0; j < A->size2; ++j )
                {
                    if ( gsl_matrix_get ( & group_view.matrix, i, j ) > 0 )
                    {
                        double a = gsl_matrix_get ( A, i, j );
                        s += a * a;
                    }
                }
            }
            nr = nr + w * sqrt ( s );
        }
        return nr;
    }
    else
    {
        double nr = 0.0;
        for ( size_t g = 0; g < weights->size1; ++g )
        {
            //Weight for group
            double w = gsl_matrix_get ( weights, g, 0 );
            //Get group
            gsl_matrix_const_view group_view = gsl_matrix_const_submatrix ( groups, 0, g * A->size2, A->size1, A->size2 );
            double s = 0.0;
            for ( size_t i = 0; i < A->size1; ++i )
            {
                for ( size_t j = 0; j < A->size2; ++j )
                {
                    if ( gsl_matrix_get ( & group_view.matrix, i, j ) > 0 )
                    {
                        double a = gsl_matrix_get ( A, i, j );
                        s += a * a;
                    }
                }
            }
            nr = nr + w * sqrt ( s );
        }
        return nr;
    }
}

double compute_penalty ( size_t penalty, const gsl_matrix * B, double lambda, const gsl_matrix * weights, double a, double gamma, double N, const gsl_matrix * groups, size_t is_covar )
{
    if ( penalty == LASSO || penalty == ADALASSO || penalty == LASSO_INVCOVAR || penalty == ADALASSO_INVCOVAR || penalty == GRAPHICAL_LASSO || penalty == GRAPHICAL_ADALASSO )
        return N * lambda * l1_norm ( B, weights, is_covar );
    else if ( penalty == GROUP_ADALASSO )
        return N * lambda * group_l2_norm ( B, groups, weights );
    else if ( penalty == SCAD || penalty == SCAD_INVCOVAR || penalty == GRAPHICAL_SCAD )
        return N * scad_penalty ( B, lambda, a, is_covar );
    else if ( penalty == MCP || penalty == MCP_INVCOVAR || penalty == GRAPHICAL_MCP )
        return N * mcp_penalty ( B, lambda, gamma, is_covar );
    else
        return 0.0;
}

double compute_penalties ( const msvar_parameters * estimates, size_t coeff_penalty, size_t covar_penalty, double coeff_lambda, double covar_lambda, const gsl_matrix * coeff_lambda_weights, const gsl_matrix * covar_lambda_weights, double a, double gamma, double N, const gsl_matrix * groups )
{
    double cp = 0.0;
    int is_precision = ( covar_penalty == ORACLE_INVCOVAR || covar_penalty == GRAPHICAL_ORACLE || INVCOVAR_REG ( covar_penalty ) );
    for ( size_t m = 0; m < estimates->M; ++m )
    {
        if ( coeff_penalty != NONE )
        {
            size_t r, s;
            if ( coeff_penalty != GROUP_ADALASSO )
            {
                r = m * estimates->D;
                s = estimates->D;
            }
            else
            {
                r = m;
                s = 1;
            }
            gsl_matrix_const_view Av = gsl_matrix_const_submatrix ( estimates->A, m * estimates->D, 1, estimates->D, estimates->D * estimates->p );
            gsl_matrix_const_view coeff_weights_v = gsl_matrix_const_submatrix ( coeff_lambda_weights, r, 0, s, coeff_lambda_weights->size2 );
            cp += compute_penalty ( coeff_penalty, & Av.matrix, coeff_lambda, & coeff_weights_v.matrix, a, gamma, N, groups, 0 );
        }
        if ( covar_penalty != NONE )
        {
            const gsl_matrix * S;
            if ( is_precision )
                S = estimates->Sigma_inv;
            else
                S = estimates->SigmaL;
            gsl_matrix_const_view Sigmav = gsl_matrix_const_submatrix ( S, m * estimates->D, 0, estimates->D, estimates->D );
            gsl_matrix_const_view covar_weights_v = gsl_matrix_const_submatrix ( covar_lambda_weights, m * estimates->D, 0, estimates->D, estimates->D );
            cp += compute_penalty ( covar_penalty, & Sigmav.matrix, covar_lambda, & covar_weights_v.matrix, a, gamma, N, NULL, 1 );
        }
    }
    return cp;
}

double inner_product ( const gsl_matrix * A, const gsl_matrix * B )
{
    double r;
    double p = 0.0;
    for(size_t j = 0; j < B->size2; ++j )
    {
        gsl_vector_const_view A_vecview = gsl_matrix_const_column( A, j );
        gsl_vector_const_view B_vecview = gsl_matrix_const_column( B, j );
        gsl_blas_ddot ( & A_vecview.vector, & B_vecview.vector, & r);
        p = p + r;
    }
    return p;
}

double frobenius_norm_diff ( const gsl_matrix * a, const gsl_matrix * b )
{
    double sum = 0;
    double ac, bc;
    for ( size_t i = 0; i < a->size1; ++i )
        for ( size_t j = 0; j < a->size2; ++j )
        {
            ac = gsl_matrix_get ( a,i,j );
            bc = gsl_matrix_get ( b,i,j );
            sum += ( ac - bc ) * ( ac - bc );
        }
    return sqrt ( sum );
}

double norm_diff ( const gsl_vector * a, const gsl_vector * b )
{
    double sum = 0;
    double ac, bc;
    for ( size_t i = 0; i < a->size; ++i )
    {
        ac = gsl_vector_get ( a,i );
        bc = gsl_vector_get ( b,i );
        sum += ( ac - bc ) * ( ac - bc );
        }
    return sqrt ( sum );
}

int kronecker_product ( CBLAS_TRANSPOSE_t TransA, double alpha, const gsl_matrix * A, const gsl_matrix * B, gsl_matrix * C )
{
    size_t p = B->size1;
    size_t q = B->size2;
    size_t n, m;
    //Check to use either A or its transpose
    if ( TransA == CblasNoTrans )
    {
        
        n = p * A->size1;
        m = q * A->size2;
    }
    else
    {
        n = p * A->size2;
        m = q * A->size1;
    }
    if ( C->size1 == n && C->size2 == m )
    {
        for ( int i = 0; i < n; ++i )
        for ( int j = 0; j < m; ++j )
        {
            double Aij;
            if ( TransA == CblasNoTrans )
                Aij = gsl_matrix_get ( A, floor ( i / p ) , floor ( j / q ) );
            else
                Aij = gsl_matrix_get ( A, floor ( j / q ) , floor ( i / p )  );
            double Bij = gsl_matrix_get ( B, i % p, j % q );
            double Cij = gsl_matrix_get ( C, i, j );
            //Compute the Kronecker producto C = A x B + alpha * C
            gsl_matrix_set ( C, i, j, Aij * Bij + Cij * alpha );
        }
        return GSL_SUCCESS;
    }
    else
        GSL_ERROR ( " Kronecker product invalid length ", GSL_EBADLEN );
}

int msvar_parameters_read_constants ( const char * filename, msvar_parameters * parameters )
{
    FILE * fin = fopen ( filename, "r" );
    if ( fin == NULL )
    {   
        printf ( " File to read %s:\n", filename );
        GSL_ERROR ( " fopen (r) failed ", GSL_EINVAL );
    }
    else
    {
        if ( ! (
            fscanf ( fin, "\nmsvar_D = %lu",    & ( parameters->D )  ) &&
            fscanf ( fin, "\nmsvar_M = %lu",    & ( parameters->M )  ) &&
            fscanf ( fin, "\nmsvar_N = %lu",    & ( parameters->T )  ) &&
            fscanf ( fin, "\nmsvar_p = %lu",    & ( parameters->p )  ) &&
            fscanf ( fin, "\nmsvar_state = %lu",    & ( parameters->initial_state )  ) &&
            fscanf ( fin, "\nmsvar_switch_coeff = %lu",    & ( parameters->switching_coeff )  ) &&
            fscanf ( fin, "\nmsvar_switch_covar = %lu",    & ( parameters->switching_covar )  )

        ) )
        {
            fclose ( fin ) ;
            printf ( " File %s:\n", filename ) ;
            GSL_ERROR ( " fscanf failed ", GSL_EINVAL );
        }

        fclose ( fin );
        int N = parameters->T - parameters->p;
        parameters->T = parameters->T - parameters->p;
        if ( parameters->M == 1 && ( parameters->switching_coeff || parameters->switching_covar )  )
        {
            parameters->switching_coeff = 0;
            parameters->switching_covar = 0;
            ///GSL_ERROR ( " Input number of states is 1; unable to use switching parameters ", GSL_EINVAL );
        }
        //Check values are adequate
        if ( parameters->D <= 0 || parameters->M <= 0 || N <= 0 || parameters->p <= 0 )
        {
            GSL_ERROR ( " Negative input parameter constant ", GSL_EINVAL );
        }
        else if ( parameters->initial_state <= 0 || parameters->initial_state > parameters->M )
        {
            GSL_ERROR ( " Markov initial state out of range ", GSL_EINVAL );
        }
        else if ( parameters->M > 1 &&  ( ! parameters->switching_coeff ) && ( ! parameters->switching_covar )  )
        {
            GSL_ERROR ( " No switching of parameters input; number of states must be 1 ", GSL_EINVAL );
        }
        else
        {
            return GSL_SUCCESS;
        }
    }
}

int msvar_parameters_read_matrices ( const char * filename, msvar_parameters * parameters )
{
    // Read matrices from corresponding files
    char subfile[ strlen ( filename ) + 64 ];
    strcpy ( subfile, filename );
    strcat ( subfile,"_A" );
    fscanf_matrix ( subfile, "A", parameters->A );
    strcpy ( subfile, filename );
    strcat ( subfile,"_Pr" );
    fscanf_matrix ( subfile, "Pr", parameters->Pr );
    
    //Check that the matrix is adequate ( row sums to 1 )
    if ( parameters->M == 1 )
        gsl_matrix_set_all ( parameters->Pr, 1.0 );
    else
    {
        for ( size_t i = 0; i < parameters->M; ++i )
        {
            gsl_vector_const_view row = gsl_matrix_const_row ( parameters->Pr, i );
            for ( size_t j = 0; j < parameters->M; ++j )
            {
                double pr = gsl_vector_get ( & row.vector, j );
                if (  ( pr <= 0.0 ) || ( pr == 1.0 && parameters->M > 1 ) )
                {
                    GSL_ERROR ( " Invalid probability matrix element ", GSL_EINVAL );
                }
            }
            // ( Absolute ) sum of row
            if ( fabs ( gsl_blas_dasum ( & row.vector )  - 1.0 ) > 1.0e-5 )
            {
                printf ( " %lu-th row sum: %e\n", i+1, gsl_blas_dasum ( & row.vector )  );
                GSL_ERROR ( " Invalid probability matrix row ", GSL_EINVAL );
            }
        }
    }
    
    strcpy ( subfile, filename );
    strcat ( subfile,"_Sigma" );
    fscanf_matrix ( subfile, "Sigma", parameters->SigmaL );
    //Compute Cholesky decomposition of every Sigma sub-matrix
    //An error occurs if the matrix is not positive definite
    size_t Nj = parameters->M;
    if ( parameters->switching_covar == 0 )
        Nj = 1;
    for ( size_t j = 0; j < Nj; ++j )
    {
        gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( parameters->SigmaL, j * parameters->D, 0, parameters->D, parameters->D );
        //Saves Cholesky factor ( in the diagonal and lower triangular part ) and the original matrix ( in upper triangular part ) of the submatrix of parameters->SigmaL
        gsl_linalg_cholesky_decomp1 ( & SigmaL_subview.matrix );
    }
    strcpy ( subfile, filename );
    strcat ( subfile,"_Sigma_inv" );
    fscanf_matrix ( subfile, "Sigma inv", parameters->Sigma_inv );
    return GSL_SUCCESS;

}

int msvar_parameters_save ( const char * filename, const msvar_parameters * parameters )
{
    FILE * f = fopen ( filename, "w" );
    if ( f == NULL )
    {
        printf ( " File to write %s:\n", filename );
        GSL_ERROR ( " fopen (w) failed ", GSL_EINVAL );
    }
    else
    {
        //Write constants
        fprintf ( f, "\nmsvar_D = %lu",   ( parameters->D )  );
        fprintf ( f, "\nmsvar_M = %lu",   ( parameters->M )  );
        fprintf ( f, "\nmsvar_N = %lu",   ( parameters->T + parameters->p )  );
        fprintf ( f, "\nmsvar_p = %lu",   ( parameters->p )  );
        fprintf ( f, "\nmsvar_state = %lu",   ( parameters->initial_state )  );
        fprintf ( f, "\nmsvar_switch_coeff = %lu",    ( parameters->switching_coeff )  );
        fprintf ( f, "\nmsvar_switch_covar = %lu\n",    ( parameters->switching_covar )  );
        fclose ( f );
        
        //Write matrix of coefficients
        char file_out[strlen ( filename ) + 64];
        strcpy ( file_out, filename );
        strcat ( file_out, "_A" );
        fprintf_matrix ( file_out, "A", parameters->A );
        
        //Write matrix of probabilities and filename
        strcpy ( file_out, filename );
        strcat ( file_out, "_Pr" );
        fprintf_matrix ( file_out, "Pr", parameters->Pr );
        
        //Write covariance matrix and filename
        size_t Nj = 1;
        if ( parameters->switching_covar )
            Nj = parameters->M;
        //Compute every Sigma sub-matrix from their Cholesky decomposition; (upper triangle already contains Sigma)
        for ( size_t j = 0; j < Nj; ++j )
        {
            gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( parameters->SigmaL, j * parameters->D, 0, parameters->D, parameters->D );
            for ( size_t ir = 0; ir < parameters->D; ++ir )
            {
                double val;
                gsl_vector_view SigmaL_vecview = gsl_matrix_subrow ( & SigmaL_subview.matrix, ir, 0, ir + 1 );
                gsl_blas_ddot ( & SigmaL_vecview.vector, & SigmaL_vecview.vector, & val );
                gsl_matrix_set ( & SigmaL_subview.matrix, ir, ir, val );
                for ( size_t ic = 0; ic < ir; ++ic )
                {
                    double s = gsl_matrix_get ( & SigmaL_subview.matrix, ic, ir );
                    gsl_matrix_set ( & SigmaL_subview.matrix, ir, ic, s );
                }
            }
        }
        strcpy ( file_out, filename );
        strcat ( file_out, "_Sigma" );
        fprintf_matrix ( file_out, "Sigma", parameters->SigmaL );
        strcat ( file_out, "_inv" );
        fprintf_matrix ( file_out, "Sigma inv", parameters->Sigma_inv );
        
        return GSL_SUCCESS;
    }
}

int generate_data ( const msvar_parameters * parameters, gsl_matrix * Y, gsl_matrix * YT, gsl_vector * mean_vec, const gsl_rng * rng, gsl_vector * regimes )
{
    size_t m = 0;
    for ( size_t t = 0; t < YT->size1; ++t )
    {
        //Sample Markov state
        if ( parameters->M > 1 )
        {
            if ( t == 0 )
            {
                //Use input parameter as initial state
                m = parameters->initial_state - 1;
            }
            else
            {
                //Condition on current state m ( the one from previous iteration )
                gsl_vector_const_view Pr_vecview = gsl_matrix_const_row ( parameters->Pr, m );
                unsigned int n[parameters->M];
                gsl_ran_multinomial ( rng, parameters->M, 1, ( & Pr_vecview.vector ) ->data, n );  //computes a random sample n from the distribution ( & Pr_vecview.vector ) ->data
                //Get sampled state ( index where n is not zero )
                for ( m = 0; ; ++m )
                {
                    if ( n[m] > 0 )
                        break;
                }
            }
        }
        
        //Sample y_t using state m
        gsl_vector_view YT_vecview = gsl_matrix_row ( YT, t );//get the pointer to t_th row
        //Get corresponding covariance matrix
        gsl_matrix_const_view SigmaL_subview = gsl_matrix_const_submatrix ( parameters->SigmaL, parameters->switching_covar * m * parameters->D, 0, parameters->D, parameters->D );
        compute_ar_sum ( m, t, parameters, NULL, YT, mean_vec );
        gsl_ran_multivariate_gaussian ( rng, mean_vec, & SigmaL_subview.matrix, & YT_vecview.vector );   //store the K dimensional Gaussian random vector value in Y_vecview
        if ( regimes != NULL )
            gsl_vector_set ( regimes, t, m + 1 );
    }
    gsl_matrix_const_view YT_subview = gsl_matrix_const_submatrix ( YT, YT->size1 - Y->size1, 0, Y->size1, parameters->D );
    gsl_matrix_memcpy ( Y, & YT_subview.matrix );
    
    return GSL_SUCCESS;
}

int generate_initial_transition_probabilities ( msvar_parameters * estimates, const gsl_rng * rng )
{
    if ( estimates->M == 1 )
    {
        gsl_matrix_set_all ( estimates->Pr, 1.0 );
        return GSL_SUCCESS;
    }
    
    //double pvar = 0.01;
    //double p = fabs ( 1.0 / estimates->M + gsl_ran_gaussian ( rng, pvar )  );
    
    //Symmetric matrix with slightly largest elements on diagonal
    double r = 2.0 / 100.0;
    double p = 1.0 / estimates->M - ( r / estimates->M ) / ( estimates->M - 1 );
    double d = round ( p * 1000 ) / 1000;
    double s = d * ( estimates->M - 1 );
    gsl_matrix_set_all ( estimates->Pr, d );
    for ( size_t i_m = 0; i_m < estimates->M; ++i_m )
        gsl_matrix_set ( estimates->Pr, i_m, i_m, 1.0 - s );
    
    return GSL_SUCCESS;
}

int forward_recursion ( em_workspace * em_wspace, const gsl_matrix * X, const gsl_matrix * Y, msvar_parameters * estimates )
{
    gsl_vector_view xi_filter_vecview = gsl_matrix_column ( em_wspace->xi_filter, 0 );
    gsl_vector_view xi_tt1_vecview    = gsl_matrix_column ( em_wspace->xi_tt1, 0 );  //xi_tt1 is xi_{t | t-1}; here xi{1|0} already contains initial state
    gsl_blas_dcopy ( & xi_tt1_vecview.vector, & xi_filter_vecview.vector );
    estimates->loglik = 0;
    for ( size_t t = 0; t < estimates->T; ++t )
    {
        gsl_vector_const_view Y_vecview = gsl_matrix_const_row ( Y, t + estimates->p );//get row ( t+p ) from matrix Y
        //create vector eta with the Gaussian density function computed at AR ( p )
        for ( size_t m = 0; m < estimates->M; ++m )
        {
            double log_pdf;
            compute_ar_sum ( m, t, estimates, X, NULL, em_wspace->mean_vec );
            // Every Sigma sub-matrix already has its Cholesky factor stored
            gsl_matrix_const_view SigmaL_subview = gsl_matrix_const_submatrix ( estimates->SigmaL, estimates->switching_covar * m * estimates->D, 0, estimates->D, estimates->D );
            //Compute log density
            gsl_ran_multivariate_gaussian_log_pdf ( & Y_vecview.vector, em_wspace->mean_vec, & SigmaL_subview.matrix, & log_pdf, em_wspace->temp_vec );
            gsl_vector_set ( em_wspace->temp_vecm, m, GSL_MAX ( exp ( log_pdf ), 1.0e-300 )  );
        }
        //Get column t from matrix xi_tt1 matrix
        xi_tt1_vecview = gsl_matrix_column ( em_wspace->xi_tt1, t );
        //We compute using xi_tt1{j} = F * xi_filter{j-1}, F^t = Pr matrix
        gsl_blas_dgemv ( CblasTrans, 1.0, estimates->Pr, & xi_filter_vecview.vector, 0.0, & xi_tt1_vecview.vector );
        xi_filter_vecview = gsl_matrix_column ( em_wspace->xi_filter, t );
        gsl_blas_dcopy ( & xi_tt1_vecview.vector, & xi_filter_vecview.vector );
        double val;
        gsl_blas_ddot ( em_wspace->temp_vecm, & xi_filter_vecview.vector, & val );
        gsl_vector_mul ( & xi_filter_vecview.vector, em_wspace->temp_vecm );
        if ( val == 0.0 )
        {
            printf ( " Division by zero in forward recursion\n");
            estimates->loglik = GSL_NEGINF;
            break;
        }
        gsl_blas_dscal ( 1.0 / val, & xi_filter_vecview.vector );
        //Increment the likelihood for datum t
        estimates->loglik = estimates->loglik + log ( val );
    }
    return GSL_SUCCESS;
}

int backward_recursion ( em_workspace * em_wspace, const gsl_matrix * Ft )
{
    size_t T = em_wspace->xi_smooth->size2;
    gsl_vector_view xi_smooth_vecview;
    for ( size_t j = 0; j < T; ++j )
    {
        gsl_vector_const_view xi_filter_vecview = gsl_matrix_const_column ( em_wspace->xi_filter, T-1-j );//index j in xi_filter vector
        if ( j == 0 )
        {
            // We set first iterate ( last element ) as xi_{T|T} ( filtered )
            xi_smooth_vecview = gsl_matrix_column ( em_wspace->xi_smooth, T-1 );
            gsl_blas_dcopy ( & xi_filter_vecview.vector, & xi_smooth_vecview.vector );
        }
        else
        {
            //Compute the smooth xi
            gsl_blas_dcopy ( & xi_smooth_vecview.vector, em_wspace->temp_vecm );//copy the smoothed value for previous index into temp_vecm
            gsl_vector_const_view xi_tt1_vecview = gsl_matrix_const_column ( em_wspace->xi_tt1, T-j );//index T-j in xi_tt1 vector, for j > 0
            gsl_vector_div ( em_wspace->temp_vecm, & xi_tt1_vecview.vector );   //Elementwise division
            xi_smooth_vecview = gsl_matrix_column ( em_wspace->xi_smooth, T-1-j );//returns a vector view of index j in smooth vector
            gsl_blas_dgemv ( CblasNoTrans, 1.0, Ft, em_wspace->temp_vecm, 0.0, & xi_smooth_vecview.vector );//Compute F' * temp_vecm and store in xi_smooth_vecview
            gsl_vector_mul ( & xi_smooth_vecview.vector, & xi_filter_vecview.vector ); //Elementwise product
        }
    }
    return GSL_SUCCESS;
}

int estimate_transition_probabilities ( em_workspace * em_wspace, msvar_parameters * estimates, double C )
{
    //Using ( 5.12 ) , page 83
    
    gsl_matrix_memcpy ( em_wspace->tmpm, em_wspace->xi_smooth );
    gsl_matrix_div_elements ( em_wspace->tmpm, em_wspace->xi_tt1 ); // xi_smooth / xi_tt1
    for ( size_t i = 0; i < estimates->M; ++i )
    {
        gsl_vector_set_zero ( em_wspace->temp_vecx );
        //Row i of transition probabilities
        gsl_vector_view Pr_vecview = gsl_matrix_row ( estimates->Pr, i );
        //Sum over all columns ( t's )
        for ( size_t t = 0; t < estimates->T; ++t )
        {
            //Get column t of xi_smooth / xi_tt1
            gsl_vector_const_view tmpm_vecview = gsl_matrix_const_column ( em_wspace->tmpm, t );
            //Multiply by Pr_vecview and save into temp_vecm:
            gsl_blas_dcopy ( & tmpm_vecview.vector, em_wspace->temp_vecm );
            gsl_vector_mul ( em_wspace->temp_vecm, & Pr_vecview.vector );
            // Finally, temp_vecm = temp_vecm + xi_filter ( i,t-1 ) * Pr_vecview * tmpm_vecview:
            double alpha;
            if ( t == 0 )
                //xi_{t|t} for time t=0 ( not index ) is column of identity matrix with entry 1 for initial state
                alpha = (  ( i == ( estimates->initial_state - 1 )  ) ? ( 1.0 ) : ( 0.0 )  );
            else
                alpha = gsl_matrix_get ( em_wspace->xi_filter, i, t - 1 );
            // temp_vecx = temp_vecx + xi_filter ( i,t-1 ) * Pr_vecview * tmpm_vecview:
            gsl_vector_axpby ( alpha, em_wspace->temp_vecm, 1.0, em_wspace->temp_vecx );
        }
        //Normalize temp_vecx
        double s = gsl_blas_dasum ( em_wspace->temp_vecx );
        //Penalization
        gsl_vector_add_constant ( em_wspace->temp_vecx, C );
        gsl_blas_dscal ( 1.0 / ( s + C * estimates->M ), em_wspace->temp_vecx );
        //Update row i of Pr matrix
        gsl_blas_dcopy ( em_wspace->temp_vecx, & Pr_vecview.vector );
    }
    
    return GSL_SUCCESS;
}

int estimate_coefficients ( em_workspace * em_wspace, const gsl_matrix * Y, const gsl_matrix * X, const gsl_vector * y, msvar_parameters * estimates )
{
    //Compute the regression coefficients as generalized linear regression ( pag. 189 )
    //Since estimates->switching_coeff = 0, here estimates->M only depends on SigmaL
    if ( estimates->M > 1 )
    {
        
        gsl_matrix_set_zero ( em_wspace->C1 );
        gsl_matrix_set_zero ( em_wspace->C2 );
        for ( size_t i = 0; i < estimates->M; ++i )
        {
            gsl_matrix_memcpy ( em_wspace->tmp, X );
            //Product X' times E_m ( E_m is diag ( xi_smooth )  )
            gsl_vector_const_view xi_vecview = gsl_matrix_const_row ( em_wspace->xi_smooth, i );
            gsl_matrix_scale_rows_0 ( em_wspace->tmp, & xi_vecview.vector );
            
            gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( estimates->SigmaL, i * estimates->D, 0, estimates->D, estimates->D );
            // We already have the Cholesky decomposition of Sigma_subview, so we can invert ( no need to restore, since in next estimation step it is modified )
            gsl_linalg_cholesky_invert ( & SigmaL_subview.matrix );
            //Kronecker with Sigma_m^{-1}
            kronecker_product ( CblasTrans, 1.0, em_wspace->tmp, & SigmaL_subview.matrix, em_wspace->C2 );
            //Add factor X to the product X' times E_m
            gsl_blas_dgemm ( CblasTrans, CblasNoTrans, 1.0, em_wspace->tmp, X, 0.0, em_wspace->Ctmp );
            //Kronecker with Sigma_m^{-1}
            kronecker_product ( CblasNoTrans, 1.0, em_wspace->Ctmp, & SigmaL_subview.matrix, em_wspace->C1 );
        }
        //Rightmost product in b = inv ( C1 ) C2 y
        gsl_blas_dgemv ( CblasNoTrans, 1.0, em_wspace->C2, y, 0.0, em_wspace->b );
        //Solve C1 * x = b, for x
        //Use HouseHolder, inline ( other options include LU, QR )
        gsl_linalg_HH_svx ( em_wspace->C1, em_wspace->b ); // ( C1 is modified by this )
        //Update coefficient matrix A
        for ( size_t i = 0; i < estimates->A->size2; ++i )
        {
            // b is column-stacked Aj's
            gsl_vector_view b_subview = gsl_vector_subvector ( em_wspace->b, i * estimates->D, estimates->D );
            gsl_matrix_set_col ( estimates->A, i, & b_subview.vector );
        }
    }
    else
    {
        //Only one regime; by properties of Kronecker product the system becomes general linear regression ( Sigma inverse disappears )
        estimate_switching_coefficients ( em_wspace, Y, X, estimates );
    }
    return GSL_SUCCESS;
}

int estimate_switching_coefficients ( em_workspace * em_wspace, const gsl_matrix * Y, const gsl_matrix * X, msvar_parameters * estimates )
{
    //Compute the regression coefficients as generalized linear regression ( pag. 188 )
    
    gsl_matrix_const_view Y_subview = gsl_matrix_const_submatrix ( Y, estimates->p, 0, estimates->T, estimates->D );
    
    for ( size_t i = 0; i < estimates->M; ++i )
    {
        gsl_matrix_memcpy ( em_wspace->tmp, X );
        //Product E_m times X ( E_m is diag ( xi_smooth ) )
        gsl_vector_const_view xi_vecview = gsl_matrix_const_row ( em_wspace->xi_smooth, i );
        gsl_matrix_scale_rows_0 ( em_wspace->tmp, & xi_vecview.vector );
        //Get X' Em X
        gsl_blas_dgemm ( CblasTrans, CblasNoTrans, 1.0, em_wspace->tmp, X, 0.0, em_wspace->Ctmp );
        //Compute  X' Em Y = tmp' Y
        gsl_blas_dgemm ( CblasTrans, CblasNoTrans, 1.0, em_wspace->tmp, & Y_subview.matrix, 0.0, em_wspace->Ctmp2 );
        //Solve Ctmp B = Ctmp2
        for ( size_t j = 0; j < estimates->D; ++j )
        {
            //Use HouseHolder for each column ( other options include LU, QR )
            gsl_matrix_memcpy ( em_wspace->Ctmp1HH, em_wspace->Ctmp ); // ( Matrix gets destroyed by HH )
            gsl_vector_const_view Ctmp2_vecview = gsl_matrix_const_column ( em_wspace->Ctmp2, j );
            gsl_vector_view A_vecview = gsl_matrix_row ( estimates->A, i * estimates->D + j );
            gsl_linalg_HH_solve ( em_wspace->Ctmp1HH, & Ctmp2_vecview.vector, & A_vecview.vector );
        }
    }
    return GSL_SUCCESS;
}

int estimate_sparse_coefficients ( em_workspace * em_wspace, regu_workspace * regu_wspace,
                                   const gsl_matrix * Y, const gsl_matrix * X, msvar_parameters * estimates,
                                   size_t penalty, double lambda, gsl_matrix * lambda_weights,
                                 size_t max_iters, double tol, double L_c, const gsl_matrix * groups )
{
    if ( penalty == ORACLE )
    {
        estimate_switching_coefficients ( em_wspace, Y, X, estimates );
        gsl_matrix_const_view Y_view = gsl_matrix_const_submatrix ( Y, estimates->p, 0, estimates->T, estimates->D );
        gsl_matrix_memcpy ( regu_wspace->Ys, & Y_view.matrix );
        gsl_matrix_view A0_subview = gsl_matrix_submatrix ( estimates->A, 0, 1, estimates->A->size1, estimates->D * estimates->p );
        gsl_matrix_mul_elements ( & A0_subview.matrix, lambda_weights );
        //Intercept
        //Get Y - XB^T
        gsl_matrix_memcpy ( em_wspace->tmp, X );
        for ( int i = 0; i < estimates->M; ++i )
        {
            gsl_vector_view intercept_view = gsl_matrix_subcolumn ( estimates->A, 0, estimates->D * i, estimates->D );
            gsl_matrix_view X_tmp_view = gsl_matrix_submatrix ( em_wspace->tmp, 0, 1, estimates->T, estimates->D * estimates->p );
            gsl_matrix_view A_subview = gsl_matrix_submatrix ( estimates->A, estimates->D * i, 1, estimates->D, estimates->D * estimates->p );
            gsl_matrix_memcpy ( regu_wspace->Ys, & Y_view.matrix );
            gsl_blas_dgemm ( CblasNoTrans, CblasTrans, -1.0, & X_tmp_view.matrix, & A_subview.matrix, 1.0, regu_wspace->Ys );
            gsl_vector_const_view xi_view = gsl_matrix_const_row ( em_wspace->xi_smooth, i );
            double dd = gsl_vector_sum_0 ( & xi_view.vector );
            for ( size_t j = 0; j < estimates->T; ++j )
            {
                double x = gsl_vector_get ( & xi_view.vector, j );
                gsl_vector_view y_view = gsl_matrix_row ( regu_wspace->Ys, j );
                gsl_blas_dscal ( x, & y_view.vector );
            }
            for ( size_t j = 0; j < estimates->D; ++j )
            {
                gsl_vector_const_view y_view = gsl_matrix_const_column ( regu_wspace->Ys, j );
                double dn = gsl_vector_sum_0 ( & y_view.vector );
                gsl_vector_set ( & intercept_view.vector, j, dn / dd );
            }
        }
        return GSL_SUCCESS;
    }

    gsl_matrix_const_view Y_view = gsl_matrix_const_submatrix ( Y, estimates->p, 0, estimates->T, estimates->D );
    
    size_t num_p = 1;
    size_t pn = 0;
#ifdef PARVAR
    num_p = estimates->M;
    pid_t pid_arr [ num_p ];
    pid_t ppid = getpid ();
    for ( size_t k = 0; k < num_p - 1; ++k )
    {
        pid_t p = fork();
        if ( p > 0 ) //caller
        {
            pid_arr [ pn ] = p;
            ++pn;
            continue;
        }
        else if ( ! p ) // child
        {
            break;
        }
        else // error
        {
            perror ( "\n fork call failed " );
            exit ( 1 );
        }
    }
    if ( getpid () == ppid )
        pn = 0;
    else
        ++pn;
#endif

    for ( int i = 0; i < estimates->M; ++i )
    {
        //Each thread does a regime
        if ( num_p > 1 && pn != i )
            continue;
        
        gsl_matrix_memcpy ( regu_wspace->Ys, & Y_view.matrix );
        gsl_matrix_memcpy ( em_wspace->tmp, X );
        gsl_matrix_view X_tmp_view = gsl_matrix_submatrix ( em_wspace->tmp, 0, 1, estimates->T, estimates->D * estimates->p );
        gsl_matrix_view A_subview = gsl_matrix_submatrix ( estimates->A, estimates->D * i, 1, estimates->D, estimates->D * estimates->p );
        gsl_vector_view intercept_view = gsl_matrix_subcolumn ( estimates->A, 0, estimates->D * i, estimates->D );

        //Product E_m times X, E_m is diag ( sqrt xi_smooth )
        for ( size_t j = 0; j < estimates->T; ++j )
        {
            double xm;
            xm = gsl_matrix_get ( em_wspace->xi_smooth, i, j );
            //Design matrix
            gsl_vector_view x_vecview = gsl_matrix_row ( & X_tmp_view.matrix, j );
            gsl_blas_dscal ( sqrt ( xm ), & x_vecview.vector );
            //Response matrix
            gsl_vector_view y_vecview = gsl_matrix_row ( regu_wspace->Ys, j );
            gsl_blas_daxpy ( -1.0, & intercept_view.vector, & y_vecview.vector );
            gsl_blas_dscal ( sqrt ( xm ), & y_vecview.vector );
        }
        
        gsl_matrix_const_view SigmaL_subview = gsl_matrix_const_submatrix ( estimates->SigmaL, estimates->D * i, 0, estimates->D, estimates->D );
        gsl_matrix_memcpy ( regu_wspace->Li, & SigmaL_subview.matrix );
        gsl_linalg_cholesky_invert ( regu_wspace->Li );
        //Weights for adalasso
        if ( penalty == GROUP_ADALASSO )
        {
            gsl_matrix_const_view Lm_subview = gsl_matrix_const_submatrix ( lambda_weights, i, 0, 1, lambda_weights->size2 );
            gsl_matrix_transpose_memcpy ( regu_wspace->lambda_weights_tr, & Lm_subview.matrix );
        }
        else// if ( penalty == ADALASSO )
        {
            gsl_matrix_const_view Lm_subview = gsl_matrix_const_submatrix ( lambda_weights, estimates->D * i, 0, estimates->D, estimates->D * estimates->p );
            gsl_matrix_transpose_memcpy ( regu_wspace->lambda_weights_tr, & Lm_subview.matrix );
        }

        //Block coordinate descent iterations
        int its;
        for ( its = 0; its < 1; ++its )
        {
            //Do for each coordinate
            for ( size_t j = 0; j < estimates->D; ++j )
            {
                gsl_matrix_set_zero ( regu_wspace->R );
                double s;
                for ( size_t k = 0; k < estimates->D; ++k )
                {
                    s = gsl_matrix_get ( regu_wspace->Li, j, k );
                    if ( k == j || s < tol )
                        continue;
                    gsl_matrix_const_view Yc_view = gsl_matrix_const_submatrix ( regu_wspace->Ys, 0, k, regu_wspace->Ys->size1, 1 );
                    gsl_matrix_scale ( regu_wspace->R, 1.0 / s );
                    gsl_matrix_add ( regu_wspace->R, & Yc_view.matrix );
                    gsl_matrix_const_view A_rview = gsl_matrix_const_submatrix ( & A_subview.matrix, k, 0, 1, ( & A_subview.matrix )->size2 );
                    gsl_blas_dgemm ( CblasNoTrans, CblasTrans, -1.0 * s, & X_tmp_view.matrix, & A_rview.matrix, s, regu_wspace->R );
                }
                s = gsl_matrix_get ( regu_wspace->Li, j, j );
                gsl_matrix_scale ( regu_wspace->R, 1.0 / ( 2.0 * s ) );
                gsl_matrix_const_view Yc_view = gsl_matrix_const_submatrix ( regu_wspace->Ys, 0, j, regu_wspace->Ys->size1, 1 );
                gsl_matrix_add ( regu_wspace->R, & Yc_view.matrix );
                //(Check that all matrices are transposed)
                gsl_matrix_view A_rview = gsl_matrix_submatrix ( & A_subview.matrix, j, 0, 1, ( & A_subview.matrix )->size2 );
                gsl_matrix_transpose_memcpy ( regu_wspace->fista_wspace->B_hat_init, & A_rview.matrix );
                gsl_matrix_view lambda_v;
                //const gsl_matrix * lambda_ptr;
                if ( penalty == GROUP_ADALASSO )
                {
                    lambda_v = gsl_matrix_submatrix ( regu_wspace->lambda_weights_tr, 0, 0, regu_wspace->lambda_weights_tr->size1, regu_wspace->lambda_weights_tr->size2 );
                    //gsl_matrix_const_view lambda_v = gsl_matrix_const_submatrix ( regu_wspace->lambda_weights_tr, 0, 0, regu_wspace->lambda_weights_tr->size1, regu_wspace->lambda_weights_tr->size2 );
                    //lambda_ptr = regu_wspace->lambda_weights_tr;
                    //Set groups
                    for ( size_t g = 0; g < regu_wspace->lambda_weights_tr->size1; ++g )
                    {
                        gsl_vector_const_view g_v = gsl_matrix_const_row ( groups, g * estimates->D + j );
                        gsl_vector_view gtr_v = gsl_matrix_column ( regu_wspace->groups_tr, g );
                        gsl_blas_dcopy ( & g_v.vector, & gtr_v.vector );
                    }
                }
                else
                {
                    lambda_v = gsl_matrix_submatrix ( regu_wspace->lambda_weights_tr, 0, j, regu_wspace->lambda_weights_tr->size1, 1 );
                    //lambda_ptr = & lambda_v.matrix;
                }
                int max_iters_0 = max_iters / 2;
                fista_estimation ( regu_wspace->fista_wspace, & X_tmp_view.matrix, regu_wspace->R, regu_wspace->groups_tr, lambda, & lambda_v.matrix,
                                regu_wspace->fista_wspace->B_hat_init, L_c, max_iters_0, tol, penalty, estimates->T / s );
                gsl_matrix_transpose_memcpy ( & A_rview.matrix, regu_wspace->fista_wspace->B_hat_init );
            }
        }

        //Intercept
        //Get Y - XB^T
        gsl_blas_dgemm ( CblasNoTrans, CblasTrans, -1.0, & X_tmp_view.matrix, & A_subview.matrix, 1.0, regu_wspace->Ys );
        gsl_vector_const_view xi_view = gsl_matrix_const_row ( em_wspace->xi_smooth, i );
        double dd = gsl_vector_sum_0 ( & xi_view.vector );
        for ( size_t j = 0; j < estimates->T; ++j )
        {
            double x = gsl_vector_get ( & xi_view.vector, j );
            gsl_vector_view y_view = gsl_matrix_row ( regu_wspace->Ys, j );
            gsl_blas_daxpy ( sqrt ( x ), & intercept_view.vector, & y_view.vector );
            gsl_blas_dscal ( sqrt ( x ), & y_view.vector );
        }
        for ( size_t j = 0; j < estimates->D; ++j )
        {
            gsl_vector_const_view y_view = gsl_matrix_const_column ( regu_wspace->Ys, j );
            double dn = gsl_vector_sum_0 ( & y_view.vector );
            gsl_vector_set ( & intercept_view.vector, j, dn / dd );
        }
        if ( estimates->M > 1 && num_p > 1 )
        {
            for ( size_t j = 0; j < estimates->D; ++j )
            {
                //Intercept
                regu_wspace->A_mm[ i * estimates->D + j ][0] = gsl_vector_get ( & intercept_view.vector, j );
                //Rest
                for ( size_t k = 0; k < estimates->A->size2 - 1; ++k )
                    regu_wspace->A_mm[ i * estimates->D + j ][ k + 1 ] = gsl_matrix_get ( & A_subview.matrix, j, k );
            }
        }
    }
#ifdef PARVAR
    if ( getpid () != ppid )
        exit ( 0 );
    //if ( getpid () == ppid )
    for ( size_t i = 0; i < num_p - 1; ++i )
    {
        int pstatus = waitpid ( pid_arr [ i ], NULL, 0 );
        if ( pstatus == -1 && errno != ECHILD )
        {
            perror ( "waitpid() error" );
            exit ( 1 );
        }
    }
#endif
    
    if ( estimates->M > 1 && num_p > 1 )
        for ( size_t j = 0; j < estimates->A->size1; ++j )
            for ( size_t k = 0; k < estimates->A->size2; ++k )
                gsl_matrix_set ( estimates->A, j, k, regu_wspace->A_mm[j][k] );
    
    return GSL_SUCCESS;
}

int estimate_covariance_precision ( em_workspace * em_wspace, const gsl_matrix * Y, const gsl_matrix * X, msvar_parameters * estimates )
{
    // Page 187
    
    gsl_matrix_const_view Y_subview = gsl_matrix_const_submatrix ( Y, estimates->p, 0, estimates->T, estimates->D );
    double ts = 0.0;
    //Since estimates->switching_covar = 0, here estimates->M only depends on A ( so no modification )
    for ( size_t i = 0; i < estimates->M; ++i )
    {
        //Compute Y - X * Bm = U
        gsl_matrix_view U_subview =  gsl_matrix_submatrix ( em_wspace->U, i * estimates->T, 0, estimates->T, estimates->D );
        gsl_matrix_memcpy ( & U_subview.matrix, & Y_subview.matrix );
        gsl_matrix_const_view A_subview = gsl_matrix_const_submatrix ( estimates->A, i * estimates->D, 0, estimates->D, 1 + estimates->D * estimates->p );
        gsl_blas_dgemm ( CblasNoTrans, CblasTrans, -1.0, X, & A_subview.matrix, 1.0, & U_subview.matrix );
        gsl_matrix_view tmp_subview = gsl_matrix_submatrix ( em_wspace->tmpc, i * estimates->T, 0, estimates->T, estimates->D );
        gsl_matrix_memcpy ( & tmp_subview.matrix, & U_subview.matrix );
        //Compute E * U, where E = diag ( vec ( xi_smooth )  )
        gsl_vector_const_view xi_vecview = gsl_matrix_const_row ( em_wspace->xi_smooth, i );
        gsl_matrix_scale_rows_0 ( & tmp_subview.matrix, & xi_vecview.vector );
        ts += gsl_vector_sum_0 ( & xi_vecview.vector );
    }
    //Sigma = U'E U / sum ( E ) = tmpc' U / sum ( E )
    gsl_blas_dgemm ( CblasTrans, CblasNoTrans, 1.0 / ts, em_wspace->tmpc, em_wspace->U, 0.0, estimates->SigmaL );
    //Compute Cholesky decomposition of Sigma and store
    gsl_linalg_cholesky_decomp1 ( estimates->SigmaL );
    gsl_matrix_memcpy ( estimates->Sigma_inv, estimates->SigmaL );
    gsl_linalg_cholesky_invert ( estimates->Sigma_inv );
    
    return GSL_SUCCESS;
}

int estimate_switching_covariance_precision ( em_workspace * em_wspace, const gsl_matrix * Y, const gsl_matrix * X, msvar_parameters * estimates, int m, size_t chol )
{
    // Page 189 ( without the Kronecker product since it is in error)
    gsl_matrix_const_view Y_subview = gsl_matrix_const_submatrix ( Y, estimates->p, 0, estimates->T, estimates->D );
    
    int m0 = 0;
    int m1 = estimates->M;
    if ( m >= 0 )
    {
        m0 = m;
        m1 = m0 + 1;
    }
    
    for ( int i = m0; i < m1; ++i )
    {
        //Here we obtain the fitted value matrix Usw = Y - XB
        if ( estimates->switching_coeff )
        {
            gsl_matrix_memcpy ( em_wspace->Usw, & Y_subview.matrix );
            gsl_matrix_const_view A_subview = gsl_matrix_const_submatrix ( estimates->A, i * estimates->D, 0, estimates->D, 1 + estimates->D * estimates->p );
            gsl_blas_dgemm ( CblasNoTrans, CblasTrans, -1.0, X, & A_subview.matrix, 1.0, em_wspace->Usw );
            gsl_matrix_memcpy ( em_wspace->tmpcsw, em_wspace->Usw ); //tmpcsw <- Usw
        }
        else
        {
            //Fitted values remain the same across regimes for non-switching coefficients
            if ( i == 0 )
            {
                gsl_matrix_memcpy ( em_wspace->Usw, & Y_subview.matrix );
                gsl_blas_dgemm ( CblasNoTrans, CblasTrans, -1.0, X, estimates->A, 1.0, em_wspace->Usw );
            }
            gsl_matrix_memcpy ( em_wspace->tmpcsw, em_wspace->Usw ); //tmpcsw <- Usw
        }
        //Product E_m Usw ( E_m is diagonal of smooth xi )
        gsl_vector_const_view xi_vecview = gsl_matrix_const_row ( em_wspace->xi_smooth, i );
        gsl_matrix_scale_rows_0 ( em_wspace->tmpcsw, & xi_vecview.vector );
        double ts = gsl_vector_sum_0 ( & xi_vecview.vector );
        //Obtain Usw' E_m Usw / T_m ( E_m is diagonal of smooth xi and T_m its sum )  ( Em Usw ) ' = Usw' Em
        gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( estimates->SigmaL, i * estimates->D, 0, estimates->D, estimates->D );
        gsl_blas_dgemm ( CblasTrans, CblasNoTrans, 1.0 / ts, em_wspace->tmpcsw, em_wspace->Usw, 0.0, & SigmaL_subview.matrix );
        
        gsl_matrix_view tmpcsw_view = gsl_matrix_submatrix ( em_wspace->tmpcsw, 0, 0, estimates->D, estimates->D );
        gsl_matrix_memcpy ( & tmpcsw_view.matrix, & SigmaL_subview.matrix );
        gsl_error_handler_t * current_handler = gsl_set_error_handler_off ( );
        int status = gsl_linalg_cholesky_decomp1 ( & tmpcsw_view.matrix );
        //If matrix is not positive definite
        while ( status == GSL_EDOM )
        {
            //printf("shift\n");
            double lmin;
            double lmax;
            gsl_matrix_memcpy ( & tmpcsw_view.matrix, & SigmaL_subview.matrix );
            //Regularize covariance
            min_max_sym_eigenvalue ( & tmpcsw_view.matrix, em_wspace->temp_vec, em_wspace->eigw, & lmin, & lmax );
            gsl_vector_view Sv = gsl_matrix_diagonal ( & SigmaL_subview.matrix );
            //gsl_vector_add_constant ( & Sv.vector, fabs ( lmin ) + 1.0e-6 );
            gsl_vector_add_constant ( & Sv.vector, fabs ( lmin ) + 1.0e-5 );
            gsl_matrix_memcpy ( & tmpcsw_view.matrix, & SigmaL_subview.matrix );
            status = gsl_linalg_cholesky_decomp1 ( & tmpcsw_view.matrix );
        }
        /*if ( status == GSL_EDOM )
        {
            gsl_matrix * S_tmp = gsl_matrix_alloc ( estimates->D, estimates->D );
            gsl_matrix * S_e = gsl_matrix_alloc ( estimates->D, estimates->D );
            gsl_matrix * W = gsl_matrix_alloc ( estimates->D, estimates->D );
            gsl_eigen_symmv_workspace * eigvw = gsl_eigen_symmv_alloc ( estimates->D );
            
            //printf("projec\n");
            //Projection onto S_D^+
            gsl_matrix_set_zero ( S_tmp );
            // S_e * S_tmp * S_e'; S_e contains eigenvectors
            gsl_vector_view Tmp_vecview = gsl_matrix_diagonal ( S_tmp );
            gsl_eigen_symmv ( & SigmaL_subview.matrix, & Tmp_vecview.vector, S_e, eigvw );
            for ( size_t j = 0; j < estimates->D; ++j )
            {
                double e = gsl_vector_get ( & Tmp_vecview.vector, j );
                if ( e < 1.0e-5 )
                {
                    e = 1.0e-5;
                    gsl_vector_set ( & Tmp_vecview.vector, j, e );
                }
            }
            // S_e * S_tmp * S_e'
            gsl_blas_dgemm ( CblasNoTrans, CblasNoTrans, 1.0, S_e, S_tmp, 0.0, W );
            gsl_blas_dgemm ( CblasNoTrans, CblasTrans, 1.0, W, S_e, 0.0, & SigmaL_subview.matrix );
            
            gsl_matrix_free ( S_tmp );
            gsl_matrix_free ( S_e );
            gsl_matrix_free ( W );
            gsl_eigen_symmv_free ( eigvw );
        }
        */
        //Default handler
        gsl_set_error_handler ( current_handler );
        gsl_matrix_view Sigma_inv_subview = gsl_matrix_submatrix ( estimates->Sigma_inv, i * estimates->D, 0, estimates->D, estimates->D );
        //Compute Cholesky decomposition of Sigma sub-matrix and store
        if ( chol > 0 )
        {
            gsl_linalg_cholesky_decomp1 ( & SigmaL_subview.matrix );
            gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, & SigmaL_subview.matrix );
            gsl_linalg_cholesky_invert ( & Sigma_inv_subview.matrix );
        }
        else
        {
            gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, & SigmaL_subview.matrix );
            gsl_linalg_cholesky_decomp1 ( & Sigma_inv_subview.matrix );
            gsl_linalg_cholesky_invert ( & Sigma_inv_subview.matrix );
        }
    }
    return GSL_SUCCESS;
}

int estimate_regularized_covariance_precision ( em_workspace * em_wspace, regu_cov_workspace * regu_cov_wspace, const gsl_matrix * Y, const gsl_matrix * X,
                                 msvar_parameters * estimates, size_t penalty, double lambda, gsl_matrix * lambda_weights, size_t max_iters, double tol )
{
    estimate_sparse_covariance_precision ( em_wspace, regu_cov_wspace, Y, X, estimates, penalty, lambda, lambda_weights, max_iters, tol );
    return GSL_SUCCESS;
}

int estimate_sparse_covariance_precision ( em_workspace * em_wspace, regu_cov_workspace * regu_cov_wspace, const gsl_matrix * Y, const gsl_matrix * X,
                                 msvar_parameters * estimates, size_t penalty, double lambda, gsl_matrix * lambda_weights, size_t max_iters, double tol )
{
    if ( penalty == ORACLE )
    {
        estimate_switching_covariance_precision ( em_wspace, Y, X, estimates, -1, 0 );
        gsl_matrix_mul_elements ( estimates->SigmaL, lambda_weights );
        for ( size_t i = 0; i < estimates->M; ++i )
        {
            gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( estimates->SigmaL, i * estimates->D, 0, estimates->D, estimates->D );
            gsl_matrix_view Sigma_inv_subview = gsl_matrix_submatrix ( estimates->Sigma_inv, i * estimates->D, 0, estimates->D, estimates->D );
            
            gsl_error_handler_t * current_handler = gsl_set_error_handler_off ( );
            int status = gsl_linalg_cholesky_decomp1 ( & SigmaL_subview.matrix );
            gsl_set_error_handler ( current_handler );
            if ( status == GSL_EDOM )
            {
                gsl_matrix_const_view SigmaL_subview2 = gsl_matrix_const_submatrix ( em_wspace->estimates_prev->SigmaL, i * estimates->D, 0, estimates->D, estimates->D );
                gsl_matrix_const_view Sigma_inv_subview2 = gsl_matrix_const_submatrix ( em_wspace->estimates_prev->Sigma_inv, i * estimates->D, 0, estimates->D, estimates->D );
                gsl_matrix_memcpy ( & SigmaL_subview.matrix, & SigmaL_subview2.matrix );
                gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, & Sigma_inv_subview2.matrix );
            }
            else
            {
                gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, & SigmaL_subview.matrix );
                gsl_linalg_cholesky_invert ( & Sigma_inv_subview.matrix );
            }
        }
        return GSL_SUCCESS;
    }
    else if ( penalty == ORACLE_INVCOVAR )
    {
        estimate_switching_covariance_precision ( em_wspace, Y, X, estimates, -1, 0 );
        gsl_matrix_mul_elements ( estimates->Sigma_inv, lambda_weights );
        
        for ( size_t i = 0; i < estimates->M; ++i )
        {
            gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( estimates->SigmaL, i * estimates->D, 0, estimates->D, estimates->D );
            gsl_matrix_view Sigma_inv_subview = gsl_matrix_submatrix ( estimates->Sigma_inv, i * estimates->D, 0, estimates->D, estimates->D );
            
            gsl_matrix_memcpy ( & SigmaL_subview.matrix, & Sigma_inv_subview.matrix );
            gsl_error_handler_t * current_handler = gsl_set_error_handler_off ( );
            int status = gsl_linalg_cholesky_decomp1 ( & SigmaL_subview.matrix );
            gsl_set_error_handler ( current_handler );
            if ( status == GSL_EDOM )
            {
                gsl_matrix_const_view SigmaL_subview2 = gsl_matrix_const_submatrix ( em_wspace->estimates_prev->SigmaL, i * estimates->D, 0, estimates->D, estimates->D );
                gsl_matrix_memcpy ( & SigmaL_subview.matrix, & SigmaL_subview2.matrix );
                gsl_matrix_const_view Sigma_inv_subview2 = gsl_matrix_const_submatrix ( em_wspace->estimates_prev->Sigma_inv, i * estimates->D, 0, estimates->D, estimates->D );
                gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, & Sigma_inv_subview2.matrix );
            }
            else
            {
                gsl_linalg_cholesky_invert  ( & SigmaL_subview.matrix );
                gsl_linalg_cholesky_decomp1 ( & SigmaL_subview.matrix );
            }
            
        }
        return GSL_SUCCESS;
    }
    
    double L_cc = 1.0 / GRAD_STEP;

    size_t num_p = 1;
    size_t pn = 0;
#ifdef PARVAR
    num_p = estimates->M;
    pid_t pid_arr [ num_p ];
    pid_t ppid = getpid ();
    for ( size_t k = 0; k < num_p - 1; ++k )
    {
        pid_t p = fork();
        if ( p > 0 ) //caller
        {
            pid_arr [ pn ] = p;
            ++pn;
            continue;
        }
        else if ( ! p ) // child
        {
            break;
        }
        else // error
        {
            perror ( "\n fork call failed " );
            exit ( 1 );
        }
    }
    if ( getpid () == ppid )
        pn = 0;
    else
        ++pn;
#endif
    
    if ( INVCOVAR_REG(penalty) )
    {
        if ( penalty == GRAPHICAL_SCAD || penalty == GRAPHICAL_MCP )
            set_covar_penalty_weights ( estimates->SigmaL, estimates->Sigma_inv, penalty, lambda_weights, lambda );
        
        for ( size_t i = 0; i < estimates->M; ++i )
        {
            //Each thread does a regime
            if ( num_p > 1 && pn != i )
                continue;
            
            gsl_vector_const_view xi_vecview = gsl_matrix_const_row ( em_wspace->xi_smooth, i );
            double Nm = estimates->T / gsl_blas_dasum ( & xi_vecview.vector );
            estimate_switching_covariance_precision ( em_wspace, Y, X, estimates, i, 0 );
            //Get current matrix for regime
            gsl_matrix_const_view L_subview = gsl_matrix_const_submatrix ( lambda_weights, i * estimates->D, 0, estimates->D, estimates->D );
            gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( estimates->SigmaL, i * estimates->D, 0, estimates->D, estimates->D );
            gsl_matrix_view Sigma_inv_subview = gsl_matrix_submatrix ( estimates->Sigma_inv, i * estimates->D, 0, estimates->D, estimates->D );
            if ( INVCOVAR_GRAPHICAL(penalty) )
            {
                /*
                */
            }
            else
            {
                //Initial estimate
                gsl_matrix_memcpy ( regu_cov_wspace->regu_cov_mats->S, & Sigma_inv_subview.matrix );
                    //estimate_switching_covariance_precision ( em_wspace, Y, X, estimates, i, 0 );
                precision_optimization ( regu_cov_wspace, & SigmaL_subview.matrix, regu_cov_wspace->regu_cov_mats->S, estimates, lambda, & L_subview.matrix, penalty, max_iters, tol, L_cc, Nm );
                gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, regu_cov_wspace->regu_cov_mats->S );
                int status;
                gsl_error_handler_t * current_handler = gsl_set_error_handler_off ( );
                status = gsl_linalg_cholesky_decomp1 ( regu_cov_wspace->regu_cov_mats->S );
                if ( status != GSL_EDOM )
                {
                    gsl_linalg_cholesky_invert  ( regu_cov_wspace->regu_cov_mats->S );
                    gsl_linalg_cholesky_decomp1 ( regu_cov_wspace->regu_cov_mats->S );
                    int status0 = gsl_linalg_cholesky_decomp1 ( regu_cov_wspace->regu_cov_mats->S );
                    if ( status0 == GSL_EDOM )
                    {
                        gsl_matrix_const_view SigmaL_subview2 = gsl_matrix_const_submatrix ( em_wspace->estimates_prev->SigmaL, i * estimates->D, 0, estimates->D, estimates->D );
                        gsl_matrix_memcpy ( regu_cov_wspace->regu_cov_mats->S, & SigmaL_subview2.matrix );
                        gsl_matrix_const_view Sigma_inv_subview2 = gsl_matrix_const_submatrix ( em_wspace->estimates_prev->Sigma_inv, i * estimates->D, 0, estimates->D, estimates->D );
                        gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, & Sigma_inv_subview2.matrix );
                    }
                    else
                        gsl_matrix_memcpy ( & SigmaL_subview.matrix, regu_cov_wspace->regu_cov_mats->S );
                }
                else
                {
                    gsl_matrix_const_view SigmaL_subview2 = gsl_matrix_const_submatrix ( em_wspace->estimates_prev->SigmaL, i * estimates->D, 0, estimates->D, estimates->D );
                    gsl_matrix_memcpy ( & SigmaL_subview.matrix, & SigmaL_subview2.matrix );
                    gsl_matrix_const_view Sigma_inv_subview2 = gsl_matrix_const_submatrix ( em_wspace->estimates_prev->Sigma_inv, i * estimates->D, 0, estimates->D, estimates->D );
                    gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, & Sigma_inv_subview2.matrix );
                }
                gsl_set_error_handler ( current_handler );
            }
            
            if ( estimates->M > 1 && num_p > 1 )
            {
                for ( size_t j = 0; j < estimates->D; ++j )
                    for ( size_t k = 0; k < estimates->D; ++k )
                    {
                        regu_cov_wspace->S_mm[ i * estimates->D + j ][k] = gsl_matrix_get ( & SigmaL_subview.matrix, j, k );
                        regu_cov_wspace->S_inv_mm[ i * estimates->D + j ][k] = gsl_matrix_get ( & Sigma_inv_subview.matrix, j, k );
                    }
            }
        }
#ifdef PARVAR
        if ( getpid () != ppid )
            exit ( 0 );
        //if ( getpid () == ppid )
        for ( size_t i = 0; i < num_p - 1; ++i )
        {
            int pstatus = waitpid ( pid_arr [ i ], NULL, 0 );
            if ( pstatus == -1 && errno != ECHILD )
            {
                perror ( "waitpid() error" );
                exit ( 1 );
            }
        }
#endif
        if ( estimates->M > 1 && num_p > 1 )
        {
            for ( size_t j = 0; j < estimates->M * estimates->D; ++j )
                for ( size_t k = 0; k < estimates->D; ++k )
                {
                    gsl_matrix_set ( estimates->SigmaL, j, k, regu_cov_wspace->S_mm[j][k] );
                    gsl_matrix_set ( estimates->Sigma_inv, j, k, regu_cov_wspace->S_inv_mm[j][k] );
                }
        }
    }
    else
    {
        for ( size_t i = 0; i < estimates->M; ++i )
        {
            //Each thread does a regime
            if ( num_p > 1 && pn != i )
                continue;
            
            gsl_vector_const_view xi_vecview = gsl_matrix_const_row ( em_wspace->xi_smooth, i );
            double Nm = estimates->T / gsl_blas_dasum ( & xi_vecview.vector );
            estimate_switching_covariance_precision ( em_wspace, Y, X, estimates, i, 0 );
            //Get current covariance matrix for regime
            gsl_matrix_const_view L_subview = gsl_matrix_const_submatrix ( lambda_weights, i * estimates->D, 0, estimates->D, estimates->D );
            gsl_matrix_view SigmaL_subview = gsl_matrix_submatrix ( estimates->SigmaL, i * estimates->D, 0, estimates->D, estimates->D );
            gsl_matrix_view Sigma_inv_subview = gsl_matrix_submatrix ( estimates->Sigma_inv, i * estimates->D, 0, estimates->D, estimates->D );
            //Initial estimate
            gsl_matrix_memcpy ( regu_cov_wspace->regu_cov_mats->S, & SigmaL_subview.matrix );
                /*
                */
            covariance_optimization ( regu_cov_wspace, & SigmaL_subview.matrix, regu_cov_wspace->regu_cov_mats->S, estimates, lambda, & L_subview.matrix, penalty, max_iters, tol, L_cc, Nm );
            
            gsl_error_handler_t * current_handler = gsl_set_error_handler_off ( );
            int status = gsl_linalg_cholesky_decomp1 ( regu_cov_wspace->regu_cov_mats->S );
            gsl_set_error_handler ( current_handler );
            if ( status == GSL_EDOM )
            {
                gsl_matrix_const_view SigmaL_subview2 = gsl_matrix_const_submatrix ( em_wspace->estimates_prev->SigmaL, i * estimates->D, 0, estimates->D, estimates->D );
                gsl_matrix_memcpy ( & SigmaL_subview.matrix, & SigmaL_subview2.matrix );
                gsl_matrix_const_view Sigma_inv_subview2 = gsl_matrix_const_submatrix ( em_wspace->estimates_prev->Sigma_inv, i * estimates->D, 0, estimates->D, estimates->D );
                gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, & Sigma_inv_subview2.matrix );
            }
            else
            {
                gsl_matrix_memcpy ( & SigmaL_subview.matrix, regu_cov_wspace->regu_cov_mats->S );
                gsl_matrix_memcpy ( & Sigma_inv_subview.matrix, regu_cov_wspace->regu_cov_mats->S );
                gsl_linalg_cholesky_invert  ( & Sigma_inv_subview.matrix );
            }
            
            if ( estimates->M > 1 && num_p > 1 )
            {
                for ( size_t j = 0; j < estimates->D; ++j )
                    for ( size_t k = 0; k < estimates->D; ++k )
                    {
                        regu_cov_wspace->S_mm[ i * estimates->D + j ][k] = gsl_matrix_get ( & SigmaL_subview.matrix, j, k );
                        regu_cov_wspace->S_inv_mm[ i * estimates->D + j ][k] = gsl_matrix_get ( & Sigma_inv_subview.matrix, j, k );
                    }
            }
        }
#ifdef PARVAR
        if ( getpid () != ppid )
            exit ( 0 );
        //if ( getpid () == ppid )
        for ( size_t i = 0; i < num_p - 1; ++i )
        {
            int pstatus = waitpid ( pid_arr [ i ], NULL, 0 );
            if ( pstatus == -1 && errno != ECHILD )
            {
                perror ( "waitpid() error" );
                exit ( 1 );
            }
        }
#endif
        if ( estimates->M > 1 && num_p > 1 )
        {
            for ( size_t j = 0; j < estimates->M * estimates->D; ++j )
                for ( size_t k = 0; k < estimates->D; ++k )
                {
                    gsl_matrix_set ( estimates->SigmaL, j, k, regu_cov_wspace->S_mm[j][k] );
                    gsl_matrix_set ( estimates->Sigma_inv, j, k, regu_cov_wspace->S_inv_mm[j][k] );
                }
        }
    }
    return GSL_SUCCESS;
}

int compute_initial_estimates ( em_workspace * em_wspace, const gsl_matrix * Y, const gsl_matrix * X, msvar_parameters * estimates, int sort_i )
{
    size_t M = estimates->M;
    estimates->M = 1;
    gsl_matrix_set_all ( em_wspace->xi_smooth, 1.0 );

    //Coefficients
    estimate_switching_coefficients ( em_wspace, Y, X, estimates );
    //Set every other regime coefficients to the estimated VAR coefficients for 1 regime
    gsl_matrix_const_view Av_subview = gsl_matrix_const_submatrix ( estimates->A, 0, 0, estimates->D, estimates->A->size2 );
    size_t Nj = M;
    if ( estimates->switching_coeff == 0 )
        Nj = 1;
    for ( size_t j = 1; j < Nj; ++j )
    {
        gsl_matrix_view A_subview = gsl_matrix_submatrix ( estimates->A, j * estimates->D, 0, estimates->D, estimates->A->size2 );
        gsl_matrix_memcpy ( & A_subview.matrix, & Av_subview.matrix );
    }
    //Intercept initialize
    if ( Nj > 1 && sort_i == 1 )
    {
        gsl_vector_view av = gsl_matrix_column ( estimates->A, 0 );
        gsl_sort_vector ( & av.vector );
    }
    if ( Nj > 1 && sort_i == 2 )
    {
        gsl_vector_view av = gsl_matrix_column ( estimates->A, 0 );
        gsl_vector_scale ( & av.vector, -1.0 );
        gsl_sort_vector ( & av.vector );
        gsl_vector_scale ( & av.vector, -1.0 );
    }
    
    //Covariance
    size_t chol = 0;
    estimate_switching_covariance_precision ( em_wspace, Y, X, estimates, -1, chol );
    
    //Compute Cholesky and set every other covariance to MLE of 1 regime
    gsl_matrix_view S_subview = gsl_matrix_submatrix ( estimates->SigmaL, 0, 0, estimates->D, estimates->D );
    gsl_matrix_view S_inv_subview = gsl_matrix_submatrix ( estimates->Sigma_inv, 0, 0, estimates->D, estimates->D );
    gsl_linalg_cholesky_decomp1 ( & S_subview.matrix );
    gsl_matrix_memcpy ( & S_inv_subview.matrix, & S_subview.matrix );
    gsl_linalg_cholesky_invert ( & S_inv_subview.matrix );
    Nj = M;
    if ( estimates->switching_covar == 0 )
        Nj = 1;
    for ( size_t j = 1; j < Nj; ++j )
    {
        gsl_matrix_view Se_subview = gsl_matrix_submatrix ( estimates->SigmaL, j * estimates->D, 0, estimates->D, estimates->D );
        gsl_matrix_view Se_inv_subview = gsl_matrix_submatrix ( estimates->Sigma_inv, j * estimates->D, 0, estimates->D, estimates->D );
        gsl_matrix_memcpy ( & Se_subview.matrix, & S_subview.matrix );
        gsl_matrix_memcpy ( & Se_inv_subview.matrix, & S_inv_subview.matrix );
    }
    estimates->M = M;
    
    return GSL_SUCCESS;
}

int expectation_maximization ( em_workspace * em_wspace, regu_workspace * regu_wspace, regu_cov_workspace * regu_cov_wspace, msvar_parameters * estimates,
                               const gsl_matrix * Y, const gsl_matrix * X, const gsl_vector * y, size_t em_max_iters, size_t max_iters, double tol, size_t coeff_penalty,
                               size_t covar_penalty, double coeff_lambda, double covar_lambda, gsl_matrix * coeff_lambda_weights, gsl_matrix * covar_lambda_weights,
                               double L_c, const gsl_matrix * groups )
{
    //Linear Regression ML ( LS )
    if ( estimates->M == 1 )
    {
        gsl_matrix_set_all ( em_wspace->xi_tt1, 1.0 );
        gsl_matrix_set_all ( em_wspace->xi_filter, 1.0 );
        gsl_matrix_set_all ( em_wspace->xi_smooth, 1.0 );
        if ( coeff_penalty != NONE && coeff_lambda > 0.0 )
            estimate_sparse_coefficients ( em_wspace, regu_wspace, Y, X, estimates, coeff_penalty, coeff_lambda, coeff_lambda_weights, max_iters, tol, L_c, groups );
        else
            estimate_switching_coefficients ( em_wspace, Y, X, estimates );
        if ( covar_penalty != NONE && covar_lambda > 0.0 )
            estimate_regularized_covariance_precision ( em_wspace, regu_cov_wspace, Y, X, estimates, covar_penalty, covar_lambda, covar_lambda_weights, max_iters, tol );
        else
            estimate_switching_covariance_precision ( em_wspace, Y, X, estimates, -1, 1 );
           
        // Get likelihood
        estimates->loglik = log_likelihood ( em_wspace, Y, estimates, estimates->p, NULL );
        estimates->pen = compute_penalties ( estimates, coeff_penalty, covar_penalty, coeff_lambda, covar_lambda, coeff_lambda_weights, covar_lambda_weights, PARAM_SCAD, PARAM_MCP, estimates->T, groups );
        //printf ( " PenLoglik: %lf\n", estimates->loglik - estimates->pen );
        if ( gsl_isnan ( estimates->loglik - estimates->pen ) || gsl_isinf ( estimates->loglik - estimates->pen ) != 0 )
            printf ( " EM likelihood value invalid\n" );
        
       return GSL_SUCCESS;
    }
    
#ifdef PARVAR
    if ( coeff_penalty != NONE && coeff_lambda > 0.0 )
    {
        regu_wspace->A_mm = ( double ** ) mmap ( NULL, estimates->A->size1 * sizeof ( double * ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
        for ( size_t k = 0; k < estimates->A->size1; ++k )
            regu_wspace->A_mm [ k ] = ( double * ) mmap ( NULL, estimates->A->size2 * sizeof ( double ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
    }
    if ( covar_penalty != NONE && covar_lambda > 0.0 )
    {
        regu_cov_wspace->S_mm = ( double ** ) mmap ( NULL, estimates->M * estimates->D * sizeof ( double * ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
        regu_cov_wspace->S_inv_mm = ( double ** ) mmap ( NULL, estimates->M * estimates->D * sizeof ( double * ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
        for ( size_t k = 0; k < estimates->M * estimates->D; ++k )
        {
            regu_cov_wspace->S_mm [ k ] = ( double * ) mmap ( NULL, estimates->D * sizeof ( double ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
            regu_cov_wspace->S_inv_mm [ k ] = ( double * ) mmap ( NULL, estimates->D * sizeof ( double ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
        }
    }
#endif

    size_t iters;
    estimates->loglik = GSL_NEGINF;
    estimates->pen = 0.0;
    em_wspace->estimates_prev->loglik = GSL_NEGINF;
    em_wspace->estimates_prev->pen = 0.0;
    
    gsl_matrix_set_zero ( em_wspace->estimates_prev->A );
    gsl_matrix_set_zero ( em_wspace->estimates_prev->Pr );
    gsl_matrix_set_zero ( em_wspace->estimates_prev->SigmaL );
    gsl_matrix_set_zero ( em_wspace->estimates_prev->Sigma_inv );
    
    for ( iters = 0; iters < em_max_iters; ++iters )
    {
        // Expectation
        
        /*    For our model, Pr = F^t ( transition probabilities and state-space matrix F )
            Initial xi_{0|0} as ( estimates->initial_state ) -th column of identity matrix; initial_state is input ( fixed ) parameter
            Page 90: xi_{1|0} = F * xi_{0|0} = Pr^T * xi_{0|0} = ( estimates->initial_state ) -th row of Pr
        */
        gsl_vector_view xi_init_view = gsl_matrix_column ( em_wspace->xi_tt1, 0 );
        //Set p-th state vector
        gsl_vector_set_zero ( & xi_init_view.vector );
        gsl_vector_set ( & xi_init_view.vector, estimates->initial_state - 1, 1.0 );
        //Forward recursion ( Filtering ) and get log likelihood there
        forward_recursion ( em_wspace, X, Y, estimates );
        if ( gsl_isnan ( estimates->loglik ) )
        {
            printf ( " EM invalid likelihood; stopping %ld\n", iters );
            break;
        }
        //if ( gsl_isinf ( estimates->loglik ) )
        //    break;
        /*Here we now have the log likelihood of the previous iteration ( obtained from corresponding filtered xi and estimates )
            and of the previous two iterations ( em_wspace->estimates_prev->loglik )*/
        estimates->pen = compute_penalties ( estimates, coeff_penalty, covar_penalty, coeff_lambda, covar_lambda, coeff_lambda_weights, covar_lambda_weights, PARAM_SCAD, PARAM_MCP, estimates->T, groups );
        //if ( iters == 1 ) printf ( " EM Iter: %lu, PenLoglik: %lf\n", iters, estimates->loglik - estimates->pen );

        //Compute norm of the difference of estimates
        double ea = frobenius_norm_diff ( estimates->A, em_wspace->estimates_prev->A ) / sqrt ( inner_product ( em_wspace->estimates_prev->A, em_wspace->estimates_prev->A ) );
        double es;
        if ( INVCOVAR_REG(covar_penalty) || covar_penalty == ORACLE_INVCOVAR || covar_penalty == GRAPHICAL_ORACLE )
            es = frobenius_norm_diff ( estimates->Sigma_inv, em_wspace->estimates_prev->Sigma_inv ) / sqrt ( inner_product ( em_wspace->estimates_prev->Sigma_inv, em_wspace->estimates_prev->Sigma_inv ) );
        else
            es = frobenius_norm_diff ( estimates->SigmaL, em_wspace->estimates_prev->SigmaL ) / sqrt ( inner_product ( em_wspace->estimates_prev->SigmaL, em_wspace->estimates_prev->SigmaL ) );
        double ep = frobenius_norm_diff ( estimates->Pr, em_wspace->estimates_prev->Pr ) / sqrt ( inner_product ( em_wspace->estimates_prev->Pr, em_wspace->estimates_prev->Pr ) );
        double el = fabs ( estimates->loglik - estimates->pen - em_wspace->estimates_prev->loglik + em_wspace->estimates_prev->pen ) / ( em_wspace->estimates_prev->loglik - em_wspace->estimates_prev->pen );

        if ( iters > 0 && el < tol && ea < tol && es < tol && ep < tol )
        //if ( iters > 0 && fabs ( estimates->loglik - em_wspace->estimates_prev->loglik ) < tol && ea < tol && es < tol && ep < tol )
            break;
        //Update previous estimates
        gsl_matrix_memcpy ( em_wspace->estimates_prev->A,  estimates->A );
        gsl_matrix_memcpy ( em_wspace->estimates_prev->Pr,  estimates->Pr );
        gsl_matrix_memcpy ( em_wspace->estimates_prev->SigmaL, estimates->SigmaL );
        gsl_matrix_memcpy ( em_wspace->estimates_prev->Sigma_inv, estimates->Sigma_inv );
        em_wspace->estimates_prev->loglik = estimates->loglik;
        em_wspace->estimates_prev->pen = estimates->pen;
        
        //Backward recursion ( Smoothing )
        backward_recursion ( em_wspace, estimates->Pr );

        //Maximization ( Normal Equations ( we could iterate a few times estimation of coefficients and covariances )  )
        
        //Hidden Markov Chain
        double C = 0.0;
        estimate_transition_probabilities ( em_wspace, estimates, C );

        //Numerical check (avoid having zero entry, numerical issues with regime covariance, etc.)
        int px = 0.0;
        for ( size_t i = 0; i < estimates->M; ++i )
        {
            gsl_vector_const_view xvec = gsl_matrix_const_row ( em_wspace->xi_smooth, i );
            double v = gsl_blas_dasum ( & xvec.vector );
            if ( v <= 1.0 || gsl_isnan ( v ) )
            {
                gsl_matrix_memcpy ( estimates->Pr, em_wspace->estimates_prev->Pr );
                px = 1;
                break;
            }
            for ( size_t j = 0; j < estimates->M; ++j )
            {
                double pp = gsl_matrix_get ( estimates->Pr, i, j );
                if ( pp <= 1.0e-5 || gsl_isnan ( pp ) )
                {
                    gsl_matrix_memcpy ( estimates->Pr, em_wspace->estimates_prev->Pr );
                    i = estimates->M;
                    px = 1;
                    break;
                }
            }
        }

        //VAR coefficients
        if ( coeff_penalty != NONE && coeff_lambda > 0.0 )
            estimate_sparse_coefficients ( em_wspace, regu_wspace, Y, X, estimates, coeff_penalty, coeff_lambda, coeff_lambda_weights, max_iters, tol, L_c, groups );
        else if ( estimates->switching_coeff > 0 )
            estimate_switching_coefficients ( em_wspace, Y, X, estimates );
        else
            estimate_coefficients ( em_wspace, Y, X, y, estimates );

        //VAR covariance
        if ( covar_penalty != NONE && covar_lambda > 0.0 )
            estimate_regularized_covariance_precision ( em_wspace, regu_cov_wspace, Y, X, estimates, covar_penalty, covar_lambda, covar_lambda_weights, max_iters, tol );
        else if ( estimates->switching_covar > 0 )
            estimate_switching_covariance_precision ( em_wspace, Y, X, estimates, -1, 1 );
        else
            estimate_covariance_precision ( em_wspace, Y, X, estimates );

        // (Here it allows one last estimation of coefficients and covariance, above)
        if ( px > 0 )
        {
            //printf ( "pr val check\n" );
            //printf_matrix("Pr", estimates->Pr);
            break;
        }
    }
    /*if ( iters == em_max_iters )
        printf ( " EM Iters: %lu (max), PenLoglik: %lf\n", iters, estimates->loglik - estimates->pen );
    else
        printf ( " EM Iters: %lu, PenLoglik: %lf\n", iters, estimates->loglik - estimates->pen );
    */
#ifdef PARVAR
    if ( coeff_penalty != NONE && coeff_lambda > 0.0 )
    {
        for ( size_t k = 0; k < estimates->A->size1; ++k )
            munmap ( regu_wspace->A_mm[k], estimates->A->size2 * sizeof ( double ) );
        munmap ( regu_wspace->A_mm, estimates->A->size1 * sizeof ( double * ) );
    }
    if ( covar_penalty != NONE && covar_lambda > 0.0 )
    {
        for ( size_t k = 0; k < estimates->M * estimates->D; ++k )
        {
            munmap ( regu_cov_wspace->S_mm[k], estimates->D * sizeof ( double ) );
            munmap ( regu_cov_wspace->S_inv_mm[k], estimates->D * sizeof ( double ) );
        }
        munmap ( regu_cov_wspace->S_mm, estimates->M * estimates->D * sizeof ( double * ) );
        munmap ( regu_cov_wspace->S_inv_mm, estimates->M * estimates->D * sizeof ( double * ) );
    }
#endif
    return GSL_SUCCESS;
}

int compute_ar_sum ( size_t m, size_t t, const msvar_parameters * parameters, const gsl_matrix * X, const gsl_matrix * Y, gsl_vector * Y_t )
{
    //This determines index to get sub-matrix of coefficients ( zero if no switching; m if so )
    m = m * parameters->switching_coeff;
    //Use predictors already stored
    if ( X != NULL )
    {
        gsl_matrix_const_view A_subview = gsl_matrix_const_submatrix ( parameters->A, m * parameters->D, 0, parameters->D, 1 + parameters->p * parameters->D );
        gsl_vector_const_view Y_lags_vecview = gsl_matrix_const_row ( X, t );
        gsl_blas_dgemv ( CblasNoTrans, 1.0, & A_subview.matrix, & Y_lags_vecview.vector, 0.0, Y_t );
    }
    else
    {
        //First col of A is intercept vector
        gsl_vector_const_view intercept_vecview = gsl_matrix_const_subcolumn ( parameters->A, 0, m * parameters->D, parameters->D );
        gsl_blas_dcopy ( & intercept_vecview.vector, Y_t );
        //Iterate min ( p,t ) times ( lag ); Y matrix has T+p rows (data points)
        for ( size_t j = 0; j < GSL_MIN ( parameters->p, t ); ++j )
        {
            gsl_vector_const_view Y_lag_vecview = gsl_matrix_const_row ( Y, t - 1 - j );
            gsl_matrix_const_view Aj_subview = gsl_matrix_const_submatrix ( parameters->A, m * parameters->D, 1 + j * parameters->D, parameters->D, parameters->D );
            gsl_blas_dgemv ( CblasNoTrans, 1.0, & Aj_subview.matrix, & Y_lag_vecview.vector, 1.0, Y_t );
        }
    }
    
    return GSL_SUCCESS;
}

int compute_regression_matrix_vector ( regu_workspace * regu_wspace, const gsl_matrix * Y, const msvar_parameters * estimates, gsl_matrix * X, gsl_vector * y, double * L_c )
{   
    /*  Obtain matrices for generalized regression y = XB ( see page 187-189 )
        We are using T+p data; the observations are T data
    */
    for ( size_t i = 0; i < estimates->T; ++i )
    {
        gsl_vector_view y_sub_vecview = gsl_vector_subvector ( y, i * estimates->D, estimates->D );
        gsl_matrix_get_row ( & y_sub_vecview.vector, Y, i + estimates->p );
    }
    gsl_vector_view X_vecview = gsl_matrix_column ( X, 0 );
    gsl_vector_set_all ( & X_vecview.vector, 1.0 );
    for ( size_t i = 0; i < estimates->p; ++i )
    {
        gsl_matrix_const_view Y_subview = gsl_matrix_const_submatrix ( Y, estimates->p - i - 1, 0, estimates->T, estimates->D );
        gsl_matrix_view X_subview   =    gsl_matrix_submatrix ( X, 0, 1 + i * estimates->D, estimates->T, estimates->D );
        gsl_matrix_memcpy ( & X_subview.matrix, & Y_subview.matrix );
    }
    
    if ( regu_wspace != NULL )
    {
        //The Lipschitz constant (for FISTA) is the highest eigenval;
        //We are bounding eigenvalues of (Em X)'(EmX) by X'X, since Em are between 0 and 1
        // temporary storage for XtX, since the matrix is modified when computing eigens
        gsl_matrix_const_view X_view = gsl_matrix_const_submatrix ( X, 0, 1, estimates->T, estimates->D * estimates->p );
        gsl_blas_dgemm ( CblasTrans, CblasNoTrans, 1.0, & X_view.matrix, & X_view.matrix, 0.0, regu_wspace->E );
        double l1, l2;
        min_max_sym_eigenvalue ( regu_wspace->E, regu_wspace->eval_vec, regu_wspace->eigw, & l1, & l2 );
        ( * L_c) = l2;
    }
    return GSL_SUCCESS;
}

void em_replications ( const msvar_parameters * parameters, msvar_parameters * estimates,
                       gsl_matrix * estimates_matrix_A, gsl_matrix * estimates_matrix_Pr, gsl_matrix * estimates_matrix_Sigma,
                       gsl_matrix * estimates_matrix_loglik, gsl_matrix * estimates_matrix_smooth_p, gsl_matrix * estimates_matrix_lambdas,
                       size_t replications, size_t em_max_iters, size_t max_iters, double tol, const gsl_rng * rng, size_t first_datapoint_t, const char * oracle_dir,
                       size_t coeff_penalty, double coeff_loglambda_step, size_t covar_penalty, double covar_loglambda_step, size_t k_folds, const char * groups_file )
{
    gsl_matrix * coeff_lambda_weights = NULL;
    gsl_matrix * covar_lambda_weights = NULL;
    gsl_matrix * groups = NULL;
    gsl_matrix * A_mle = gsl_matrix_alloc ( estimates->A->size1, estimates->A->size2 );
    gsl_matrix * SL_mle = gsl_matrix_alloc ( estimates->SigmaL->size1, estimates->SigmaL->size2 );
    gsl_matrix * S_inv_mle = gsl_matrix_alloc ( estimates->SigmaL->size1, estimates->SigmaL->size2 );
    gsl_matrix * Pr_mle = gsl_matrix_alloc ( estimates->Pr->size1, estimates->Pr->size2 );
    gsl_matrix * X  = gsl_matrix_alloc ( estimates->T, 1 + estimates->D * estimates->p ); //Design matrix for generalized regression
    gsl_vector * y = gsl_vector_alloc ( estimates->D * estimates->T );
    gsl_matrix * YT = gsl_matrix_alloc ( parameters->T + parameters->p + first_datapoint_t - 1, parameters->D );
    gsl_matrix * Y = gsl_matrix_alloc ( parameters->T + parameters->p, parameters->D );

    em_workspace * em_wspace = NULL;
    regu_workspace * regu_wspace = NULL;
    regu_cov_workspace * regu_cov_wspace = NULL;
    
    em_wspace = ( em_workspace * ) calloc ( 1, sizeof ( em_workspace )  );
    em_workspace_alloc ( em_wspace, estimates );
    if ( coeff_penalty != NONE )
    {
        size_t g_size = 0;
        if ( coeff_penalty == GROUP_ADALASSO )
        {
            read_alloc_groups ( groups_file, & groups, estimates->A->size2 - 1 );
            g_size = groups->size1 / estimates->D;
            coeff_lambda_weights = gsl_matrix_alloc ( estimates->M, g_size );
            printf ( " Groups read from file\n" );
        }
        else
        {
            coeff_lambda_weights = gsl_matrix_alloc ( estimates->A->size1, estimates->A->size2 - 1 );
            if ( coeff_penalty == ORACLE )
            {
                //char file_oracle[ strlen ( oracle_dir ) + 128 ];
                //strcpy ( file_oracle, oracle_dir );
                //strcat ( file_oracle, "msvar_parameters_A_oracle" );
                //fscanf_matrix ( file_oracle, " A oracle ", coeff_lambda_weights );
                gsl_matrix_set_zero ( coeff_lambda_weights );
                for ( size_t i = 0; i < parameters->A->size1; ++i )
                    for ( size_t j = 1; j < parameters->A->size2; ++j )
                        if ( fabs ( gsl_matrix_get ( parameters->A, i, j ) ) > 0.0 )
                            gsl_matrix_set ( coeff_lambda_weights, i, j - 1, 1.0 );

            }
        }
        regu_wspace = ( regu_workspace * ) calloc ( 1, sizeof ( regu_workspace )  );
        regu_workspace_alloc ( regu_wspace, estimates, coeff_penalty, g_size );
    }
    if ( covar_penalty != NONE )
    {
        regu_cov_wspace = ( regu_cov_workspace * ) calloc ( 1, sizeof ( regu_cov_workspace )  );
        regu_cov_workspace_alloc ( regu_cov_wspace, estimates, covar_penalty );
        
        covar_lambda_weights = gsl_matrix_alloc ( estimates->SigmaL->size1, estimates->SigmaL->size2 );
        if ( covar_penalty == ORACLE )
        {
            //char file_oracle[ strlen ( oracle_dir ) + 128 ];
            //strcpy ( file_oracle, oracle_dir );
            //strcat ( file_oracle, "msvar_parameters_Sigma_oracle" );
            //fscanf_matrix ( file_oracle, " S oracle ", covar_lambda_weights );
            gsl_matrix_set_all ( covar_lambda_weights, 1.0 );
            for ( size_t m = 0; m < parameters->M; ++m )
            {
                gsl_matrix_const_view SigmaL_subview = gsl_matrix_const_submatrix ( parameters->SigmaL, m * estimates->D, 0, estimates->D, estimates->D );
                gsl_matrix_view weights_subview = gsl_matrix_submatrix ( covar_lambda_weights, m * estimates->D, 0, estimates->D, estimates->D );
                for ( size_t i = 0; i < parameters->D; ++i )
                    for ( size_t j = 0; j < i; ++j )
                        if ( ! ( fabs ( gsl_matrix_get ( & SigmaL_subview.matrix, i, j ) ) > 0.0 ) )
                        {
                            gsl_matrix_set ( & weights_subview.matrix, i, j, 0.0 );
                            gsl_matrix_set ( & weights_subview.matrix, j, i, 0.0 );
                        }
            }
        }
        else if ( covar_penalty == ORACLE_INVCOVAR || covar_penalty == GRAPHICAL_ORACLE )
        {
            //char file_oracle[ strlen ( oracle_dir ) + 128 ];
            //strcpy ( file_oracle, oracle_dir );
            //strcat ( file_oracle, "msvar_parameters_Sigma_inv_oracle" );
            //fscanf_matrix ( file_oracle, " S inv oracle ", covar_lambda_weights );
            gsl_matrix_set_all ( covar_lambda_weights, 1.0 );
            for ( size_t m = 0; m < parameters->M; ++m )
            {
                gsl_matrix_const_view Sigma_inv_subview = gsl_matrix_const_submatrix ( parameters->Sigma_inv, m * estimates->D, 0, estimates->D, estimates->D );
                gsl_matrix_view weights_subview = gsl_matrix_submatrix ( covar_lambda_weights, m * estimates->D, 0, estimates->D, estimates->D );
                for ( size_t i = 0; i < parameters->D; ++i )
                    for ( size_t j = 0; j < i; ++j )
                        if ( ! ( fabs ( gsl_matrix_get ( & Sigma_inv_subview.matrix, i, j ) ) > 0.0 ) )
                        {
                            gsl_matrix_set ( & weights_subview.matrix, i, j, 0.0 );
                            gsl_matrix_set ( & weights_subview.matrix, j, i, 0.0 );
                        }
            }
        }
    }
    int max_i = 1;
    int max_j = 1;
    double min_loglambda_coeff = MIN_LOGLAMBDA_COEFF;
    double min_loglambda_covar = MIN_LOGLAMBDA_COVAR;
    if ( coeff_penalty == ADALASSO || coeff_penalty == GROUP_ADALASSO )
        min_loglambda_coeff = MIN_LOGLAMBDA_COEFF_ALASSO;
    if ( covar_penalty == ADALASSO || covar_penalty == ADALASSO_INVCOVAR || covar_penalty == GRAPHICAL_ADALASSO )
        min_loglambda_covar = MIN_LOGLAMBDA_COVAR_ALASSO;
    if ( coeff_penalty != NONE && coeff_penalty != ORACLE )
        max_i = ( int ) ( ( MAX_LOGLAMBDA - min_loglambda_coeff ) / coeff_loglambda_step );
    if ( covar_penalty != NONE && covar_penalty != ORACLE )
        max_j = ( int ) ( ( MAX_LOGLAMBDA - min_loglambda_covar ) / covar_loglambda_step );
    double L_c = 0;
    
    double ** ic_array;
    ic_array = ( double ** ) calloc ( max_i, sizeof ( double * ) );
    for ( size_t i = 0; i < max_i; ++i )
    {
        ic_array[ i ] = ( double * ) calloc ( max_j, sizeof ( double ) );
        for ( int j = 0; j < max_j; ++j )
            ic_array[ i ][ j ] = GSL_NEGINF;
    }

    size_t num_p = 1;
    size_t pn = 0;

    double * estimates_mem_loglik;
    double ** estimates_mem_smooth_p;
    double ** estimates_mem_A;
    double ** estimates_mem_Pr;
    double ** estimates_mem_Sigma;
    int * coeff_lambda_ind_pn;
    int * covar_lambda_ind_pn;
    double * coeff_lambda_ic_pn;
    double * covar_lambda_ic_pn;
#ifdef PARVAR
    estimates_mem_loglik = ( double * ) mmap ( NULL, replications * sizeof ( double ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
    estimates_mem_smooth_p = ( double ** ) mmap ( NULL, replications * sizeof ( double * ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
    estimates_mem_A = ( double ** ) mmap ( NULL, replications * sizeof ( double * ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
    estimates_mem_Pr = ( double ** ) mmap ( NULL, replications * sizeof ( double * ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
    estimates_mem_Sigma = ( double ** ) mmap ( NULL, replications * sizeof ( double * ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
    coeff_lambda_ind_pn = ( int * ) mmap ( NULL, replications * sizeof ( int ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
    covar_lambda_ind_pn = ( int * ) mmap ( NULL, replications * sizeof ( int ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
    coeff_lambda_ic_pn = ( double * ) mmap ( NULL, replications * sizeof ( double ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
    covar_lambda_ic_pn = ( double * ) mmap ( NULL, replications * sizeof ( double ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );

    for ( size_t i = 0; i < replications; ++i )
    {
        coeff_lambda_ind_pn[ i ] = 0;
        covar_lambda_ind_pn[ i ] = 0;
        coeff_lambda_ic_pn[ i ] = 0;
        covar_lambda_ic_pn[ i ] = 0;
        //Size is columns of output matrices; total number of elements of estimator
        estimates_mem_smooth_p[i] = ( double * ) mmap ( NULL, estimates_matrix_smooth_p->size2 * sizeof ( double ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
        estimates_mem_A[i] = ( double * ) mmap ( NULL, estimates_matrix_A->size2 * sizeof ( double ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
        estimates_mem_Pr[i] = ( double * ) mmap ( NULL, estimates_matrix_Pr->size2 * sizeof ( double ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
        estimates_mem_Sigma[i] = ( double * ) mmap ( NULL, estimates_matrix_Sigma->size2 * sizeof ( double ), PROT_READ | PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0 );
        for ( size_t j = 0; j < estimates_matrix_smooth_p->size2; ++j )
            estimates_mem_smooth_p[i][j] = 0.0;
        for ( size_t j = 0; j < estimates_matrix_A->size2; ++j )
            estimates_mem_A[i][j] = 0.0;
        for ( size_t j = 0; j < estimates_matrix_Pr->size2; ++j )
            estimates_mem_Pr[i][j] = 0.0;
        for ( size_t j = 0; j < estimates_matrix_Sigma->size2; ++j )
            estimates_mem_Sigma[i][j] = 0.0;
        estimates_mem_loglik [ i ] = GSL_NEGINF;
    }
#else
    estimates_mem_loglik = ( double * ) calloc ( replications, sizeof ( double )  );
    estimates_mem_smooth_p = ( double ** ) calloc ( replications, sizeof ( double * )  );
    estimates_mem_A = ( double ** ) calloc ( replications, sizeof ( double * )  );
    estimates_mem_Pr = ( double ** ) calloc ( replications, sizeof ( double * )  );
    estimates_mem_Sigma = ( double ** ) calloc ( replications, sizeof ( double * )  );
    coeff_lambda_ind_pn = ( int * ) calloc ( replications, sizeof ( int )  );
    covar_lambda_ind_pn = ( int * ) calloc ( replications, sizeof ( int )  );
    coeff_lambda_ic_pn = ( double * ) calloc ( replications, sizeof ( double )  );
    covar_lambda_ic_pn = ( double * ) calloc ( replications, sizeof ( double )  );
    for ( size_t i = 0; i < replications; ++i )
    {
        coeff_lambda_ind_pn[ i ] = 0;
        covar_lambda_ind_pn[ i ] = 0;
        coeff_lambda_ic_pn[ i ] = 0;
        covar_lambda_ic_pn[ i ] = 0;
        //Size is columns of output matrices; total number of elements of estimator
        estimates_mem_smooth_p[i] = ( double * ) calloc ( estimates_matrix_smooth_p->size2, sizeof ( double )  );
        estimates_mem_A[i] = ( double * ) calloc ( estimates_matrix_A->size2, sizeof ( double )  );
        estimates_mem_Pr[i] = ( double * ) calloc ( estimates_matrix_Pr->size2, sizeof ( double )  );
        estimates_mem_Sigma[i] = ( double * ) calloc ( estimates_matrix_Sigma->size2, sizeof ( double )  );
        for ( size_t j = 0; j < estimates_matrix_smooth_p->size2; ++j )
            estimates_mem_smooth_p[i][j] = 0.0;
        for ( size_t j = 0; j < estimates_matrix_A->size2; ++j )
            estimates_mem_A[i][j] = 0.0;
        for ( size_t j = 0; j < estimates_matrix_Pr->size2; ++j )
            estimates_mem_Pr[i][j] = 0.0;
        for ( size_t j = 0; j < estimates_matrix_Sigma->size2; ++j )
            estimates_mem_Sigma[i][j] = 0.0;
        estimates_mem_loglik [ i ] = GSL_NEGINF;
    }
#endif


#ifdef PARVAR
    //Threads
    num_p = GSL_MAX ( 1, GSL_MIN ( replications, NUM_THREADS ) );
    pid_t pid_arr [ num_p ];
    pid_t ppid = getpid ();
    for ( size_t i = 0; i < num_p - 1; ++i )
    {
        pid_t p = fork();
        if ( p > 0 ) //caller
        {
            pid_arr [ pn ] = p;
            ++pn;
            continue;
        }
        else if ( ! p ) // child
        {
            break;
        }
        else // error
        {
            perror ( "\n fork call failed " );
            exit ( 1 );
        }
    }
    if ( getpid () == ppid )
        pn = 0;
    else
        ++pn;
#endif

    // Replications
    for ( size_t i = 0; i < replications; ++i )
    {
        generate_data ( parameters, Y, YT, em_wspace->mean_vec, rng, NULL );
        
        //Each thread process does a specific range of replicas
        if ( num_p > 1 && ( floor ( ( 1.0 * pn / num_p ) * replications ) > i || i >= floor ( ( 1.0 * ( pn + 1 ) / num_p ) * replications ) ) )
            continue;
        
        compute_regression_matrix_vector ( regu_wspace, Y, estimates, X, y, & L_c );

        double c1 = gsl_matrix_get ( estimates_matrix_lambdas, i, 0 );
        double c2 = gsl_matrix_get ( estimates_matrix_lambdas, i, 1 );
        if ( c1 < 0.0 || c2 < 0.0 )
        {
            coeff_lambda_ic_pn[i] = 0.0;
            covar_lambda_ic_pn[i] = 0.0;
            continue;
        }

        if ( num_p > 1 )
            printf ( " Replication %lu by process %lu\n", i + 1, pn );
        else
            printf ( " Replication %lu\n", i + 1 );

        if ( c1 > 0.0 || c2 > 0.0 )
        {
            coeff_lambda_ic_pn[i] = fabs ( c1 );
            covar_lambda_ic_pn[i] = fabs ( c2 );
        }
        else if ( ( coeff_penalty != NONE && coeff_penalty != ORACLE ) || ( covar_penalty != NONE && covar_penalty != ORACLE && covar_penalty != ORACLE_INVCOVAR && covar_penalty != GRAPHICAL_ORACLE ) )
        {
            double llmax = GSL_NEGINF;
            for ( int in = 0; in < 2; ++in )
            {
                generate_initial_transition_probabilities ( estimates, NULL );
                compute_initial_estimates ( em_wspace, Y, X, estimates, in );
                expectation_maximization ( em_wspace, NULL, NULL, estimates, Y, X, y, em_max_iters, max_iters, tol, NONE, NONE, 0.0, 0.0, NULL, NULL, 0.0, NULL );
                if ( estimates->loglik > llmax && ! gsl_isnan ( estimates->loglik ) )
                {
                    gsl_matrix_memcpy ( A_mle, estimates->A );
                    gsl_matrix_memcpy ( Pr_mle, estimates->Pr );
                    gsl_matrix_memcpy ( SL_mle, estimates->SigmaL );
                    gsl_matrix_memcpy ( S_inv_mle, estimates->Sigma_inv );
                    llmax = estimates->loglik;
                }
            }
            gsl_matrix_memcpy ( estimates->A, A_mle );
            gsl_matrix_memcpy ( estimates->Pr, Pr_mle );
            gsl_matrix_memcpy ( estimates->SigmaL, SL_mle );
            gsl_matrix_memcpy ( estimates->Sigma_inv, S_inv_mle );
            set_penalty_weights ( estimates->D, estimates->A, estimates->SigmaL, estimates->Sigma_inv, coeff_penalty, covar_penalty, coeff_lambda_weights, covar_lambda_weights, groups );
            



            size_t pnC = ICRIT;
 
            
            
            size_t ir = i;
            coeff_lambda_ind_pn[ ir ] = 0;
            covar_lambda_ind_pn[ ir ] = 0;
            
            printf(" Performing tuning parameter selection...\n");

            double resol_step = 4.0;
            compute_information_descent_optimum ( em_wspace, regu_wspace, regu_cov_wspace, estimates, A_mle, SL_mle, S_inv_mle, Pr_mle, Y, NULL, X, y, ic_array, pnC,
                coeff_penalty, coeff_loglambda_step, covar_penalty, covar_loglambda_step, coeff_lambda_weights, covar_lambda_weights,
                em_max_iters, max_iters, tol, & ( coeff_lambda_ind_pn[ ir ] ), & ( covar_lambda_ind_pn[ ir ] ), 0, groups, resol_step, L_c );
            resol_step = 2.0;
            //coeff_lambda_ind_pn[ ir ] = ( resol_step * ( coeff_lambda_ind_pn[ ir ] - 1 ) + 1 ) * ( coeff_penalty != NONE );
            //covar_lambda_ind_pn[ ir ] = ( resol_step * ( covar_lambda_ind_pn[ ir ] - 1 ) + 1 ) * ( covar_penalty != NONE );
            compute_information_descent_optimum ( em_wspace, regu_wspace, regu_cov_wspace, estimates, A_mle, SL_mle, S_inv_mle, Pr_mle, Y, NULL, X, y, ic_array, pnC,
                coeff_penalty, coeff_loglambda_step, covar_penalty, covar_loglambda_step, coeff_lambda_weights, covar_lambda_weights,
                em_max_iters, max_iters, tol, & ( coeff_lambda_ind_pn[ ir ] ), & ( covar_lambda_ind_pn[ ir ] ), SUBGRID_WIDTH, groups, resol_step, L_c );
            //coeff_lambda_ind_pn[ ir ] = ( resol_step * ( coeff_lambda_ind_pn[ ir ] - 1 ) + 1 ) * ( coeff_penalty != NONE );
            //covar_lambda_ind_pn[ ir ] = ( resol_step * ( covar_lambda_ind_pn[ ir ] - 1 ) + 1 ) * ( covar_penalty != NONE );
            resol_step = 1.0;
            compute_information_descent_optimum ( em_wspace, regu_wspace, regu_cov_wspace, estimates, A_mle, SL_mle, S_inv_mle, Pr_mle, Y, NULL, X, y, ic_array, pnC,
                coeff_penalty, coeff_loglambda_step, covar_penalty, covar_loglambda_step, coeff_lambda_weights, covar_lambda_weights,
                em_max_iters, max_iters, tol, & ( coeff_lambda_ind_pn[ ir ] ), & ( covar_lambda_ind_pn[ ir ] ), SUBGRID_WIDTH, groups, resol_step, L_c );
            
            
            coeff_lambda_ic_pn[ir] = ( coeff_lambda_ind_pn[ir] > 0 ) * exp ( min_loglambda_coeff + ( coeff_lambda_ind_pn[ir] - 1 ) * coeff_loglambda_step );
            covar_lambda_ic_pn[ir] = ( covar_lambda_ind_pn[ir] > 0 ) * exp ( min_loglambda_covar + ( covar_lambda_ind_pn[ir] - 1 ) * covar_loglambda_step );
            //printf( " Replication %lu lambdas: (%d) %lf, (%d) %lf\n",
            //        i + 1, coeff_lambda_ind_pn[ir], coeff_lambda_ic_pn[ir], covar_lambda_ind_pn[ir], covar_lambda_ic_pn[ir] );
        }
        //Proceed
        
        if ( coeff_penalty == ORACLE && ( covar_penalty == ORACLE || covar_penalty == ORACLE_INVCOVAR || covar_penalty == GRAPHICAL_ORACLE ) )
        {
            generate_initial_transition_probabilities ( estimates, NULL );
            compute_initial_estimates ( em_wspace, Y, X, estimates, 0 );
            gsl_matrix_set_all ( em_wspace -> xi_smooth, 1.0 / estimates->M );
            gsl_matrix_memcpy ( em_wspace->estimates_prev->SigmaL, estimates->SigmaL );
            gsl_matrix_memcpy ( em_wspace->estimates_prev->Sigma_inv, estimates->Sigma_inv );
            estimate_sparse_coefficients ( em_wspace, regu_wspace, Y, X, estimates, coeff_penalty, 1.0, coeff_lambda_weights, max_iters, tol, L_c, groups );
            estimate_regularized_covariance_precision ( em_wspace, regu_cov_wspace, Y, X, estimates, covar_penalty, 1.0, covar_lambda_weights, max_iters, tol );
            gsl_matrix_memcpy ( A_mle, estimates->A );
            gsl_matrix_memcpy ( Pr_mle, estimates->Pr );
            gsl_matrix_memcpy ( SL_mle, estimates->SigmaL );
            gsl_matrix_memcpy ( S_inv_mle, estimates->Sigma_inv );
        }
        else
        {
            double llmax = GSL_NEGINF;
            for ( int in = 0; in < 2; ++in )
            {
                generate_initial_transition_probabilities ( estimates, NULL );
                compute_initial_estimates ( em_wspace, Y, X, estimates, in );
                expectation_maximization ( em_wspace, NULL, NULL, estimates, Y, X, y, em_max_iters, max_iters, tol, NONE, NONE, 0.0, 0.0, NULL, NULL, 0.0, NULL );
                if ( estimates->loglik > llmax && ! gsl_isnan ( estimates->loglik ) )
                {
                    gsl_matrix_memcpy ( A_mle, estimates->A );
                    gsl_matrix_memcpy ( Pr_mle, estimates->Pr );
                    gsl_matrix_memcpy ( SL_mle, estimates->SigmaL );
                    gsl_matrix_memcpy ( S_inv_mle, estimates->Sigma_inv );
                    llmax = estimates->loglik;
                }
            }
            gsl_matrix_memcpy ( estimates->A, A_mle );
            gsl_matrix_memcpy ( estimates->Pr, Pr_mle );
            gsl_matrix_memcpy ( estimates->SigmaL, SL_mle );
            gsl_matrix_memcpy ( estimates->Sigma_inv, S_inv_mle );
            set_penalty_weights ( estimates->D, estimates->A, estimates->SigmaL, estimates->Sigma_inv, coeff_penalty, covar_penalty, coeff_lambda_weights, covar_lambda_weights, groups );
        }
        
        if ( coeff_penalty == NONE && covar_penalty == NONE )
        {
            set_mem_matrices ( i, 0, estimates, estimates_mem_A, estimates_mem_Pr, estimates_mem_Sigma, 0 );
            for ( size_t j = 0; j < estimates->M; ++j )
            {
                gsl_vector_const_view p_vec = gsl_matrix_const_row ( em_wspace->xi_smooth, j );
                estimates_mem_smooth_p [ i ][ j ] = gsl_blas_dasum ( & p_vec.vector ) / estimates->T;
            }
            estimates_mem_loglik [ i ] = estimates->loglik;
            continue;
        }
        
        // Estimate with penalty
        if ( coeff_penalty == ORACLE )
        {
            coeff_lambda_ic_pn[ i ] = 1.0;
        }
        if ( covar_penalty == ORACLE || covar_penalty == ORACLE_INVCOVAR || covar_penalty == GRAPHICAL_ORACLE )
        {
            covar_lambda_ic_pn[ i ] = 1.0;
        }
        
        
        

        //expectation_maximization ( em_wspace, NULL, NULL, estimates, Y, X, y, em_max_iters, max_iters, tol, NONE, NONE, 0.0, 0.0, NULL, NULL, 0.0, NULL );
        expectation_maximization ( em_wspace, regu_wspace, regu_cov_wspace, estimates, Y, X, y, em_max_iters, max_iters, tol, coeff_penalty, covar_penalty,
                                coeff_lambda_ic_pn[ i ], covar_lambda_ic_pn[ i ], coeff_lambda_weights, covar_lambda_weights, L_c, groups );
        int is_precision = ( covar_penalty == ORACLE_INVCOVAR || covar_penalty == GRAPHICAL_ORACLE || INVCOVAR_REG ( covar_penalty ) );
        set_mem_matrices ( i, 0, estimates, estimates_mem_A, estimates_mem_Pr, estimates_mem_Sigma, is_precision );
        for ( size_t j = 0; j < estimates->M; ++j )
        {
            gsl_vector_const_view p_vec = gsl_matrix_const_row ( em_wspace->xi_smooth, j );
            estimates_mem_smooth_p [ i ][ j ] = gsl_blas_dasum ( & p_vec.vector ) / estimates->T;
        }
        estimates_mem_loglik [ i ] = estimates->loglik;
        
        printf ( " MSVAR fitted\n" );
        
    }//replicas end
#ifdef PARVAR
    if ( getpid () != ppid )
        exit ( 0 );
    //if ( getpid () == ppid )
    for ( size_t i = 0; i < num_p - 1; ++i )
    {
        int pstatus = waitpid ( pid_arr [ i ], NULL, 0 );
        if ( pstatus == -1 && errno != ECHILD )
        {
            perror ( "waitpid() error" );
            exit ( 1 );
        }
    }
#endif
    
    for ( size_t i = 0; i < replications; ++i )
    {
        size_t ir = i;
        gsl_matrix_set ( estimates_matrix_lambdas, i, 0, coeff_lambda_ic_pn[ir] + ( coeff_penalty == ORACLE ) );
        gsl_matrix_set ( estimates_matrix_lambdas, i, 1, covar_lambda_ic_pn[ir] + ( covar_penalty == ORACLE || ( covar_penalty == ORACLE_INVCOVAR ) ) );
    }
    
    for ( size_t i = 0; i < replications; ++i )
    {
        for ( size_t j = 0; j < estimates_matrix_smooth_p->size2; ++j )
        {
            gsl_matrix_set ( estimates_matrix_smooth_p, i , j, estimates_mem_smooth_p[ i ][j] );
        }
        for ( size_t j = 0; j < estimates_matrix_A->size2; ++j )
        {
            gsl_matrix_set ( estimates_matrix_A, i , j, estimates_mem_A[ i ][j] );
        }
        for ( size_t j = 0; j < estimates_matrix_Pr->size2; ++j )
        {
            gsl_matrix_set ( estimates_matrix_Pr, i , j, estimates_mem_Pr[ i ][j] );
        }
        for ( size_t j = 0; j < estimates_matrix_Sigma->size2; ++j )
        {
            gsl_matrix_set ( estimates_matrix_Sigma, i , j, estimates_mem_Sigma[ i ][j] );
        }
    }
    
    for ( size_t i = 0; i < replications; ++i )
        gsl_matrix_set ( estimates_matrix_loglik, i, 0, estimates_mem_loglik[ i ] );

#ifdef PARVAR
    for ( size_t i = 0; i < replications; ++i )
    {
        munmap ( estimates_mem_smooth_p[i], estimates_matrix_smooth_p->size2 * sizeof ( double ) );
        munmap ( estimates_mem_A[i], estimates_matrix_A->size2 * sizeof ( double ) );
        munmap ( estimates_mem_Pr[i], estimates_matrix_Pr->size2 * sizeof ( double ) );
        munmap ( estimates_mem_Sigma[i], estimates_matrix_Sigma->size2 * sizeof ( double ) );
    }
    munmap ( estimates_mem_smooth_p, replications * sizeof ( double * ) );
    munmap ( estimates_mem_A, replications * sizeof ( double * ) );
    munmap ( estimates_mem_Pr, replications * sizeof ( double * ) );
    munmap ( estimates_mem_Sigma, replications * sizeof ( double * ) );
    munmap ( estimates_mem_loglik, replications * sizeof ( double ) );
    munmap ( coeff_lambda_ind_pn, replications * sizeof ( int ) );
    munmap ( covar_lambda_ind_pn, replications * sizeof ( int ) );
    munmap ( coeff_lambda_ic_pn, replications * sizeof ( double ) );
    munmap ( covar_lambda_ic_pn, replications * sizeof ( double ) );
#else
    for ( size_t i = 0; i < replications; ++i )
    {
        free ( estimates_mem_smooth_p[i] );
        free ( estimates_mem_A[i] );
        free ( estimates_mem_Pr[i] );
        free ( estimates_mem_Sigma[i] );
    }
    free ( estimates_mem_smooth_p );
    free ( estimates_mem_A );
    free ( estimates_mem_Pr );
    free ( estimates_mem_Sigma );
    free ( estimates_mem_loglik );
    free ( coeff_lambda_ind_pn );
    free ( covar_lambda_ind_pn );
    free ( coeff_lambda_ic_pn );
    free ( covar_lambda_ic_pn );
#endif
    
    for ( size_t i = 0; i < max_i; ++i )
        free ( ic_array[i] );
    free ( ic_array );
    
    em_workspace_free ( em_wspace, estimates->M );
    free ( em_wspace );
    if ( coeff_penalty != NONE )
    {
        gsl_matrix_free ( coeff_lambda_weights );
        regu_workspace_free ( regu_wspace, estimates );
        free ( regu_wspace );
        if ( coeff_penalty == GROUP_ADALASSO )
            gsl_matrix_free ( groups );
    }
    if ( covar_penalty != NONE )
    {
        regu_cov_workspace_free ( regu_cov_wspace, covar_penalty, estimates );
        free (regu_cov_wspace);
        gsl_matrix_free ( covar_lambda_weights );
    }
    
    gsl_matrix_free ( A_mle );
    gsl_matrix_free ( SL_mle );
    gsl_matrix_free ( S_inv_mle );
    gsl_matrix_free ( Pr_mle );
    gsl_matrix_free ( Y );
    gsl_matrix_free ( YT );
    gsl_matrix_free ( X );
    gsl_vector_free ( y );
    return;
}

void set_mem_matrices ( size_t i, int ic_c, const msvar_parameters * estimates,
                        double ** estimates_mem_A, double ** estimates_mem_Pr, double ** estimates_mem_Sigma, int is_precision )
{
    size_t shift = 1;
    size_t row = ic_c;
    
    size_t Nj = estimates->M;
    if ( estimates->switching_coeff == 0 )
        Nj = 1;
    for ( size_t jm = 0; jm < Nj; ++jm )
    {
        size_t jj = jm * estimates->D * estimates->A->size2;
        for ( size_t j = 0; j < estimates->D; ++j )
        {
            for ( size_t jk = 0; jk < estimates->A->size2; ++jk )
                //Concatenate rows of A
                estimates_mem_A [ i * shift + row ][ jj + j * estimates->A->size2 + jk ] = gsl_matrix_get ( estimates->A, jm * estimates->D + j, jk );
        }
    }
    for ( size_t jr = 0; jr < estimates->Pr->size1; ++jr )
        for ( size_t jc = 0; jc < estimates->Pr->size2; ++jc )
        {
            //Concatenate rows of Pr
            estimates_mem_Pr[ i * shift + row ][ jr * estimates->Pr->size1 + jc ] = gsl_matrix_get ( estimates->Pr, jr, jc );
        }
    Nj = estimates->M;
    if ( estimates->switching_covar == 0 )
        Nj = 1;
    for ( size_t j = 0; j < Nj; ++j )
    {
        if ( is_precision )
        {
            gsl_matrix_const_view Sigma_inv_subview = gsl_matrix_const_submatrix ( estimates->Sigma_inv, j * estimates->D, 0, estimates->D, estimates->D );
            for ( size_t ir = 0; ir < estimates->D; ++ir )
            {
                size_t jj, jj2;
                jj = j * estimates->D * estimates->D + ir * estimates->D + ir;
                estimates_mem_Sigma[ i * shift + row ][ jj ] = gsl_matrix_get ( & Sigma_inv_subview.matrix, ir, ir );
                for ( size_t ic = 0; ic < ir; ++ic )
                {
                    double val = gsl_matrix_get ( & Sigma_inv_subview.matrix, ic, ir );
                    jj = j * estimates->D * estimates->D + ir * estimates->D + ic;
                    jj2 = j * estimates->D * estimates->D + ic * estimates->D + ir;
                    estimates_mem_Sigma[ i * shift + row ][ jj ] = val;
                    estimates_mem_Sigma[ i * shift + row ][ jj2 ] = val;
                }
            }
        }
        else
        {
            gsl_matrix_const_view SigmaL_subview = gsl_matrix_const_submatrix ( estimates->SigmaL, j * estimates->D, 0, estimates->D, estimates->D );
            for ( size_t ir = 0; ir < estimates->D; ++ir )
            {
                double val;
                size_t jj, jj2;
                gsl_vector_const_view SigmaL_vecview = gsl_matrix_const_subrow ( & SigmaL_subview.matrix, ir, 0, ir + 1 );
                gsl_blas_ddot ( & SigmaL_vecview.vector, & SigmaL_vecview.vector, & val );
                jj = j * estimates->D * estimates->D + ir * estimates->D + ir;
                estimates_mem_Sigma[ i * shift + row ][ jj ] = val;
                for ( size_t ic = 0; ic < ir; ++ic )
                {
                    val = gsl_matrix_get ( & SigmaL_subview.matrix, ic, ir );
                    jj = j * estimates->D * estimates->D + ir * estimates->D + ic;
                    jj2 = j * estimates->D * estimates->D + ic * estimates->D + ir;
                    estimates_mem_Sigma[ i * shift + row ][ jj ] = val;
                    estimates_mem_Sigma[ i * shift + row ][ jj2 ] = val;
                }
            }
        }
        
    }
    return;
}

void compute_df ( const msvar_parameters * estimates, int * params_A, int * params_S, int * params_A_min, int * params_S_min, gsl_vector * df_m,
                  size_t coeff_penalty, size_t covar_penalty, double tol, const gsl_matrix * coeff_lambda_weights, const gsl_matrix * groups )
{
    //Count number of (nonzero) parameters
    ( * params_A_min ) = estimates->D * ( estimates->D * estimates->p + 1 );
    ( * params_A ) = ( * params_A_min ) * estimates->M;
    ( * params_S_min ) = ( estimates->D * ( estimates->D + 1 ) ) / 2;
    ( * params_S ) = ( * params_S_min ) * estimates->M;
    if ( coeff_penalty != NONE )
    {
        if ( coeff_penalty == GROUP_ADALASSO )
        {
            ( * params_A ) = 0;
            for ( size_t m = 0; m < estimates->M; ++m )
            {
                double pam = 0.0;
                for ( size_t g = 0; g < coeff_lambda_weights->size2; ++g )
                {
                    double s = 0.0;
                    int d = 0;
                    gsl_matrix_const_view group_view = gsl_matrix_const_submatrix ( groups, g * estimates->D, 0, estimates->D, estimates->A->size2 - 1 );
                    for ( size_t i = 0; i < ( & group_view.matrix )->size1; ++i )
                    {
                        for ( size_t j = 0; j < ( & group_view.matrix )->size2; ++j )
                        {
                            if ( gsl_matrix_get ( & group_view.matrix, i, j ) > 0 )
                            {
                                ++d;
                                double v = gsl_matrix_get ( estimates->A, m * estimates->D + i, j + 1 );
                                s += v * v;
                            }
                        }   
                    }
                    double t = sqrt ( s );
                    if ( t >= tol )
                    pam += 1 + t * gsl_matrix_get ( coeff_lambda_weights, m, g ) * ( d - 1.0 );
                }
                gsl_vector_set ( df_m, m, pam );
                ( * params_A ) += pam;
                if ( pam < ( * params_A_min ) )
                    ( * params_A_min ) = pam;
            }
        }
        else
        {
            ( * params_A ) = 0;
            for ( size_t m = 0; m < estimates->M; ++m )
            {
                int pam = 0;
                for ( size_t r = 0; r < estimates->D; ++r )
                    for ( size_t c = 0; c < estimates->A->size2; ++c )
                    {
                        int e = (int) ( fabs ( gsl_matrix_get ( estimates->A, m * estimates->D + r, c ) ) >= tol );
                        pam += e;
                    }
                gsl_vector_set ( df_m, m, pam );
                ( * params_A ) += pam;
                if ( pam < ( * params_A_min ) )
                    ( * params_A_min ) = pam;
            }
        }
    }
    if ( covar_penalty != NONE )
    {
        ( * params_S ) = 0;
        for ( size_t m = 0; m < estimates->M; ++m )
        {
            const gsl_matrix * S_p;
            gsl_matrix_const_view S0_view = gsl_matrix_const_submatrix ( estimates->SigmaL, m * estimates->D, 0, estimates->D, estimates->D );
            gsl_matrix_const_view S0_inv_view = gsl_matrix_const_submatrix ( estimates->Sigma_inv, m * estimates->D, 0, estimates->D, estimates->D );
            if ( INVCOVAR_REG ( covar_penalty ) )
                S_p = & S0_inv_view.matrix;
            else
                S_p = & S0_view.matrix;
            
            int psm = 0;
            for ( size_t r = 0; r < estimates->D; ++r )
            {
                //Access covariance on the upper triangle of Cholesky matrix
                for ( size_t c = 0; c < r; ++c )
                    psm += (int) ( fabs ( gsl_matrix_get ( S_p, c, r ) ) >= tol );
            }
            psm += estimates->D;
            gsl_vector_set ( df_m, m, psm + gsl_vector_get ( df_m, m ) );
            ( * params_S ) += psm;
            if ( psm < ( * params_S_min ) )
                ( * params_S_min ) = psm;
        }
    }
    return;
}

double compute_ic ( em_workspace * em_wspace, const msvar_parameters * estimates, const gsl_matrix * Y, size_t ic, size_t params_A, size_t params_S,
                    gsl_vector * T_m, const gsl_vector * df_m )
{
    if ( ic == AIC )
        return 2.0 * estimates->loglik - 2.0 * ( params_A + params_S + estimates->M * ( estimates->M - 1 ) );
    else if ( ic == BIC )
        return 2.0 * estimates->loglik - log ( estimates->T ) * ( params_A + params_S + estimates->M * ( estimates->M - 1 ) );
    else if ( ic == MDL)
    {
        gsl_vector_set_zero ( T_m );
        //Regime marginal log sample sizes
        for ( size_t m = 0; m < estimates->M; ++m )
        {
            gsl_vector_const_view p_vec = gsl_matrix_const_row ( em_wspace->xi_smooth, m);
            gsl_vector_set ( T_m, m, log ( gsl_blas_dasum ( & p_vec.vector ) ) );
        }
        double dd;
        gsl_blas_ddot ( T_m, df_m, & dd );
        return 2.0 * estimates->loglik - log ( estimates->T ) * estimates->M * ( estimates->M - 1 ) - dd;
    }
    else // if ( ic_c == LIC )
        return log_likelihood ( em_wspace, Y, estimates, 0, NULL );
}

void compute_information_descent_optimum ( em_workspace * em_wspace, regu_workspace * regu_wspace, regu_cov_workspace * regu_cov_wspace, msvar_parameters * estimates,
                                const gsl_matrix * A_mle, const gsl_matrix * SL_mle, const gsl_matrix * S_inv_mle, const gsl_matrix * Pr_mle,
                                const gsl_matrix * Y_tr, const gsl_matrix * Y_vd, gsl_matrix * X, gsl_vector * y, double ** ic_arr, size_t ic_c,
                                size_t coeff_penalty, double coeff_loglambda_step, size_t covar_penalty, double covar_loglambda_step,
                                gsl_matrix * coeff_lambda_weights, gsl_matrix * covar_lambda_weights, size_t em_max_iters, size_t max_iters,
                                double tol, int * coeff_lambda_ind, int * covar_lambda_ind, double grid_width, const gsl_matrix * groups, double resol_, double L_c )
{
    int min_i = 0;
    int min_j = 0;
    int max_i = 0;
    int max_j = 0;
    int resol = ( int ) resol_;

    double min_loglambda_coeff = MIN_LOGLAMBDA_COEFF;
    double min_loglambda_covar = MIN_LOGLAMBDA_COVAR;
    if ( coeff_penalty == ADALASSO || coeff_penalty == GROUP_ADALASSO )
        min_loglambda_coeff = MIN_LOGLAMBDA_COEFF_ALASSO;
    if ( covar_penalty == ADALASSO || covar_penalty == ADALASSO_INVCOVAR || covar_penalty == GRAPHICAL_ADALASSO )
        min_loglambda_covar = MIN_LOGLAMBDA_COVAR_ALASSO;

    if ( coeff_penalty != NONE )
    {
        if ( ( * coeff_lambda_ind ) > 0 && grid_width > 1.0 )
        {
            min_i = ( int ) ( ( * coeff_lambda_ind ) - resol * ( grid_width - 1.0 ) / 2.0 );
            max_i = ( int ) ( ( * coeff_lambda_ind ) + resol * ( grid_width - 1.0 ) / 2.0 ) + resol;
            if ( min_i <= 0 )
            {
                min_i = 1;
                max_i = resol * grid_width + 1;
            }
            * coeff_lambda_ind = min_i;
        }
        else
        {
            min_i = 1;
            max_i = ( int ) ( ( MAX_LOGLAMBDA - min_loglambda_coeff ) / coeff_loglambda_step );
        }
    }
    if ( covar_penalty != NONE )
    {
        if ( ( * covar_lambda_ind ) > 0 && grid_width > 1.0 )
        {
            min_j = ( int ) ( ( * covar_lambda_ind ) - resol * ( grid_width - 1.0 ) / 2.0 );
            max_j = ( int ) ( ( * covar_lambda_ind ) + resol * ( grid_width - 1.0 ) / 2.0 ) + resol;
            if ( min_j <= 0 )
            {
                min_j = 1;
                max_j = resol * grid_width + 1;
            }
            * covar_lambda_ind = min_j;
        }
        else
        {
            min_j = 1;
            max_j = ( int ) ( ( MAX_LOGLAMBDA - min_loglambda_covar ) / covar_loglambda_step );
        }
    }

    gsl_matrix_memcpy ( estimates->A, A_mle );
    gsl_matrix_memcpy ( estimates->Pr, Pr_mle );
    gsl_matrix_memcpy ( estimates->SigmaL, SL_mle );
    gsl_matrix_memcpy ( estimates->Sigma_inv, S_inv_mle );

    for ( int il = 0; il < GSL_MAX (1, max_i ); ++il )
        for ( int jl = 0; jl < GSL_MAX (1, max_j ); ++jl )
            ic_arr[il][jl] = GSL_NEGINF;

    double max_ic = GSL_NEGINF;
    double lambda_i;
    double lambda_j;
    double lambda_i_max = GSL_POSINF;
    double lambda_j_max = GSL_POSINF;
    int max_i_z = max_i;
    int max_j_z = max_j;
    int min_i_z = min_i;
    int min_j_z = min_j;

    //Get lambda aprox limits
    if ( grid_width <= 1.0 )
    {
        //Coeffs
        for ( size_t i = 0; i < estimates->A->size1; ++i )
            for ( size_t j = 0; j < estimates->A->size2 - 1; ++j )
            {
                double a = fabs ( gsl_matrix_get ( estimates->A, i, j + 1 ) );
                lambda_i_max = GSL_MAX ( a, lambda_i_max );
            }
        //Covars
        for ( size_t k = 0; k < estimates->M; ++k )
        {
            const gsl_matrix * S_p;
            gsl_matrix_const_view S_view = gsl_matrix_const_submatrix ( estimates->SigmaL, k * estimates->D, 0, estimates->D, estimates->D );
            gsl_matrix_const_view S_inv_view = gsl_matrix_const_submatrix ( estimates->Sigma_inv, k * estimates->D, 0, estimates->D, estimates->D );
            if ( INVCOVAR_REG ( covar_penalty ) )
                S_p = & S_inv_view.matrix;
            else
                S_p = & S_view.matrix;
            for ( size_t i = 0; i < estimates->D; ++i )
                for ( size_t j = 0; j < i; ++j )
                {
                    double a = fabs ( gsl_matrix_get ( S_p, j, i ) );
                    lambda_j_max = GSL_MAX ( a, lambda_j_max );
                }
        }
    }

    int ip = -1;
    int jp = -1;
    int iq = min_i_z;
    int jq = min_j_z;
    while ( ! ( ip == iq && jp == jq ) )
    {
        jp = jq;
        lambda_i = ( iq > 0 ) * exp ( min_loglambda_coeff + ( iq - 1 ) * coeff_loglambda_step );
        gsl_matrix_memcpy ( estimates->A, A_mle );
        gsl_matrix_memcpy ( estimates->Pr, Pr_mle );
        gsl_matrix_memcpy ( estimates->SigmaL, SL_mle );
        gsl_matrix_memcpy ( estimates->Sigma_inv, S_inv_mle );
        for ( int jl = min_j_z; jl < max_j_z; jl += resol )
        {
            if ( ! gsl_isinf ( ic_arr[ iq ][ jl ] ) )
            {
                if ( ic_arr[iq][jl] > max_ic )
                {
                    max_ic = ic_arr[iq][jl];
                    ( * covar_lambda_ind ) = jq = jl;
                }
                continue;
            }
            lambda_j = ( jl > 0 ) * exp ( min_loglambda_covar + ( jl - 1 ) * covar_loglambda_step );
            //printf (" i(%d): %lf, j(%d): %lf\n", iq, lambda_i, jl, lambda_j );
            //gsl_matrix_memcpy ( estimates->A, A_mle );
            //gsl_matrix_memcpy ( estimates->Pr, Pr_mle );
            //gsl_matrix_memcpy ( estimates->SigmaL, SL_mle );
            //gsl_matrix_memcpy ( estimates->Sigma_inv, S_inv_mle );
            //Regularized
            expectation_maximization ( em_wspace, regu_wspace, regu_cov_wspace, estimates, Y_tr, X, y, em_max_iters, max_iters, tol, coeff_penalty, covar_penalty,
                                    lambda_i, lambda_j, coeff_lambda_weights, covar_lambda_weights, L_c, groups );
            int params_A, params_S, params_A_min, params_S_min;
            compute_df ( estimates, & params_A, & params_S, & params_A_min, & params_S_min, em_wspace->temp_vecx, coeff_penalty, covar_penalty, tol, coeff_lambda_weights, groups );

            if ( params_S_min == estimates->D * ( estimates->D + 1 ) / 2 )
            {
                min_j_z = jl;
                continue;
            }
            if ( params_S_min <= estimates->D || lambda_j > lambda_j_max )
            {
                max_j_z = jl;
                //printf ( " Max covar lambda index: %d, lambda: %lf; stopping here\n", jl, lambda_j );
                //Discard S all zeros case?
                //break;
            }

            double vl = compute_ic ( em_wspace, estimates, Y_vd, ic_c, params_A, params_S, em_wspace->temp_vecm, em_wspace->temp_vecx );
            if ( gsl_isinf ( vl ) || gsl_isnan ( vl ) )
            {
                printf ( " Invalid likelihood value %lf; continuing. Index j: %d, lambda: %lf:\n", vl, jl, lambda_j );

                //printf_matrix("A", estimates->A);
                //printf_matrix("Pr", estimates->Pr);
                //printf_matrix("S", estimates->SigmaL);
                //printf("L%lf\n", estimates->loglik);
                //getchar();

                gsl_matrix_memcpy ( estimates->A, A_mle );
                gsl_matrix_memcpy ( estimates->Pr, Pr_mle );
                gsl_matrix_memcpy ( estimates->SigmaL, SL_mle );
                gsl_matrix_memcpy ( estimates->Sigma_inv, S_inv_mle );
                continue;
            }
            ic_arr[iq][jl] = vl;
            if ( vl > max_ic )
            {
                max_ic = vl;
                ( * covar_lambda_ind ) = jq = jl;
            }
        }

        ip = iq;
        lambda_j = ( jq > 0 ) * exp ( min_loglambda_covar + ( jq - 1 ) * coeff_loglambda_step );
        gsl_matrix_memcpy ( estimates->A, A_mle );
        gsl_matrix_memcpy ( estimates->Pr, Pr_mle );
        gsl_matrix_memcpy ( estimates->SigmaL, SL_mle );
        gsl_matrix_memcpy ( estimates->Sigma_inv, S_inv_mle );
        for ( int il = min_i_z; il < max_i_z; il += resol )
        {
            if ( ! gsl_isinf ( ic_arr[il][jq] ) )
            {
                if ( ic_arr[il][jq] > max_ic )
                {
                    max_ic = ic_arr[il][jq];
                    ( * coeff_lambda_ind ) = iq = il;
                }
                continue;
            }
            lambda_i = ( il > 0 ) * exp ( min_loglambda_coeff + ( il - 1 ) * coeff_loglambda_step );
            //printf (" i(%d): %lf, j(%d): %lf\n", il, lambda_i, jq, lambda_j );
            //gsl_matrix_memcpy ( estimates->A, A_mle );
            //gsl_matrix_memcpy ( estimates->Pr, Pr_mle );
            //gsl_matrix_memcpy ( estimates->SigmaL, SL_mle );
            //gsl_matrix_memcpy ( estimates->Sigma_inv, S_inv_mle );
            //Regularized
            expectation_maximization ( em_wspace, regu_wspace, regu_cov_wspace, estimates, Y_tr, X, y, em_max_iters, max_iters, tol, coeff_penalty, covar_penalty,
                                    lambda_i, lambda_j, coeff_lambda_weights, covar_lambda_weights, L_c, groups );
            int params_A, params_S, params_A_min, params_S_min;
            compute_df ( estimates, & params_A, & params_S, & params_A_min, & params_S_min, em_wspace->temp_vecx, coeff_penalty, covar_penalty, tol, coeff_lambda_weights, groups );

            if ( ( params_A_min == estimates->D * ( estimates->D * estimates->p + 1 ) && coeff_penalty != GROUP_ADALASSO ) || ( params_A_min == coeff_lambda_weights->size2 && coeff_penalty == GROUP_ADALASSO ) )
            {
                min_i_z = il;
                continue;
            }
            if ( ( params_A_min <= estimates->D && coeff_penalty != GROUP_ADALASSO ) || ( params_A_min == 0 && coeff_penalty == GROUP_ADALASSO ) || lambda_i > lambda_i_max )
            {
                max_i_z = il;
                //printf ( " Max coeff lambda index: %d, lambda: %lf; stopping here\n", il, lambda_i );
                //Discard A all zeros case? (Sigma can become singular)
                //break;
            }

            double vl = compute_ic ( em_wspace, estimates, Y_vd, ic_c, params_A, params_S, em_wspace->temp_vecm, em_wspace->temp_vecx );
            if ( gsl_isinf ( vl ) || gsl_isnan ( vl ) )
            {
                printf ( " Invalid likelihood value %lf; continuing. Index i: %d, lambda: %lf:\n", vl, il, lambda_i );

                //printf_matrix("A", estimates->A);
                //printf_matrix("Pr", estimates->Pr);
                //printf_matrix("S", estimates->SigmaL);
                //printf("L%lf\n", estimates->loglik);
                //getchar();

                gsl_matrix_memcpy ( estimates->A, A_mle );
                gsl_matrix_memcpy ( estimates->Pr, Pr_mle );
                gsl_matrix_memcpy ( estimates->SigmaL, SL_mle );
                gsl_matrix_memcpy ( estimates->Sigma_inv, S_inv_mle );
                continue;
            }
            ic_arr[il][jq] = vl;
            if ( vl > max_ic )
            {
                max_ic = vl;
                ( * coeff_lambda_ind ) = iq = il;
            }
        }

    } // while
    //printf ( " Indexes at %d with value %lf: (%d, %d)\n", resol, max_ic, * coeff_lambda_ind, * covar_lambda_ind );
    return;
}

double log_likelihood ( em_workspace * em_wspace, const gsl_matrix * Y, const msvar_parameters * estimates, size_t t_1, const gsl_vector * s_0 )
{
    // Forward recursion similar procedure essentially
    gsl_vector_view xi_vecview = gsl_matrix_column ( em_wspace->xi_filter, 0 );
    gsl_vector_view xi_vecview_tmp = gsl_matrix_column ( em_wspace->xi_tt1, 0 );
    if ( s_0 == NULL )
    {
        gsl_vector_set_zero ( & xi_vecview.vector );
        gsl_vector_set ( & xi_vecview.vector, estimates->initial_state - 1, 1.0 );
    }
    else
        gsl_blas_dcopy ( s_0, & xi_vecview.vector );
    double loglik = 0.0;
    for ( size_t t = t_1; t < Y->size1; ++t )
    {
        gsl_vector_const_view Y_vecview = gsl_matrix_const_row ( Y, t);
        for ( size_t m = 0; m < estimates->M; ++m )
        {
            double log_pdf;
            compute_ar_sum ( m, t, estimates, NULL, Y, em_wspace->mean_vec );
            gsl_matrix_const_view SigmaL_subview = gsl_matrix_const_submatrix ( estimates->SigmaL, estimates->switching_covar * m * estimates->D, 0, estimates->D, estimates->D );
            gsl_ran_multivariate_gaussian_log_pdf ( & Y_vecview.vector, em_wspace->mean_vec, & SigmaL_subview.matrix, & log_pdf, em_wspace->temp_vec );
            gsl_vector_set ( em_wspace->temp_vecm, m, GSL_MAX ( exp ( log_pdf ), 1.0e-300 ) );
        }
        gsl_blas_dgemv ( CblasTrans, 1.0, estimates->Pr, & xi_vecview.vector, 0.0, & xi_vecview_tmp.vector );
        gsl_blas_dcopy ( & xi_vecview_tmp.vector, & xi_vecview.vector );
        double val;
        gsl_blas_ddot ( em_wspace->temp_vecm, & xi_vecview.vector, & val );
        if ( val == 0.0 )
        {
            printf ( " (Division by zero in likelihood)\n" );
            return GSL_NEGINF;
        }
        gsl_vector_mul ( & xi_vecview.vector, em_wspace->temp_vecm );
        gsl_blas_dscal ( 1.0 / val, & xi_vecview.vector );
        loglik = loglik + log ( val );
    }
    return loglik;
}

void min_max_sym_eigenvalue ( gsl_matrix * M, gsl_vector * eval_vec, gsl_eigen_symm_workspace * eigw, double * lmin, double * lmax)
{
    gsl_eigen_symm ( M, eval_vec, eigw );
    gsl_vector_minmax ( eval_vec, lmin, lmax );
    return;
}

void msvar_parameters_alloc_matrices ( msvar_parameters * parameters )
{
    size_t n;
    parameters->Pr   = gsl_matrix_alloc ( parameters->M, parameters->M );
    if ( parameters->switching_coeff )
        n = parameters->M;
    else
        n = 1;
    parameters->A   = gsl_matrix_alloc ( n * parameters->D, parameters->D * parameters->p + 1 );
    if ( parameters->switching_covar )
        n = parameters->M;
    else
        n = 1;
    parameters->SigmaL = gsl_matrix_alloc ( n * parameters->D, parameters->D );
    parameters->Sigma_inv = gsl_matrix_alloc ( n * parameters->D, parameters->D );
    return;
}

void msvar_parameters_free_matrices ( msvar_parameters * parameters )
{
    gsl_matrix_free ( parameters->Pr );
    gsl_matrix_free ( parameters->A );
    gsl_matrix_free ( parameters->SigmaL );
    gsl_matrix_free ( parameters->Sigma_inv );
    return;
}

void em_workspace_alloc ( em_workspace * em_wspace, const msvar_parameters * estimates )
{
    em_wspace->xi_smooth = gsl_matrix_alloc ( estimates->M, estimates->T );
    em_wspace->xi_filter = gsl_matrix_alloc ( estimates->M, estimates->T );
    em_wspace->xi_tt1 = gsl_matrix_alloc ( estimates->M, estimates->T );

    //Forward and backward recursion  auxiliary memory
    em_wspace->temp_vec = gsl_vector_alloc ( estimates->D );
    em_wspace->mean_vec = gsl_vector_alloc ( estimates->D );
    em_wspace->temp_vecm = gsl_vector_alloc ( estimates->M );
    em_wspace->temp_vecx = gsl_vector_alloc ( estimates->M );

    // Transition probabs auxiliary memory
    if ( estimates->M > 1 )
    {
        em_wspace->tmpm = gsl_matrix_alloc ( estimates->M, estimates->T );
    }
    
    //Coefficients auxiliary memory
    em_wspace->tmp = gsl_matrix_alloc ( estimates->T, 1 + estimates->D * estimates->p );
    em_wspace->Ctmp = gsl_matrix_alloc ( 1 + estimates->D * estimates->p, 1 + estimates->D * estimates->p );
    //if ( estimates->M > 1 )
    //{
        em_wspace->C1 = gsl_matrix_alloc ( estimates->D * ( 1 + estimates->D * estimates->p ) , estimates->D * ( 1 + estimates->D * estimates->p )  );
        em_wspace->C2 = gsl_matrix_alloc ( estimates->D * ( 1 + estimates->D * estimates->p ) , estimates->D * estimates->T );
        em_wspace->b = gsl_vector_alloc ( estimates->D * ( 1 + estimates->D * estimates->p )  );
    //}
    //else
    //{
        em_wspace->Ctmp1HH = gsl_matrix_alloc ( 1 + estimates->D * estimates->p, 1 + estimates->D * estimates->p );
        em_wspace->Ctmp2 = gsl_matrix_alloc ( 1 + estimates->D * estimates->p, estimates->D );
    //}
    
    //Covariance auxiliary memory
    //if ( estimates->switching_covar )
    //{
        em_wspace->Usw = gsl_matrix_alloc ( estimates->T, estimates->D );
        em_wspace->tmpcsw = gsl_matrix_alloc ( estimates->T, estimates->D );
        em_wspace->eigw = gsl_eigen_symm_alloc ( estimates->D );
    //}
    //else
    //{
        em_wspace->U = gsl_matrix_alloc ( estimates->T * estimates->M, estimates->D );
        em_wspace->tmpc = gsl_matrix_alloc ( estimates->T * estimates->M, estimates->D );
    //}
    
    //EM  auxiliary memory
    
    em_wspace->estimates_prev = ( msvar_parameters * ) calloc ( 1, sizeof ( msvar_parameters )  );
    em_wspace->estimates_prev->D = estimates->D;
    em_wspace->estimates_prev->M = estimates->M;
    em_wspace->estimates_prev->T = estimates->T;
    em_wspace->estimates_prev->p = estimates->p;
    em_wspace->estimates_prev->initial_state = estimates->initial_state;
    em_wspace->estimates_prev->switching_coeff = estimates->switching_coeff;
    em_wspace->estimates_prev->switching_covar = estimates->switching_covar;
    msvar_parameters_alloc_matrices ( em_wspace->estimates_prev );
    
    return;
}

void em_workspace_free ( em_workspace * em_wspace, size_t M )
{
    gsl_matrix_free ( em_wspace->xi_smooth );
    gsl_matrix_free ( em_wspace->xi_filter );
    gsl_matrix_free ( em_wspace->xi_tt1 );
    
    gsl_vector_free ( em_wspace->temp_vec );
    gsl_vector_free ( em_wspace->mean_vec );
    gsl_vector_free ( em_wspace->temp_vecm );
    gsl_vector_free ( em_wspace->temp_vecx );
    
    if ( M > 1 )
        gsl_matrix_free ( em_wspace->tmpm );
    
    gsl_matrix_free ( em_wspace->Usw );
    gsl_matrix_free ( em_wspace->tmpcsw );
    
    gsl_matrix_free ( em_wspace->Ctmp );
    gsl_matrix_free ( em_wspace->tmp );
    //if ( M > 1 )
    //{
        gsl_matrix_free ( em_wspace->C1 );
        gsl_matrix_free ( em_wspace->C2 );
        gsl_vector_free ( em_wspace->b );
    //}
    //else
    //{
        gsl_matrix_free ( em_wspace->Ctmp1HH );
        gsl_matrix_free ( em_wspace->Ctmp2 );
    //}
    
    gsl_matrix_free ( em_wspace->U );
    gsl_matrix_free ( em_wspace->tmpc );
    
    gsl_eigen_symm_free ( em_wspace->eigw );
    
    msvar_parameters_free_matrices ( em_wspace->estimates_prev );
    free ( em_wspace->estimates_prev );
    return;
}

int read_alloc_groups ( const char * groups_file, gsl_matrix ** groups, size_t A_cols )
{
    int c;
    size_t rows = 0;
    FILE * f = fopen ( groups_file, "r" );
    if ( f == NULL )
    {
        printf ( " File %s: %s:\n", groups_file, " groups file " );
        GSL_ERROR ( " fopen (r) failed ", GSL_EINVAL );
    }
    else
    {
        do
        {
            c = getc ( f );
            if ( c == '\n')
                ++rows;
        } while ( c != EOF );
        fclose ( f );
        * groups = gsl_matrix_alloc ( rows, A_cols );
        fscanf_matrix ( groups_file, " A groups ", * groups );
        return GSL_SUCCESS;
    }
}

void printf_matrix ( const char * info, const gsl_matrix * m )
{
    printf ( " %s\n",info );
    for ( size_t i = 0; i < m->size1; ++i )
    {
        for ( size_t j = 0; j < m->size2; ++j )
        {
            printf ( "%lf ", gsl_matrix_get ( m,i,j )  );
        }
        printf ( "\n" );
    }
    return;
}

void printf_vector ( const char * info, const gsl_vector * m )
{
    printf ( " %s\n",info );
    for ( size_t i = 0; i < m->size; ++i )
    {
        printf ( "%lf ", gsl_vector_get ( m,i )  );
    }
    printf ( "\n" );
    return;
}

int fprintf_matrix ( const char * filename, const char * info, const gsl_matrix * m )
{
    //Prin gsl matrix to file
    FILE * f = fopen ( filename, "w" );
    if ( f == NULL )
    {
        printf ( " File %s: %s:\n", filename, info );
        GSL_ERROR ( " fopen (w) failed ", GSL_EINVAL );
    }
    else
    {
        // gsl_matrix_fprintf ( f, m, "%f" );
        for ( size_t i = 0; i < m->size1; ++i )
        {
            for ( size_t j = 0; j < m->size2; ++j )
            {
                fprintf ( f, "%lf ", gsl_matrix_get ( m,i,j )  );
            }
            fprintf ( f, "\n" );
        }
        fclose ( f );
        return GSL_SUCCESS;
    }
}

int fprintf_vector ( const char * filename, const char * info, const gsl_vector * m )
{
    //Print gsl vector to file
    FILE * f = fopen ( filename, "w" );
    if ( f == NULL )
    {
        printf ( " File %s: %s:\n", filename, info );
        GSL_ERROR ( " fopen (w) failed ", GSL_EINVAL );
    }
    else
    {
        // gsl_vector_fprintf ( f, m, "%f" );
        for ( size_t i = 0; i < m->size; ++i )
             fprintf ( f, "%lf\n", gsl_vector_get ( m,i )  );
        fclose ( f );
        return GSL_SUCCESS;
    }
}

int fscanf_matrix ( const char * filename, const char * info, gsl_matrix * m )
{
    //Scan gsl matrix from file
    FILE * f = fopen ( filename, "r" );
    if ( f == NULL )
    {   
        printf ( " File %s: %s:\n", filename, info );
        GSL_ERROR ( " fopen (r) failed ", GSL_EINVAL );
    }
    else
    {
        gsl_matrix_fscanf ( f, m );
        fclose ( f );
        return GSL_SUCCESS;
    }
}

int fscanf_vector ( const char * filename, const char * info, gsl_vector * m )
{
    //Scan gsl matrix from file
    FILE * f = fopen ( filename, "r" );
    if ( f == NULL )
    {   
        printf ( " File %s: %s:\n", filename, info );
        GSL_ERROR ( " fopen (r) failed ", GSL_EINVAL );
    }
    else
    {
        gsl_vector_fscanf ( f, m );
        fclose ( f );
        return GSL_SUCCESS;
    }
}

///******************** Reg functions **********************************/


void fista_workspace_alloc ( fista_workspace * fista_wspace, size_t T, size_t size1, size_t size2 )
{
    fista_wspace->Z = gsl_matrix_alloc ( size1 * size2, 1 );
    fista_wspace->B_prev = gsl_matrix_alloc ( size1 * size2, 1 );
    fista_wspace->B_hat_init = gsl_matrix_alloc ( size1 * size2, 1 );
    fista_wspace->XB = gsl_matrix_alloc ( T * size2, 1 );
    fista_wspace->XZ = gsl_matrix_alloc ( T * size2, 1 );
    fista_wspace->BmZ = gsl_matrix_alloc ( size1 * size2, 1 );
    
    fista_wspace->Grad = gsl_matrix_alloc ( size1 * size2, 1 );
    fista_wspace->XtY = gsl_matrix_alloc ( size1 * size2, 1 );
    fista_wspace->XtXZ = gsl_matrix_alloc ( size1 * size2, 1 );
    fista_wspace->XtX = gsl_matrix_alloc ( size1 * size2, size1 * size2 );
    
    return;
}

void fista_workspace_free ( fista_workspace * fista_wspace )
{
    gsl_matrix_free ( fista_wspace->Z );
    gsl_matrix_free ( fista_wspace->B_prev );
    gsl_matrix_free ( fista_wspace->B_hat_init );
    gsl_matrix_free ( fista_wspace->XB );
    gsl_matrix_free ( fista_wspace->XZ );
    gsl_matrix_free ( fista_wspace->BmZ );
    
    gsl_matrix_free ( fista_wspace->Grad );
    gsl_matrix_free ( fista_wspace->XtX );
    gsl_matrix_free ( fista_wspace->XtXZ );
    gsl_matrix_free ( fista_wspace->XtY );
    
    return;
}

void regu_cov_matrices_alloc ( regu_cov_matrices * regu_cov_mats, const msvar_parameters * parameters, size_t penalty )
{
    if ( INVCOVAR_GRAPHICAL(penalty) )
    {
        regu_cov_mats->W11 = gsl_matrix_alloc ( parameters->D - 1, parameters->D - 1 );
        regu_cov_mats->W = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->S_0 = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->Beta = gsl_matrix_alloc ( parameters->D - 1, parameters->D  );
        regu_cov_mats->beta_prev = gsl_matrix_alloc ( parameters->D - 1, 1 );
        regu_cov_mats->w = gsl_matrix_alloc ( parameters->D - 1, 1 );
        regu_cov_mats->w_prev = gsl_matrix_alloc ( parameters->D - 1, 1 );
        regu_cov_mats->s = gsl_matrix_alloc ( parameters->D - 1, 1 );
        regu_cov_mats->U = gsl_matrix_alloc ( parameters->D - 1, parameters->D - 1 );
        regu_cov_mats->Tmp = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->Th = gsl_matrix_alloc ( parameters->D - 1, parameters->D - 1 );
    }
    else if ( penalty != NONE )
    {
        regu_cov_mats->S = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->S_prev = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->S_0 = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->S_inv = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->S_0_inv = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->Tmp = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->Th = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->U = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->W = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->Beta = gsl_matrix_alloc ( parameters->D, parameters->D );
        regu_cov_mats->S_tmp = gsl_matrix_alloc ( parameters->D, parameters->D );
    }    
    return;
}

void regu_cov_matrices_free ( regu_cov_matrices * regu_cov_mats, size_t penalty )
{
    if ( INVCOVAR_GRAPHICAL ( penalty ) )
    {
        gsl_matrix_free (regu_cov_mats->W11);
        gsl_matrix_free (regu_cov_mats->W);
        gsl_matrix_free (regu_cov_mats->S_0);
        gsl_matrix_free (regu_cov_mats->Beta);
        gsl_matrix_free (regu_cov_mats->beta_prev);
        gsl_matrix_free (regu_cov_mats->w);
        gsl_matrix_free (regu_cov_mats->w_prev);
        gsl_matrix_free (regu_cov_mats->s);
        gsl_matrix_free (regu_cov_mats->U);
        gsl_matrix_free (regu_cov_mats->Tmp);
        gsl_matrix_free (regu_cov_mats->Th);
    }
    else if ( penalty != NONE )
    {
        gsl_matrix_free (regu_cov_mats->S);
        gsl_matrix_free (regu_cov_mats->S_prev);
        gsl_matrix_free (regu_cov_mats->S_0);
        gsl_matrix_free (regu_cov_mats->S_inv);
        gsl_matrix_free (regu_cov_mats->S_0_inv);
        gsl_matrix_free (regu_cov_mats->Tmp);
        gsl_matrix_free (regu_cov_mats->Th);
        gsl_matrix_free (regu_cov_mats->U);
        gsl_matrix_free (regu_cov_mats->W);
        gsl_matrix_free (regu_cov_mats->Beta);
        gsl_matrix_free (regu_cov_mats->S_tmp);
    }
    return;
}

void regu_workspace_alloc ( regu_workspace * regu_wspace, const msvar_parameters * parameters, size_t penalty, size_t g_size )
{
    regu_wspace->groups_tr = NULL;
    regu_wspace->A_mm = NULL;
    size_t r1 = parameters->D * parameters->p;
    size_t r2 = parameters->D;
    regu_wspace->Ys = gsl_matrix_alloc ( parameters->T, r2 );
    regu_wspace->E = gsl_matrix_alloc ( r1, r1 );
    regu_wspace->Li = gsl_matrix_alloc ( r2, r2 );
    regu_wspace->R = gsl_matrix_alloc ( parameters->T, 1 );
    
    regu_wspace->fista_wspace = ( fista_workspace * ) calloc ( 1, sizeof ( fista_workspace )  );
    //fista_workspace_alloc ( regu_wspace->fista_wspace, parameters->T, r1, r2 );
    fista_workspace_alloc ( regu_wspace->fista_wspace, parameters->T, r1, 1 );
    if ( penalty == GROUP_ADALASSO )
    {
        regu_wspace->lambda_weights_tr = gsl_matrix_alloc ( g_size, 1 );
        regu_wspace->groups_tr = gsl_matrix_alloc ( r1, g_size * 1 );
    }
    else
        regu_wspace->lambda_weights_tr = gsl_matrix_alloc ( r1, r2 );
    gsl_matrix_set_all( regu_wspace->lambda_weights_tr, 1.0 );
    regu_wspace->eval_vec = gsl_vector_alloc ( r1 );
    regu_wspace->eigw = gsl_eigen_symm_alloc ( r1 );
    
    return;
}

void regu_workspace_free ( regu_workspace * regu_wspace, const msvar_parameters * parameters )
{
    fista_workspace_free ( regu_wspace->fista_wspace );
    free ( regu_wspace->fista_wspace );

    gsl_matrix_free ( regu_wspace->Li );
    gsl_matrix_free ( regu_wspace->R );
    gsl_matrix_free ( regu_wspace->Ys );
    gsl_matrix_free ( regu_wspace->E );
    gsl_matrix_free ( regu_wspace->lambda_weights_tr );
    if ( regu_wspace->groups_tr != NULL )
        gsl_matrix_free ( regu_wspace->groups_tr );
    gsl_vector_free ( regu_wspace->eval_vec );
    gsl_eigen_symm_free ( regu_wspace->eigw );
    
    return;
}

void regu_cov_workspace_alloc ( regu_cov_workspace * regu_cov_wspace, const msvar_parameters * parameters, size_t penalty )
{
    regu_cov_wspace->S_mm = NULL;
    regu_cov_wspace->S_inv_mm = NULL;
    regu_cov_wspace->regu_cov_mats = ( regu_cov_matrices * ) calloc ( 1, sizeof ( regu_cov_matrices )  );
    regu_cov_matrices_alloc ( regu_cov_wspace->regu_cov_mats, parameters, penalty );
    //regu_cov_wspace->Se = gsl_matrix_alloc ( parameters->D, parameters->D );
    if ( ! INVCOVAR_GRAPHICAL(penalty) && penalty != NONE )
    {
        regu_cov_wspace->eval_vec = gsl_vector_alloc ( parameters->D );
        regu_cov_wspace->eigw = gsl_eigen_symm_alloc ( parameters->D );
        regu_cov_wspace->eigvw = gsl_eigen_symmv_alloc ( parameters->D );
    }
    return;
}

void regu_cov_workspace_free ( regu_cov_workspace * regu_cov_wspace, size_t penalty, const msvar_parameters * parameters )
{
    regu_cov_matrices_free ( regu_cov_wspace->regu_cov_mats, penalty );
    free ( regu_cov_wspace->regu_cov_mats );
    
    if ( ! INVCOVAR_GRAPHICAL(penalty) && penalty != NONE )
    {
        gsl_vector_free ( regu_cov_wspace->eval_vec );
        gsl_eigen_symm_free ( regu_cov_wspace->eigw );
        gsl_eigen_symmv_free ( regu_cov_wspace->eigvw );
    }
    //gsl_matrix_free ( regu_cov_wspace->Se );
    return;
}

void set_penalty_weights ( size_t D, const gsl_matrix * A, const gsl_matrix * SL, const gsl_matrix * S_inv, size_t coeff_penalty, size_t covar_penalty,
                         gsl_matrix * coeff_lambda_weights, gsl_matrix * covar_lambda_weights, const gsl_matrix * groups )
{
    if ( A != NULL && coeff_penalty != NONE )
        set_coeff_penalty_weights ( A, D, coeff_penalty, coeff_lambda_weights, 0.0, groups );
    if ( SL != NULL && S_inv != NULL && covar_penalty != NONE )
        set_covar_penalty_weights ( SL, S_inv, covar_penalty, covar_lambda_weights, 0.0 );
    return;
}

void set_coeff_penalty_weights ( const gsl_matrix * A, size_t D, size_t coeff_penalty, gsl_matrix * coeff_lambda_weights, double coeff_lambda, const gsl_matrix * groups )
{
    double gamma = 1.0; //0.5, 1.0, 2.0, as in Zou
    size_t M = A->size1 / D;
    if ( coeff_penalty == GROUP_ADALASSO )
    {
        for ( size_t m = 0; m < M; ++m )
        {
            gsl_matrix_const_view A_view = gsl_matrix_const_submatrix ( A, m * D, 1, D, A->size2 - 1 );
            for ( size_t g = 0; g < coeff_lambda_weights->size2; ++g )
            {
                gsl_matrix_const_view group_view = gsl_matrix_const_submatrix ( groups, g * D, 0, D, A->size2 - 1 );
                double s = 0.0;
                for ( size_t i = 0; i < ( & group_view.matrix )->size1; ++i )
                    for ( size_t j = 0; j < ( & group_view.matrix )->size2; ++j )
                    {
                        if ( gsl_matrix_get ( & group_view.matrix, i, j ) > 0 )
                        {
                            double v = gsl_matrix_get ( & A_view.matrix, i, j );
                            s += v * v;
                        }
                    }
                gsl_matrix_set ( coeff_lambda_weights, m, g, 1.0 / pow ( sqrt ( s ), gamma ) );
            }
        }
    }
    else if ( coeff_penalty == ADALASSO )
    {
        for ( size_t i = 0; i < A->size1; ++i )
            for ( size_t j = 1; j < A->size2; ++j )
                gsl_matrix_set ( coeff_lambda_weights, i, j - 1, 1.0 / pow ( fabs ( gsl_matrix_get ( A, i, j ) ), gamma ) );
    }
    else if ( coeff_penalty != NONE && coeff_penalty != ORACLE )
        gsl_matrix_set_all ( coeff_lambda_weights, 1.0 );
    
    return;
}

void set_covar_penalty_weights ( const gsl_matrix * SL, const gsl_matrix * S_inv, size_t covar_penalty, gsl_matrix * covar_lambda_weights, double covar_lambda )
{
    double a = PARAM_SCAD;
    double gamma = 1.0; //0.5, 1.0, 2.0, as in Zou
    double delta = PARAM_MCP;
    size_t D = SL->size2;
    size_t M = SL->size1 / D;
        
    if ( covar_penalty == LASSO_INVCOVAR || covar_penalty == LASSO || covar_penalty == GRAPHICAL_LASSO )
    {
        gsl_matrix_set_all ( covar_lambda_weights, 1.0 );
    }
    else if ( ( covar_penalty == ADALASSO_INVCOVAR || covar_penalty == GRAPHICAL_ADALASSO ) )
    {
        for ( size_t m = 0; m < M; ++m )
        {
            gsl_matrix_const_view S_inv_subview = gsl_matrix_const_submatrix ( S_inv, m * D, 0, D, D );
            for ( size_t i = 0; i < D; ++i )
            {
                double s = 1.0 / pow ( fabs ( gsl_matrix_get ( & S_inv_subview.matrix, i, i ) ), gamma );
                gsl_matrix_set ( covar_lambda_weights, m * D + i, i, s );
                for ( size_t j = 0; j < i; ++j )
                {
                    s = 1.0 / pow ( fabs ( gsl_matrix_get ( & S_inv_subview.matrix, i, j ) ), gamma );
                    gsl_matrix_set ( covar_lambda_weights, m * D + i, j, s );
                    gsl_matrix_set ( covar_lambda_weights, m * D + j, i, s );
                }
            }
        }
    }
    else if ( ( covar_penalty == GRAPHICAL_SCAD ) && covar_lambda > 0.0 )
    {
        //LLA (see paper "network exploration via scad" and "one step sparse estimates")
        //The weight is derivative on estimate

        for ( size_t m = 0; m < M; ++m )
        {
            gsl_matrix_const_view S_inv_subview = gsl_matrix_const_submatrix ( S_inv, m * D, 0, D, D );
            for ( size_t i = 0; i < D; ++i )
            {
                double w, lj;
                w = gsl_matrix_get ( & S_inv_subview.matrix, i, i );
                //The weight is scad derivative on estimate
                lj = covar_lambda;
                if ( fabs ( w ) > covar_lambda )
                {
                    lj = ( covar_lambda * a - fabs ( w ) ) / ( a - 1.0 );
                    if ( covar_lambda * a - fabs( w ) < 0.0 )
                        lj = 0.0;
                }
                gsl_matrix_set ( covar_lambda_weights, m * D + i, i, lj * GSL_SIGN ( 1.0 ) );
                for ( size_t j = 0; j < i; ++j )
                {
                    w = gsl_matrix_get ( & S_inv_subview.matrix, i, j );
                    lj = covar_lambda;
                    if ( fabs ( w ) > covar_lambda )
                    {
                        lj = ( covar_lambda * a - fabs ( w ) ) / ( a - 1.0 );
                        if ( covar_lambda * a - fabs( w ) < 0.0 )
                            lj = 0.0;
                    }
                    gsl_matrix_set ( covar_lambda_weights, m * D + i, j, lj * GSL_SIGN ( 1.0 ) );
                    gsl_matrix_set ( covar_lambda_weights, m * D + j, i, lj * GSL_SIGN ( 1.0 ) );
                }
            }
        }
    }
    else if ( ( covar_penalty == GRAPHICAL_MCP ) && covar_lambda > 0.0 )
    {
        //LLA (see paper "network exploration via scad" and "one step sparse estimates")
        //The weight is derivative on estimate
        for ( size_t m = 0; m < M; ++m )
        {
            gsl_matrix_const_view S_inv_subview = gsl_matrix_const_submatrix ( S_inv, m * D, 0, D, D );
            for ( size_t i = 0; i < D; ++i )
            {
                double w, lj;
                w = gsl_matrix_get ( & S_inv_subview.matrix, i, i );
                //The weight is mcp derivative on estimate
                lj = ( covar_lambda - fabs ( w ) / delta );
                if ( fabs (w) > delta * covar_lambda )
                    lj = 0;
                gsl_matrix_set ( covar_lambda_weights, m * D + i, i, lj  * GSL_SIGN ( 1.0 ) );
                for ( size_t j = 0; j < i; ++j )
                {
                    w = gsl_matrix_get ( & S_inv_subview.matrix, i, j );
                    lj = ( covar_lambda - fabs ( w ) / delta );
                    if ( fabs (w) > delta * covar_lambda )
                        lj = 0;
                    gsl_matrix_set ( covar_lambda_weights, m * D + i, j, lj * GSL_SIGN ( 1.0 ) );
                    gsl_matrix_set ( covar_lambda_weights, m * D + j, i, lj * GSL_SIGN ( 1.0 ) );
                }
            }
        }
    }
    else if ( covar_penalty == ADALASSO )
    {
        for ( size_t m = 0; m < M; ++m )
        {
            gsl_matrix_const_view SL_subview = gsl_matrix_const_submatrix ( SL, m * D, 0, D, D );
            gsl_matrix_view L_subview = gsl_matrix_submatrix ( covar_lambda_weights, m * D, 0, D, D );
            for ( size_t i = 0; i < D; ++i )
            {
                double s;
                gsl_vector_const_view SL_vecview = gsl_matrix_const_subrow ( & SL_subview.matrix, i, 0, i + 1 );
                gsl_blas_ddot ( & SL_vecview.vector, & SL_vecview.vector, & s );
                gsl_matrix_set ( & L_subview.matrix, i, i , 1.0 / pow ( fabs ( s ), gamma ) );
                for ( size_t j = 0; j < i; ++j )
                {
                    s = 1.0 / pow ( fabs ( gsl_matrix_get ( & SL_subview.matrix, j, i ) ), gamma );
                    gsl_matrix_set ( & L_subview.matrix, i, j, s );
                    gsl_matrix_set ( & L_subview.matrix, j, i, s );
                }
            }
        }
    }
    else if ( covar_penalty == GRAPHICAL_ORACLE )
    {
        if ( gsl_isinf ( gsl_matrix_max ( covar_lambda_weights ) ) )
            return;
        for ( size_t i = 0; i < covar_lambda_weights->size1; ++i )
        {
            for ( size_t j = 0; j < covar_lambda_weights->size2; ++j )
            {
                if ( gsl_matrix_get ( covar_lambda_weights, i, j ) > 0.0 )
                    gsl_matrix_set ( covar_lambda_weights, i, j, 0.0 );
                else
                    gsl_matrix_set ( covar_lambda_weights, i, j, GSL_POSINF );
            }
        }
    }
    return;
}

void fista_estimation ( fista_workspace * fista_wspace, const gsl_matrix * X, const gsl_matrix * Y, const gsl_matrix * groups, double lambda,
                        const gsl_matrix * lambda_weights, gsl_matrix * B_init_hat, double L_0, size_t max_iters, double tol, size_t penalty, double N )
{
    gsl_matrix_memcpy ( fista_wspace->Z, B_init_hat );
    gsl_matrix * B = B_init_hat;
    //printf_matrix ( " B ", B );

    gsl_blas_dgemm ( CblasTrans, CblasNoTrans, 1.0, X, Y, 0.0, fista_wspace->XtY );
    gsl_blas_dgemm ( CblasTrans, CblasNoTrans, 1.0, X, X, 0.0, fista_wspace->XtX );
    
    double alpha = 1.0;
    double alpha_next = 1.0;
    double o_f = GSL_POSINF;
    double o_f_prev = o_f;
    double a_param = PARAM_SCAD;
    double gamma_param = PARAM_MCP;
    double c_pen = 0.0;
    if ( penalty == SCAD )
        c_pen = 1.0 / ( a_param - 1.0 );
    else if ( penalty == MCP )
        c_pen = 1.0 / gamma_param;
    
    if ( L_0 < c_pen )
        printf (" Warning: function is not strictly convex\n");
    double L = GSL_MAX ( L_0, c_pen );
    //Theorem 4.4 of paper on strong + semi convex FISTA
    double s = 0.0;
    if ( c_pen == 0.0 )
        s = L;
    else if ( c_pen <= L / 2.0 )
        s = L - c_pen;
    else if ( c_pen < L )
        s = sqrt ( L * L - c_pen * c_pen );
    else if ( c_pen == L )
        s = 0.0;
    double L_a = ( 2.0 * c_pen + s ) / 2.0 + tol;
    
    int monotone = 1;
    double backtrack = 0.8; // in [0, 1)
    size_t const k_restart = GSL_MAX ( 400, max_iters / 4 );

    //If this is not zero, it becomes Loh and Wainwright
    c_pen = 0.0;
    
    int i;
    for ( i = 1; i < max_iters; ++i )
    {
        double L_bar = L_a;
        double o_f_tmp = o_f;
        double q_d = o_f_tmp;
        gsl_blas_dsymm ( CblasLeft, CblasUpper, 1.0, fista_wspace->XtX, fista_wspace->Z, 0, fista_wspace->XtXZ );
        // Grad = Xt X Z - XtY ( - c_pen Z)
        gsl_matrix_memcpy ( fista_wspace->Grad, fista_wspace->XtXZ );
        gsl_matrix_sub ( fista_wspace->Grad, fista_wspace->XtY );
        if ( c_pen > 0.0 )
        {
            gsl_matrix_scale ( fista_wspace->Grad, 1.0 / c_pen );
            gsl_matrix_sub ( fista_wspace->Grad, fista_wspace->Z );
            gsl_matrix_scale ( fista_wspace->Grad, c_pen );
        }
        if ( backtrack > 0.0 )
            gsl_blas_dgemm ( CblasNoTrans, CblasNoTrans, 1.0, X, fista_wspace->Z, 0.0, fista_wspace->XZ );
        do
        {
            // B  = Z - Grad / L_bar
            gsl_matrix_memcpy ( B, fista_wspace->Z );
            gsl_matrix_scale ( B, L_bar );
            gsl_matrix_sub ( B, fista_wspace->Grad );
            gsl_matrix_scale ( B, 1.0 / ( L_bar * ( 1.0 + c_pen / L_bar ) ) );
            proximal_operator ( B, penalty, lambda, lambda_weights, L_bar, a_param, gamma_param, c_pen, N, groups, 0 );
            o_f_tmp = ls_objective_function ( fista_wspace, B, Y, X, groups, lambda, lambda_weights, penalty, a_param, gamma_param, N );
            
            // Hasta aqui const step
            if ( backtrack > 0.0 )
            {
                //B-Z
                gsl_matrix_memcpy ( fista_wspace->BmZ, B );
                gsl_matrix_sub ( fista_wspace->BmZ, fista_wspace->Z );
                q_d =  Quad ( fista_wspace, B, Y, groups, lambda, lambda_weights, L_bar, penalty, a_param, gamma_param, N );
            }
            if ( o_f_tmp <= q_d || backtrack == 0.0 || 1.0 / L_bar < 1.0e-16 )
                break;
            L_bar = L_bar / backtrack;
        } while ( backtrack > 0.0 );

        L_a = L_bar;
        alpha = alpha_next;
        alpha_next = ( 1.0 + sqrt ( 1.0 + 4.0 * alpha * alpha )  ) / 2.0;
        gsl_matrix_memcpy ( fista_wspace->B_prev, fista_wspace->Z );
        //Monotone FISTA: return to previous point, update direction
        if ( monotone && o_f_tmp > o_f )
        {
            o_f_prev = o_f;
            //Z = B_prev + ( B - B_prev ) * ( alpha / alpha_next )
            //  = B * ( alpha / alpha_next ) + B_prev * ( 1.0 - alpha / alpha_next  );
            gsl_matrix_memcpy ( fista_wspace->Z, fista_wspace->B_prev );
            gsl_matrix_scale ( fista_wspace->Z, ( 1.0 - alpha / alpha_next) * ( alpha_next / alpha ) );
            gsl_matrix_add ( fista_wspace->Z, B );
            gsl_matrix_scale ( fista_wspace->Z, alpha / alpha_next );
        }
        else
        {
            o_f_prev = o_f;
            o_f = o_f_tmp;
            //Z = B + ( B - B_prev ) * ( alpha - 1 ) / alpha_next
            //  = B * ( 1 + ( alpha-1 ) /alpha_next ) - B_prev * ( alpha - 1 ) / alpha_next;
            gsl_matrix_memcpy ( fista_wspace->Z, fista_wspace->B_prev );
            gsl_matrix_scale ( fista_wspace->Z, ( ( 1.0 - alpha ) / alpha_next ) * ( 1.0 / ( 1.0 + ( alpha - 1.0 ) / alpha_next ) ) );
            gsl_matrix_add ( fista_wspace->Z, B );
            gsl_matrix_scale ( fista_wspace->Z, 1.0 + ( alpha - 1.0 ) / alpha_next );
        }
        // Convergence
        //if ( ( fabs ( o_f_prev - o_f ) < tol  && frobenius_norm_diff ( fista_wspace->Z, fista_wspace->B_prev ) < tol ) || 1.0 / L_a < 1.0e-16 )
        if ( ( fabs ( ( o_f_prev - o_f ) / o_f_prev ) < tol  && frobenius_norm_diff ( fista_wspace->Z, fista_wspace->B_prev ) / sqrt ( inner_product ( fista_wspace->B_prev, fista_wspace->B_prev ) ) < tol ) || 1.0 / L_a < 1.0e-16 )
            break;
        //"Fista with restart"
        if ( k_restart && ( i % k_restart == 0 ) )
            alpha_next = 1.0;
    }
    return; //B_hat is B
}

void covariance_optimization ( regu_cov_workspace * regu_cov_wspace, const gsl_matrix * Sigma, gsl_matrix * S_init, msvar_parameters * estimates, double lambda, const gsl_matrix * lambda_weights, size_t penalty, size_t max_iters, double tol, double L_0, double N )
{
    gsl_matrix * S = S_init;
    regu_cov_matrices * mats = regu_cov_wspace->regu_cov_mats;
    gsl_matrix_memcpy ( mats->S_inv, S );
    gsl_linalg_cholesky_decomp1 ( mats->S_inv );
    gsl_linalg_cholesky_invert ( mats->S_inv );
    
    double c = 0;
    //c = compute_penalty();
    for ( size_t i = 0; i < estimates->D; ++i )
    {
        double s = gsl_matrix_get ( Sigma, i, i );
        c += log ( s ) + 1.0;
    }
    //Lipschitz constant (matrix gets destroyed so let us copy)
    gsl_matrix_memcpy ( mats->S_0, Sigma );
    double l_min, l_max;
    min_max_sym_eigenvalue ( mats->S_0, regu_cov_wspace->eval_vec, regu_cov_wspace->eigw, & l_min, & l_max );
    c = c - ( estimates->D - 1 ) * ( log ( l_min ) + 1.0 );
    double delta = compute_min_delta ( l_min, c, max_iters, tol, 0 );
    //delta = tol;
    
    //double rho = 0.1; // Stepsize for alternating multipliers ( default value from spcov)
    double a = PARAM_SCAD;
    double gamma = PARAM_MCP;
    double c_pen = 0.0;
    if ( penalty == SCAD )
        c_pen = 1.0 / ( a - 1.0 );
    else if ( penalty == MCP )
        c_pen = 1.0 / gamma;
    
    double backtrack = 0.6;
    int const k_restart = max_iters;//GSL_MAX ( 400, max_iters / 4 );
    
    double F_val = GSL_POSINF;
    double pen_val = 0.0;
    int acceleration = 0; // 0, 1, 2
    
    double L_k = 1.0 / ( backtrack * backtrack );
    gsl_error_handler_t * current_handler = gsl_set_error_handler_off ( );
    /*
    double F_0;
    for (size_t its1 = 0; its1 < max_iters; ++its1 )
    {
        gsl_matrix_memcpy ( mats->S_0, S );
        F_0 = F_val;
    */
        int admm = 0;
        double t_0 = 1.0;
        double t_1;
        F_val = F_val - pen_val;
        int its2;
        for ( its2 = 0; its2 < max_iters; ++its2 )
        {
            gsl_matrix_memcpy ( mats->S_prev, S );
            double F_prev = F_val;
            //Set grad = S_inv - S_inv * Scov * S_inv - C * S (Loh and Wainwright)
            //(originally grad = S_0_inv - S_inv * Scov * S_inv (Bien and Tibshirani))
            //We have S_inv from previous iteration
            gsl_matrix_memcpy ( mats->Beta, mats->S_inv );
            gsl_blas_dgemm ( CblasNoTrans, CblasNoTrans, 1.0, Sigma, mats->S_inv, 0, mats->Tmp );
            gsl_blas_dgemm ( CblasNoTrans, CblasNoTrans, -1.0, mats->S_inv, mats->Tmp, 1.0, mats->Beta );
            if ( c_pen > 0.0 )
            {
                gsl_matrix_scale ( mats->Beta, 1.0 / c_pen );
                gsl_matrix_sub ( mats->Beta, S );
                gsl_matrix_scale ( mats->Beta, c_pen );
            }
            double inner_product_prev = inner_product ( mats->S_prev, mats->Beta );
            double Q_val;
            
            L_k = 1.0 / ( backtrack * backtrack );
            
            do
            {
                gsl_matrix_memcpy ( S, mats->S_prev );
                // Set (S - grad / L_k) / ( 1 + c_pen / L_k )
                gsl_matrix_scale ( S, L_k );
                gsl_matrix_sub ( S, mats->Beta );
                gsl_matrix_scale ( S, ( 1.0 / L_k ) * ( 1.0 / ( 1.0 + c_pen / L_k ) ) );
                proximal_operator ( S, penalty, lambda, lambda_weights, L_k, a, gamma, c_pen, N, NULL, 1 );
                //Check delta constraint
                F_val = 0.0;
                gsl_matrix_memcpy ( mats->S_tmp, S );
                gsl_eigen_symm ( mats->S_tmp, regu_cov_wspace->eval_vec, regu_cov_wspace->eigw );
                for ( size_t i = 0; i < estimates->D; ++i )
                {
                    double e = gsl_vector_get ( regu_cov_wspace->eval_vec, i);
                    if ( e < delta )
                    {
                        admm = 1;
                        //break;
                    }
                    F_val += log ( e );
                }
                gsl_matrix_memcpy ( mats->S_inv, S );
                int status = gsl_linalg_cholesky_decomp1 ( mats->S_inv );
                if ( status == GSL_EDOM || admm == 1 )
                {
                    admm = 0;
                    
                    L_k = L_k / backtrack;
                    if ( backtrack > 0.0 && 1.0 / L_k < 1.0e-16 )
                    {
                        gsl_matrix_memcpy ( S, mats->S_prev );
                        break;
                    }
                    continue;
                }
                
                gsl_linalg_cholesky_invert ( mats->S_inv );
                // (see also Beck and Tebouille, 2009)
                F_val = F_val + inner_product ( mats->S_inv, Sigma);
                Q_val = F_prev + inner_product ( S, mats->Beta ) - inner_product_prev
                        + gsl_pow_2 ( frobenius_norm_diff( S, mats->S_prev ) )  * L_k / 2.0;
                if ( F_val <= Q_val )
                    break;
                L_k = L_k / backtrack;
                if ( backtrack > 0.0 && 1.0 / L_k < 1.0e-16 )
                {
                    //printf ( " Small step size reached and not taken\n" );
                    break;
                }
            } while ( backtrack > 0.0 );
            
            if ( frobenius_norm_diff( S, mats->S_prev  ) / sqrt ( inner_product ( mats->S_prev, mats->S_prev ) ) < tol || 1.0 / L_k < 1.0e-16 )
            //if ( frobenius_norm_diff( S, mats->S_prev  ) < tol || 1.0 / L_k < 1.0e-16 )
            {
                //printf("it2: %d\n", its2);
                break;
            }
            
            //Acceleration
            double w = 0.0;
            if ( acceleration == 1 )
                w = ( its2 % k_restart ) / ( its2 % k_restart + 3.0 );
            else if ( acceleration == 2 )
            {
                t_1 = 0.5 * ( 1.0 + sqrt ( 1.0 + 4 * gsl_pow_2 ( t_0 ) ) );
                w = ( t_0 - 1.0 ) / t_1;
                if ( its2 % k_restart == 0 )
                    t_1 = 1.0;
                t_0 = t_1;
            }
            if ( w > 0.0 ) // S = S + w (S - S_prev) = (1 + w) S - w S_prev
            {
                //Accelerate
                gsl_matrix_scale ( S, ( 1.0 + w ) / w );
                gsl_matrix_sub ( S, mats->S_prev );
                gsl_matrix_scale ( S, w );
                gsl_matrix_memcpy ( mats->S_inv, S );
                gsl_linalg_cholesky_decomp1 ( mats->S_inv );
                gsl_linalg_cholesky_invert ( mats->S_inv );
                //Check delta constraint..?
            }
        }
        //printf ( " C iters: %d\n", its2 );
        //printf(" Step: %lf\n", L_k);
        //printf(" F: %lf\n", F_0);
    /*
        pen_val = compute_penalty ( penalty, S, lambda, lambda_weights, a, gamma, N, NULL );
        F_val = F_val + pen_val;
        if ( F_0 >= F_val )
        { 
            if ( frobenius_norm_diff( S, mats->S_0 ) / sqrt ( inner_product ( mats->S_0, mats->S_0 ) ) < tol  )
            //if ( frobenius_norm_diff( S, mats->S_0 )  < tol  )
            {
                //printf("it: %lu, F: %lf\n", its1, F_val);
                break;
            }
        }
        else
        {
            if ( acceleration > 0 )
            {
                //printf ( " Deactivating acceleration ");
                acceleration = 0;
                gsl_matrix_memcpy ( S, mats->S_0 );
                gsl_matrix_memcpy ( mats->S_inv, S );
                gsl_linalg_cholesky_decomp1 ( mats->S_inv );
                gsl_linalg_cholesky_invert ( mats->S_inv );
            }
            else
            {
                //printf ( " Iter %lu: objective did not decrease further\n", its1);
                //gsl_matrix_memcpy ( S, Sigma );
                gsl_matrix_memcpy ( S, mats->S_0 );
                break;
            }
        }
    }
    */
    gsl_set_error_handler ( current_handler );
    //printf("Lk %lf\n", L_k);
    return;
}

void precision_optimization ( regu_cov_workspace * regu_cov_wspace, const gsl_matrix * Sigma, gsl_matrix * S_init, msvar_parameters * estimates, double lambda, const gsl_matrix * lambda_weights, size_t penalty, size_t max_iters, double tol, double L_0, double N )
{
    gsl_matrix * S = S_init;
    regu_cov_matrices * mats = regu_cov_wspace->regu_cov_mats;
    
    gsl_matrix_memcpy ( mats->S_0, Sigma );
    double l_min, l_max;
    min_max_sym_eigenvalue ( mats->S_0, regu_cov_wspace->eval_vec, regu_cov_wspace->eigw, & l_min, & l_max );
    gsl_matrix_memcpy ( mats->S_0, Sigma );
    gsl_linalg_cholesky_decomp1 ( mats->S_0 );
    gsl_linalg_cholesky_invert ( mats->S_0 );
    double c = 0;
    //c = compute_penalty();
    for ( size_t i = 0; i < estimates->D; ++i )
    {
        double s = gsl_matrix_get ( mats->S_0, i, i );
        c += -1.0 * log ( s ) + s * gsl_matrix_get ( Sigma, i, i );
    }
    c = c - ( estimates->D - 1 ) * ( log ( l_min ) + 1.0 );
    double delta = compute_min_delta ( l_min, c, max_iters, tol, 1 );
    //delta = tol;
    
    //double rho = 0.1; // Stepsize for alternating multipliers ( default value from spcov)
    double a = PARAM_SCAD;
    double gamma = PARAM_MCP;
    double c_pen = 0.0;
    if ( penalty == SCAD_INVCOVAR )
        c_pen = 1.0 / ( a - 1.0 );
    else if ( penalty == MCP_INVCOVAR )
        c_pen = 1.0 / gamma;
    
    double backtrack = 0.6;
    int const k_restart = max_iters;//GSL_MAX ( 400, max_iters / 4 );
    
    double F_val = GSL_POSINF;
    double pen_val = 0.0;
    int acceleration = 0; // 0, 1, 2
    
    double L_k = 1.0 / ( backtrack * backtrack );
    gsl_error_handler_t * current_handler = gsl_set_error_handler_off ( );
    
    // Throughout, S is the precision matrix estimate
    /*
    double F_0;
    for (size_t its1 = 0; its1 < max_iters; ++its1 )
    {
        gsl_matrix_memcpy ( mats->S_0, S );
        F_0 = F_val;
    */
        int admm = 0;
        double t_0 = 1.0;
        double t_1;
        F_val = F_val - pen_val;
        int its2;
        for ( its2 = 0; its2 < max_iters; ++its2 )
        {
            gsl_matrix_memcpy ( mats->S_prev, S );
            double F_prev = F_val;
            //Set grad = -S_inv + Scov - C * S (Loh and Wainwright), S is the precision matrix
            gsl_matrix_memcpy ( mats->Beta, S );
            gsl_linalg_cholesky_decomp1 ( mats->Beta );
            gsl_linalg_cholesky_invert ( mats->Beta );
            if ( c_pen > 0.0 )
            {
                gsl_matrix_scale ( mats->Beta, 1.0 / c_pen);
                gsl_matrix_add ( mats->Beta, S );
                gsl_matrix_scale ( mats->Beta, c_pen);
            }
            gsl_matrix_sub ( mats->Beta, Sigma );
            gsl_matrix_scale ( mats->Beta, -1.0 );
            double inner_product_prev = inner_product ( mats->S_prev, mats->Beta );
            double Q_val;
            
            L_k = 1.0 / ( backtrack * backtrack );
            
            do
            {
                gsl_matrix_memcpy ( S, mats->S_prev );
                // Set (S - grad / L_k) / ( 1 + c_pen / L_k )
                gsl_matrix_scale ( S, L_k );
                gsl_matrix_sub ( S, mats->Beta );
                gsl_matrix_scale ( S, ( 1.0 / L_k ) * ( 1.0 / ( 1.0 + c_pen / L_k ) ) );
                proximal_operator ( S, penalty, lambda, lambda_weights, L_k, a, gamma, c_pen, N, NULL, 1 );
                //Check delta constraint
                F_val = 0.0;
                gsl_matrix_memcpy ( mats->S_tmp, S );
                gsl_eigen_symm ( mats->S_tmp, regu_cov_wspace->eval_vec, regu_cov_wspace->eigw );
                for ( size_t i = 0; i < estimates->D; ++i )
                {
                    double e = gsl_vector_get ( regu_cov_wspace->eval_vec, i);
                    if ( e < delta )
                    {
                        admm = 1;
                        //break;
                    }
                    F_val -= log ( e );
                }

                gsl_matrix_memcpy ( mats->S_tmp, S );
                int status = gsl_linalg_cholesky_decomp1 ( mats->S_tmp );
                if ( status == GSL_EDOM || admm == 1 )
                {
                    admm = 0;
                    
                    L_k = L_k / backtrack;
                    if ( backtrack > 0.0 && 1.0 / L_k < 1.0e-16 )
                    {
                        gsl_matrix_memcpy ( S, mats->S_prev );
                        break;
                    }
                    continue;
                    
                }
                
                // (see also Beck and Tebouille, 2009)
                F_val = F_val + inner_product ( S, Sigma);
                Q_val = F_prev + inner_product ( S, mats->Beta ) - inner_product_prev
                        + gsl_pow_2 ( frobenius_norm_diff( S, mats->S_prev ) )  * L_k / 2.0;
                if ( F_val <= Q_val || backtrack == 0.0 )
                    break;
                L_k = L_k / backtrack;
                if ( backtrack > 0.0 && 1.0 / L_k < 1.0e-16 )
                {
                    //printf ( " Small step size reached and not taken\n" );
                    break;
                }
            } while ( backtrack > 0.0 );
            
            if ( frobenius_norm_diff( S, mats->S_prev ) / sqrt ( inner_product ( mats->S_prev, mats->S_prev ) ) < tol || 1.0 / L_k < 1.0e-16 )
            //if ( frobenius_norm_diff( S, mats->S_prev  ) < tol || 1.0 / L_k < 1.0e-16 )
            {
                //printf("it2: %d\n", its2);
                break;
            }
            
            //Acceleration
            double w = 0.0;
            if ( acceleration == 1 )
                w = ( its2 % k_restart ) / ( its2 % k_restart + 3.0 );
            else if ( acceleration == 2 )
            {
                t_1 = 0.5 * ( 1.0 + sqrt ( 1.0 + 4 * gsl_pow_2 ( t_0 ) ) );
                w = ( t_0 - 1.0 ) / t_1;
                if ( its2 % k_restart == 0 )
                    t_1 = 1.0;
                t_0 = t_1;
            }
            if ( w > 0.0 ) // S = S + w (S - S_prev) = (1 + w) S - w S_prev
            {
                //Accelerate
                gsl_matrix_scale ( S, ( 1.0 + w ) / w );
                gsl_matrix_sub ( S, mats->S_prev );
                gsl_matrix_scale ( S, w );
                
                //Check delta constraint..?
            }
        }
        //printf ( " C iters: %d\n", its2 );
        //printf(" Step: %lf\n", L_k);
        //printf(" F: %lf\n", F_0);
    /*
        pen_val = compute_penalty ( penalty, S, lambda, lambda_weights, a, gamma, N, NULL );
        F_val = F_val + pen_val;
        if ( F_0 >= F_val )
        { 
            if ( frobenius_norm_diff( S, mats->S_0 ) / sqrt ( inner_product ( mats->S_0, mats->S_0 ) ) < tol  )
            //if ( frobenius_norm_diff( S, mats->S_0 )  < tol  )
            {
                //printf("it: %lu, F: %lf\n", its1, F_val);
                break;
            }
        }
        else
        {
            if ( acceleration > 0 )
            {
                //printf ( " Deactivating acceleration ");
                acceleration = 0;
                gsl_matrix_memcpy ( S, mats->S_0 );
            }
            else
            {
                //printf ( " Iter %lu: objective did not decrease further\n", its1);
                //gsl_matrix_memcpy ( S, Sigma );
                gsl_matrix_memcpy ( S, mats->S_0 );
                break;
            }
        }
    }
    */
    gsl_set_error_handler ( current_handler );
    //printf("Lk %lf\n", L_k);
    //* L_0 = L_k;
    return;
}

double h ( double x, double a, int is_precision )
{
    if ( ! is_precision )
        return log ( x ) + a / x;
    else
        return -1.0 * log ( x ) + a * x;
}
double dh ( double x, double a, int is_precision )
{
    if ( ! is_precision )
        return ( 1.0 - a / x ) / x;
    return -1.0 / x + a;
}

double compute_min_delta ( double a, double c, size_t max_iters, double tol, int is_precision )
{
    if ( ! is_precision )
    {
        // Adapted from spcov package
        if ( h ( a, a, is_precision ) > c )
            return tol; //return GSL_NAN;
        if ( h ( a, a, is_precision ) == c )
            return a;
        
        double e = 1e-4; // sqrt ( tol );
        double x = e * a;
        for ( size_t i = 0; i < max_iters; ++i )
        {
            x = x - ( h ( x, a, is_precision ) - c ) / dh ( x, a, is_precision );
            if ( x <= 0.0 )
                x = e * a;
            else if ( x >= a )
                x = ( 1.0 - e ) * a;
            if ( fabs ( h ( x, a, is_precision ) - c ) < tol && h ( x, a, is_precision ) >= c )
                break;
        }
        x = GSL_MAX ( x, tol );
        if ( gsl_isinf ( x ) || gsl_isnan ( x ) )
            x = tol;
        return x;
    }
    else
    {
        //Here min h is at 1/a
        if ( h ( 1.0 / a, a, is_precision ) > c )
            return tol; //return GSL_NAN;
        if ( h ( 1.0 / a, a, is_precision ) == c )
            return 1.0 / a;
        
        double e = 1e-4; // sqrt ( tol );
        double x = e * ( 1.0 / a);
        for ( size_t i = 0; i < max_iters; ++i )
        {
            x = x - ( h ( x, a, is_precision ) - c ) / dh ( x, a, is_precision );
            if ( x <= 0.0 )
                x = e * ( 1.0 / a );
            else if ( x >= 1.0 / a )
                x = ( 1.0 - e ) * ( 1.0 / a );
            if ( fabs ( h ( x, a, is_precision ) - c ) < tol && h ( x, a, is_precision ) >= c )
                break;
        }
        x = GSL_MAX ( x, tol );
        if ( gsl_isinf ( x ) || gsl_isnan ( x ) )
            x = tol;
        return x;
    }
}

void proximal_operator ( gsl_matrix * B, size_t penalty, double lambda, const gsl_matrix * lambda_weights, double eta, double a_param, double gamma_param,
                         double c_pen, double N, const gsl_matrix * groups, int is_covar )
{
    if ( penalty == LASSO || penalty == ADALASSO || penalty == LASSO_INVCOVAR || penalty == ADALASSO_INVCOVAR || penalty == ORACLE || penalty == ORACLE_INVCOVAR )
        p_operator_lasso ( B, lambda, lambda_weights, eta, N, is_covar );
    else if ( penalty == GROUP_ADALASSO )
        p_operator_group_lasso ( B, lambda, lambda_weights, eta, N, groups );
    else if ( penalty == SCAD || penalty == SCAD_INVCOVAR )
        p_operator_scad ( B, lambda, eta, a_param, c_pen, N, is_covar );
    else if ( penalty == MCP || penalty == MCP_INVCOVAR )
        p_operator_mcp ( B, lambda, eta, gamma_param, c_pen, N, is_covar );
    return;
}

void p_operator_lasso ( gsl_matrix * B_p, double lambda, const gsl_matrix * lambda_weights, double eta, double N, int is_covar )
{
    for ( size_t i = 0; i < B_p->size1; ++i )
        for ( size_t j = 0; j < B_p->size2; ++j )
        {
            if ( is_covar && i == j )
                continue;
            double b, w, v;
            int s;
            b = gsl_matrix_get ( B_p, i, j );
            w = gsl_matrix_get ( lambda_weights, i, j );
            if ( b == 0.0 || w == 0.0 )
                continue;
            s = GSL_SIGN ( b );
            v =  fabs ( b ) - lambda * w / ( eta / N );
            if ( v <= 0.0 || gsl_isinf ( w ) )
                gsl_matrix_set ( B_p, i, j, 0.0 );
            else
                gsl_matrix_set ( B_p, i, j, s * v );
        }
    return;
}

void p_operator_scad ( gsl_matrix * B_p, double lambda, double eta, double a, double c_pen, double N, int is_covar )
{
    double nu = 1.0 / ( eta / N + c_pen );
    for ( size_t i = 0; i < B_p->size1; ++i )
        for ( size_t j = 0; j < B_p->size2; ++j )
        {
            if ( is_covar && i == j )
                continue;
            double b;
            int s;
            b = gsl_matrix_get ( B_p, i, j );
            s = GSL_SIGN ( b );
            if ( fabs ( b ) <= nu * lambda )
            {
                gsl_matrix_set ( B_p, i, j, 0.0 );
            }
            else if ( fabs ( b ) <= ( nu + 1.0 ) * lambda )
            {
                gsl_matrix_set ( B_p, i, j, b - s * lambda * nu );
            }
            else if ( fabs ( b ) <= a * lambda )
            {
                gsl_matrix_set ( B_p, i, j, ( ( a - 1 ) * b - s * a * nu * lambda ) / ( a - 1.0 - nu ) );
            }
        }
    return;
}

void p_operator_mcp ( gsl_matrix * B_p, double lambda, double eta, double gamma, double c_pen, double N, int is_covar )
{
    double nu = 1.0 / ( eta / N + c_pen );
    for ( size_t i = 0; i < B_p->size1; ++i )
        for ( size_t j = 0; j < B_p->size2; ++j )
        {
            if ( is_covar && i == j )
                continue;
            double b;
            int s;
            b = gsl_matrix_get ( B_p, i, j );
            s = GSL_SIGN ( b );
            if ( fabs ( b ) <= gamma * lambda )
            {
                if ( fabs ( b ) <= nu * lambda )
                    gsl_matrix_set ( B_p, i, j, 0.0 );
                else
                    gsl_matrix_set ( B_p, i, j, ( gamma * b - s * gamma * nu * lambda ) / ( gamma - nu ) );
            }
        }
    return;
}

void p_operator_group_lasso ( gsl_matrix * B_p, double lambda, const gsl_matrix * lambda_weights, double eta, double N, const gsl_matrix * groups )
{
    for ( size_t g = 0; g < lambda_weights->size1; ++g )
    {
        gsl_matrix_const_view group_view = gsl_matrix_const_submatrix ( groups, 0, g * B_p->size2, B_p->size1, B_p->size2 );
        double s = 0.0;
        for ( size_t i = 0; i < B_p->size1; ++i )
            for ( size_t j = 0; j < B_p->size2; ++j )
            {
                if ( gsl_matrix_get ( & group_view.matrix, i, j ) > 0 )
                {
                    double v = gsl_matrix_get ( B_p, i, j );
                    s += v * v;
                }
            }
        double w = gsl_matrix_get ( lambda_weights, g, 0 );
        if ( s == 0.0 || w == 0.0 )
            continue;
        double c = 1.0  - lambda * w / ( s * eta / N );
        for ( size_t i = 0; i < B_p->size1; ++i )
            for ( size_t j = 0; j < B_p->size2; ++j )
            {
                if ( gsl_matrix_get ( & group_view.matrix, i, j ) > 0 )
                {
                    if ( c <= 0.0 || gsl_isinf ( w ) )
                        gsl_matrix_set ( B_p, i, j, 0.0 );
                    else
                        gsl_matrix_set ( B_p, i, j, c * gsl_matrix_get ( B_p, i, j ) );
                }
            }
    }
    return;
}

double Quad ( fista_workspace * fista_wspace, const gsl_matrix * B, const gsl_matrix * Y, const gsl_matrix * groups,
              double lambda, const gsl_matrix * lambda_weights, double t, size_t penalty, double a, double gamma, double N )
{   
    double fy, fb, Pv;
    fb = sqrt ( inner_product ( fista_wspace->BmZ, fista_wspace->BmZ )  );
    fy = frobenius_norm_diff ( Y, fista_wspace->XZ );
    /*  Pv = fy * fy / 2 + inner_product ( B, Grad ) - inner_product ( Z, Grad ) + fb * fb * t / 2;
           = fy * fy / 2 + inner_product ( B - Z, Grad ) + fb * fb * t / 2;
    */
    Pv = fy * fy / 2.0 + inner_product ( fista_wspace->BmZ, fista_wspace->Grad ) + fb * fb * t / 2;
    return ( Pv + compute_penalty ( penalty, B, lambda, lambda_weights, a, gamma, N, groups, 0 ) );
}

double ls_objective_function ( fista_workspace * fista_wspace, const gsl_matrix * B, const gsl_matrix * Y, const gsl_matrix * X, const gsl_matrix * groups,
                            double lambda, const gsl_matrix * lambda_weights, size_t penalty, double a, double gamma, double N )
{
    gsl_blas_dgemm ( CblasNoTrans, CblasNoTrans, 1.0, X, B, 0.0, fista_wspace->XB );
    double f = frobenius_norm_diff ( Y, fista_wspace->XB );
    return ( f * f / 2.0 + compute_penalty ( penalty, B, lambda, lambda_weights, a, gamma, N, groups, 0 ) );
}

//
//
/////


